
Require Decidable.

Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import Formation.
Require Import Closure.


Inductive skind : Set :=
| sk_type  : skind
| sk_pi    : skind -> skind -> skind
| sk_sigma : skind -> skind -> skind
| sk_one   : skind
.


Fixpoint simp (k : (* kd *) term) {struct k} : skind :=
  (match k with
   | kd_type => sk_type
   | kd_sing c => sk_type
   | kd_pi k1 k2 => sk_pi (simp k1) (simp k2)
   | kd_sigma k1 k2 => sk_sigma (simp k1) (simp k2)
   | _ => sk_one
   end).


Definition trelation := (* tm *) term -> (* tm *) term -> Prop.


(* The paper calls these pre-candidates. *)
Fixpoint precand (sk : skind) {struct sk} : Type :=
  (match sk with
   | sk_type => trelation
       
   | sk_pi sk1 sk2 => 
       (* cn *) term -> (* cn *) term -> precand sk1 -> precand sk2

   | sk_sigma sk1 sk2 =>
       (precand sk1 * precand sk2)%type

   | sk_one =>
       unit
   end).


(* The paper calls these pre-candidates as well.  That is, it makes
   no distinction between these and pre-candidates.
*)
Inductive cand :=
  Cand (sk : skind) : precand sk -> cand.


Definition per := (* cn *) term -> (* cn *) term -> cand -> cand -> Prop.


Definition mrelation := (* md *) term -> (* md *) term -> cand -> Prop.


Inductive entry :=
| Dyn : term -> term -> entry
| Sta : (* cn *) term -> (* cn *) term -> cand -> entry.


Definition env := list entry.


Fixpoint maybe_cast (sk sk' : skind) (A : skind -> Type) : option (A sk -> A sk') :=
  (match sk with
   | sk_type =>
       (match sk' with
        | sk_type => Some (fun x => x)
        | _ => None
        end)

   | sk_pi sk1 sk2 =>
       (match sk' with
        | sk_pi sk1' sk2' =>
            (match maybe_cast sk1 sk1' (fun s => A (sk_pi s sk2)) with
             | None => None
             | Some f =>
                 (match maybe_cast sk2 sk2' (fun s => A (sk_pi sk1' s)) with
                  | None => None
                  | Some g => Some (fun x => g (f x))
                  end)
             end)

        | _ => None
        end)

   | sk_sigma sk1 sk2 =>
       (match sk' with
        | sk_sigma sk1' sk2' =>
            (match maybe_cast sk1 sk1' (fun s => A (sk_sigma s sk2)) with
             | None => None
             | Some f =>
                 (match maybe_cast sk2 sk2' (fun s => A (sk_sigma sk1' s)) with
                  | None => None
                  | Some g => Some (fun x => g (f x))
                  end)
             end)

        | _ => None
        end)

   | sk_one =>
       (match sk' with
        | sk_one => Some (fun x => x)
        | _ => None
        end)
   end).


Fixpoint bogus_precand (sk : skind) : precand sk :=
  (match sk with
   | sk_type => (fun _ _ => False)
   | sk_pi sk1 sk2 => (fun _ _ _ => bogus_precand sk2)
   | sk_sigma sk1 sk2 => (bogus_precand sk1, bogus_precand sk2)
   | sk_one => tt
   end).


Definition cast (sk sk' : skind) (Q : precand sk) : precand sk' :=
  (match maybe_cast sk sk' precand with
   | None => bogus_precand sk'
   | Some f => f Q
   end).


Definition out (sk : skind) (Q : cand) : precand sk :=
  (match Q with
   | Cand sk' Q' => cast sk' sk Q'
   end). 


Fixpoint left (E : env) {struct E} :=
  (match E with
   | nil => id
   | E'; Dyn m _ => dot m (left E')
   | E'; Sta c _ _ => dot c (left E')
   end).


Fixpoint right (E : env) {struct E} :=
  (match E with
   | nil => id
   | E'; Dyn _ m => dot m (right E')
   | E'; Sta _ c _ => dot c (right E')
   end).


Lemma eq_skind_dec :
  forall (sk sk' : skind),
    Decidable.decidable (sk = sk').
Proof.
intros sk.
induct sk.

(* type *)
intro sk'.
destruct sk'; try (right; intro; discriminate); [].
left; reflexivity.

(* pi *)
intros sk1 IH1 sk2 IH2 sk'.
revert IH1 IH2.
case sk'; try (intros; right; intro; discriminate); [].
intros sk1' sk2' IH1 IH2.
destruct (IH1 sk1') as [Heq | Hneq].
  subst sk1'.
  destruct (IH2 sk2') as [Heq | Hneq].
    subst sk2'.
    left; reflexivity.

    right.
    contradict Hneq.
    injection Hneq; auto; done.

  right.
  contradict Hneq.
  injection Hneq; auto; done.

(* sigma *)
intros sk1 IH1 sk2 IH2 sk'.
revert IH1 IH2.
case sk'; try (intros; right; intro; discriminate); [].
intros sk1' sk2' IH1 IH2.
destruct (IH1 sk1') as [Heq | Hneq].
  subst sk1'.
  destruct (IH2 sk2') as [Heq | Hneq].
    subst sk2'.
    left; reflexivity.

    right.
    contradict Hneq.
    injection Hneq; auto; done.

  right.
  contradict Hneq.
  injection Hneq; auto; done.

(* one *)
intro sk'.
destruct sk'; try (right; intro; discriminate); [].
left; reflexivity.
Qed.


Lemma subst_simp :
  forall s k, hkd k -> simp (subst s k) = simp k.
Proof.
intros s k Hkd.
revert s.
induct Hkd; intros; simp_sub; simpl; try f_equal; auto.
Qed.


Lemma eqk_simp :
  forall G k k',
    eqk G k k' -> simp k = simp k'.
Proof.
intros G k k' Heqk.
induct Heqk.

(* refl *)
intros; reflexivity.

(* symm *)
intros; symmetry; assumption.

(* trans *)
intros; etransitivity; eassumption.

(* sing *)
intros; simpl; reflexivity.

(* pi *)
intros; simpl; f_equal; auto; done.

(* sigma *)
intros; simpl; f_equal; auto; done.
Qed.


Lemma subk_simp :
  forall G k k',
    subk G k k' -> simp k = simp k'.
Proof.
intros G k k' Hsubk.
induct Hsubk.

(* refl *)
intros.
eapply eqk_simp; eauto; done.

(* trans *)
intros; etransitivity; eassumption.

(* sing_t *)
intros; simpl; reflexivity.

(* pi *)
intros; simpl; f_equal; auto; done.

(* sigma *)
intros; simpl; f_equal; auto; done.
Qed.


Lemma maybe_cast_id :
  forall sk A, maybe_cast sk sk A = Some (fun x => x).
Proof.
intros sk.
induct sk.

(* type *)
intros A.
simpl.
reflexivity.

(* pi *)
intros sk1 IH1 sk2 IH2 A.
simpl.
rewrite -> IH1; [].
rewrite -> IH2; [].
reflexivity.

(* sigma *)
intros sk1 IH1 sk2 IH2 A.
simpl.
rewrite -> IH1; [].
rewrite -> IH2; [].
reflexivity.

(* one *)
intros A.
simpl.
reflexivity.
Qed.


Lemma cast_id :
  forall sk Q, cast sk sk Q = Q.
Proof.
intros sk Q.
unfold cast.
rewrite -> maybe_cast_id; [].
reflexivity.
Qed.


Lemma cand_inject :
  forall sk Q Q',
    Cand sk Q = Cand sk Q'
    -> Q = Q'.
Proof.
intros sk Q Q' Heq.
f_apply (out sk) in Heq as Heq'.
unfold out in Heq'.
rewrite -> ! cast_id in Heq'.
assumption.
Qed.


Lemma cand_ext_1 :
  forall sk sk' Q Q',
    Cand sk Q = Cand sk' Q'
    -> sk = sk'.
Proof.
intros sk sk' Q Q'.
cut (forall Z Z', Z = Z' -> match Z with Cand x _ => x end = match Z' with Cand x _ => x end).
  intros H Heq.
  exact (H _ _ Heq).
intros Z Z' HeqZ.
subst Z'.
reflexivity.
Qed.


Lemma cand_ext_2 :
  forall sk sk' Q,
    sk = sk'
    -> Cand sk Q = Cand sk' (cast sk sk' Q).
Proof.
intros sk sk' Q H.
subst sk'.
rewrite -> cast_id.
reflexivity.
Qed.


Definition csk (Q : cand) :=
  (match Q with Cand sk _ => sk end).


Lemma cand_eta :
  forall Q sk,
    csk Q = sk
    -> Cand sk (out sk Q) = Q.
Proof.
intros Q sk <-.
destruct Q as [sk Z].
simpl.
rewrite -> cast_id; [].
reflexivity.
Qed.


Lemma cand_eta' :
  forall Q sk Z,
    Cand sk Z = Q
    -> Cand sk (out sk Q) = Q.
Proof.
intros Q sk Z <-.
simpl.
rewrite -> cast_id; [].
reflexivity.
Qed.
    

Lemma cand_beta :
  forall sk (Z : precand sk),
    out sk (Cand sk Z) = Z.
Proof.
intros sk Z.
simpl.
rewrite -> cast_id; reflexivity.
Qed.


Lemma subst_simp_cand :
  forall s k Q,
    hkd k
    -> Cand (simp (subst s k)) (cast (simp k) (simp (subst s k)) Q) = Cand (simp k) Q.
Proof.
intros s k Q H.
rewrite <- (cand_ext_2 (simp k) (simp (subst s k))); auto; [].
rewrite -> subst_simp; auto.
Qed.
       

Lemma maybe_cast_none :
  forall sk sk' A,
    sk <> sk' -> maybe_cast sk sk' A = None.
Proof.
intros sk sk' A Hneq.
revert sk' A Hneq.
induct sk.

(* type *)
intros sk' A Hneq.
destruct sk'; try (intros; reflexivity); [].
destruct Hneq; auto; done.

(* pi *)
intros sk1 IH1 sk2 IH2 sk' A.
case sk'; try (intros; reflexivity); [].
intros sk1' sk2' Hneq.
so (eq_skind_dec sk1 sk1') as [Heqsk | Hneqsk].
  subst sk1'.
  simpl.
  rewrite -> maybe_cast_id; [].
  rewrite -> IH2; auto; [].
  contradict Hneq.
  f_equal; auto; done.

  simpl.
  rewrite -> IH1; auto; done.

(* sigma *)
intros sk1 IH1 sk2 IH2 sk' A.
case sk'; try (intros; reflexivity); [].
intros sk1' sk2' Hneq.
so (eq_skind_dec sk1 sk1') as [Heqsk | Hneqsk].
  subst sk1'.
  simpl.
  rewrite -> maybe_cast_id; [].
  rewrite -> IH2; auto; [].
  contradict Hneq.
  f_equal; auto; done.

  simpl.
  rewrite -> IH1; auto; done.

(* one *)
intros sk' A Hneq.
destruct sk'; try (intros; reflexivity); [].
destruct Hneq; auto; done.
Qed.


Lemma left_app :
  forall E1 E2,
    left (E1;; E2) = Subst.compose (under (length E2) (left E1)) (left E2).
Proof.
intros E1 E2.
induct E2.

(* nil *)
simpl.
simp_sub.
reflexivity.

(* cons *)
intros S E IH.
#2 destruct S as [| c c' Q].
  simpl.
  simp_sub.
  f_equal; [].
  assumption.

  simpl.
  simp_sub.
  f_equal; [].
  assumption.
Qed.


Lemma right_app :
  forall E1 E2,
    right (E1;; E2) = Subst.compose (under (length E2) (right E1)) (right E2).
Proof.
intros E1 E2.
induct E2.

(* nil *)
simpl.
simp_sub.
reflexivity.

(* cons *)
intros S E IH.
#2 destruct S as [| c c' Q].
  simpl.
  simp_sub.
  f_equal; [].
  assumption.

  simpl.
  simp_sub.
  f_equal; [].
  assumption.
Qed.


Lemma insert_left :
  forall E1 c E2 Q,
    Subst.compose (under (length E2) (dot c id)) (left (E1;; E2))
    = left (E1; Sta (subst (left E1) c) (subst (right E1) c) Q;; E2).
Proof.
intros E1 c E2 Q.
rewrite -> !left_app; [].
rewrite <- compose_assoc; [].
rewrite -> dist_compose_under; [].
reflexivity.
Qed.


Lemma insert_right :
  forall E1 c E2 Q,
    Subst.compose (under (length E2) (dot c id)) (right (E1;; E2))
    = right (E1; Sta (subst (left E1) c) (subst (right E1) c) Q;; E2).
Proof.
intros E1 c E2 Q.
rewrite -> !right_app; [].
rewrite <- compose_assoc; [].
rewrite -> dist_compose_under; [].
reflexivity.
Qed.
