
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.


Inductive skeleton : Set :=
| sk_type  : skeleton
| sk_sing  : skeleton
| sk_pi    : skeleton -> skeleton -> skeleton
| sk_sigma : skeleton -> skeleton -> skeleton
| sk_one   : skeleton
.


Fixpoint skel (k : (* kd *) term) {struct k} : skeleton :=
  (match k with
   | kd_type => sk_type
   | kd_sing c => sk_sing
   | kd_pi k1 k2 => sk_pi (skel k1) (skel k2)
   | kd_sigma k1 k2 => sk_sigma (skel k1) (skel k2)
   | _ => sk_one
   end).


Lemma eqk_skel :
  forall G k k',
    eqk G k k' -> skel k = skel k'.
Proof.
intros G k k' Heqk.
induct Heqk.

(* refl *)
intros; reflexivity.

(* symm *)
intros; symmetry; assumption.

(* trans *)
intros; etransitivity; eassumption.

(* sing *)
intros; simpl; reflexivity.

(* pi *)
intros; simpl; f_equal; auto; done.

(* sigma *)
intros; simpl; f_equal; auto; done.
Qed.


(* Doesn't need any premises. *)
Lemma shift_skel :
  forall i k,
    skel (subst (sh i) k) = skel k.
Proof.
intros i k.
rewrite <- (under_zero (sh i)).
generalize 0.
revert i.
induct k; try (intros; simp_sub; simpl; auto; done).

(* var *)
intros n i j.
simp_sub.
so (Compare_dec.dec_lt n j) as [Hlt | Hgeq].
  rewrite -> project_under_lt; auto; done.

  rewrite -> project_under_ge; [| omega].
  simp_sub.
  simpl; reflexivity.

(* pi *)
intros k1 IH1 k2 IH2 i n.
simp_sub.
simpl.
f_equal; auto; [].
rewrite <- under_succ.
apply IH2; done.

(* sigma *)
intros k1 IH1 k2 IH2 i n.
simp_sub.
simpl.
f_equal; auto; [].
rewrite <- under_succ.
apply IH2; done.
Qed.


Lemma subst_skel :
  forall s k, hkd k -> skel (subst s k) = skel k.
Proof.
intros s k Hkd.
revert s.
induct Hkd; intros; simp_sub; simpl; try f_equal; auto.
Qed.


Fixpoint inh (k : (* kd *) term) {struct k} :=
  (match k with
   | kd_pi k1 k2 =>
       cn_lam k1 (inh k2)

   | kd_sigma k1 k2 =>
       cn_pair (inh k1) (subst1 (inh k1) (inh k2))

   | kd_sing c => c

   | kd_type => cn_unit

   | _ => cn_star
   end).


(* Doesn't need any premises. *)
Lemma shift_under_inh :
  forall i j k, subst (under i (sh j)) (inh k) = inh (subst (under i (sh j)) k).
Proof.
intros i j k.
revert i.
#3 induct k; try (intros; simp_sub; simpl; auto; done).

(* var *)
intros n i.
simp_sub.
so (Compare_dec.dec_lt n i) as [Hlt | Hgeq].
  rewrite -> project_under_lt; auto; done.

  rewrite -> project_under_ge; [| omega].
  simp_sub.
  simpl; reflexivity.

(* pi *)
intros k1 _ k2 IH i.
simp_sub.
simpl.
simp_sub.
f_equal; [].
rewrite <- under_succ; [].
auto; done.

(* sigma *)
intros k1 IH1 k2 IH2 i.
simp_sub.
simpl.
simp_sub.
f_equal; auto; [].
rewrite <- under_succ; [].
rewrite <- IH2; [].
simp_sub.
rewrite <- IH1; [].
reflexivity.
Qed.


Lemma inh_ofc :
  forall G k,
    ofk G k
    -> ofc G (inh k) k.
Proof.
intros G k Hof.
induct Hof.

(* type *)
intros G.
simpl.
apply ofc_unit; done.

(* sing *)
intros G c Hof.
simpl.
apply ofc_sing; auto; done.

(* pi *)
intros G k1 k2 Hof1 _ Hof2 IH.
simpl.
apply ofc_lam; auto; done.

(* sigma *)
intros G k1 k2 Hof1 IH1 Hof2 IH2.
simpl.
apply ofc_pair; eauto using subst1_ofc; done.

(* one *)
intros G.
simpl.
apply ofc_star; done.
Qed.


Lemma inhabitation :
  forall G k,
    ofk G k
    -> exists c, ofc G c k.
Proof.
intros G k H.
exists (inh k).
apply inh_ofc; auto; done.
Qed.


Lemma eqk_type_invert :
  forall G k,
    eqk G kd_type k
    -> k = kd_type

with eqk_type_invert' :
  forall G k,
    eqk G k kd_type
    -> k = kd_type.
Proof.
intros G k Heqk.
remember kd_type as e eqn:Heqe in Heqk.
revert Heqe.
#3 induct Heqk; try (intros; discriminate).
  (* refl *)
  intros G k _ ->.
  reflexivity.

  (* symm *)
  intros G k k' Heqk _ ->.
  eapply eqk_type_invert'; eauto; done.

  (* trans *)
  intros G k1 k2 k3 _ IH1 _ IH2 ->.
  so (IH1 (eq_refl _)) as ->.
  so (IH2 (eq_refl _)) as ->.
  reflexivity.

intros G k Heqk.
remember kd_type as e eqn:Heqe in Heqk.
revert Heqe.
#3 induct Heqk; try (intros; discriminate).
  (* refl *)
  intros G k _ ->.
  reflexivity.

  (* symm *)
  intros G k k' Heqk _ ->.
  eapply eqk_type_invert; eauto; done.

  (* trans *)
  intros G k1 k2 k3 _ IH1 _ IH2 ->.
  so (IH2 (eq_refl _)) as ->.
  so (IH1 (eq_refl _)) as ->.
  reflexivity.
Qed.


Lemma eqk_sing_invert :
  forall G c k,
    eqk G (kd_sing c) k
    -> exists c', k = kd_sing c' /\ eqc G c c' kd_type

with eqk_sing_invert' :
  forall G c k,
    eqk G k (kd_sing c)
    -> exists c', k = kd_sing c' /\ eqc G c c' kd_type.
Proof.
intros G c k Heqk.
remember (kd_sing c) as e eqn:Heqe.
revert c Heqe.
#4 induct Heqk; try (intros; discriminate).
  (* refl *)
  intros G k Hofk c ->.
  invert Hofk.
  intro Hofc.
  exists c; split; auto; [].
  apply eqc_refl; auto; done.
  
  (* symm *)
  intros G k k' Heqk _ c ->.
  so (eqk_sing_invert' _#3 Heqk) as (c' & Heq & Heqc).
  exists c'; split; auto; done.
  
  (* trans *)
  intros G k1 k2 k3 H1 IH1 H2 IH2 c1 ->.
  so (IH1 c1 (eq_refl _)) as (c2 & -> & Heqc12).
  so (IH2 c2 (eq_refl _)) as (c3 & -> & Heqc23).
  exists c3; split; eauto with static; done.
  
  (* sing *)
  intros G b b' H c Heq.
  injection Heq; intros; subst c.
  exists b'; split; auto; done.
  
intros G c k Heqk.
remember (kd_sing c) as e eqn:Heqe.
revert c Heqe.
#4 induct Heqk; try (intros; discriminate).
  (* refl *)
  intros G k Hofk c ->.
  invert Hofk.
  intro Hofc.
  exists c; split; auto; [].
  apply eqc_refl; auto; done.
  
  (* symm *)
  intros G k k' Heqk _ c ->.
  so (eqk_sing_invert _#3 Heqk) as (c' & Heq & Heqc).
  exists c'; split; auto; done.
  
  (* trans *)
  intros G k1 k2 k3 H1 IH1 H2 IH2 c3 ->.
  so (IH2 c3 (eq_refl _)) as (c2 & -> & Heqc23).
  so (IH1 c2 (eq_refl _)) as (c1 & -> & Heqc12).
  exists c1; split; eauto with static; done.

  (* sing *)
  intros G b b' H c Heq.
  injection Heq; intros; subst c.
  exists b; split; auto; [].
  apply eqc_symm; auto; done.
Qed.


Lemma sing_injective :
  forall G c c',
    eqk G (kd_sing c) (kd_sing c')
    -> eqc G c c' kd_type.
Proof.
intros G c c' Heqk.
so (eqk_sing_invert _#3 Heqk) as (c'' & Heq & Heqc).
injection Heq; intros; subst c''.
assumption.
Qed.


Lemma eqk_pi_invert :
  forall G k1 k2 k',
    eqk G (kd_pi k1 k2) k'
    -> exists k1' k2', k' = kd_pi k1' k2' /\ eqk G k1 k1' /\ eqk (G; cl_cn k1) k2 k2'

with eqk_pi_invert' :
  forall G k1 k2 k',
    eqk G k' (kd_pi k1 k2)
    -> exists k1' k2', k' = kd_pi k1' k2' /\ eqk G k1 k1' /\ eqk (G; cl_cn k1) k2 k2'.
Proof.
intros G k1 k2 k' Heqk.
remember (kd_pi k1 k2) as e eqn:Heqe.
revert k1 k2 Heqe.
#4 induct Heqk; try (intros; discriminate).
  (* refl *)
  intros G k Hofk k1 k2 ->.
  invert Hofk.
  intros Hofk1 Hofk2.
  exists k1, k2.
  do2 2 split; auto with static; done.

  (* symm *)
  intros G k k' Heqk _ k1 k2 ->.
  so (eqk_pi_invert' _#4 Heqk) as (k1' & k2' & Heq & Heqk1 & Heqk2).
  eauto; done.
  
  (* trans *)
  intros G k k' k'' _ IH1 _ IH2 k1 k2 ->.
  so (IH1 _ _ (eq_refl _)) as (k1' & k2' & -> & Heqk1 & Heqk2).
  so (IH2 _ _ (eq_refl _)) as (k1'' & k2'' & -> & Heqk1' & Heqk2').
  exists k1'', k2''.
  do2 2 split; eauto with static; [].
  so (subof_conv_0 _#3 Heqk1) as Hsubof.
  so (subst_eqk _#5 Hsubof Heqk2') as Heqk2''.
  rewrite -> !subst_under_id in Heqk2''.
  eauto with static; done.
  
  (* pi *)
  intros G e1 k1' e2 k2' Heqk1 _ Heqk2 _ k1 k2 Heqe.
  injection Heqe; intros; subst e1 e2.
  eauto; done.

intros G k1 k2 k' Heqk.
remember (kd_pi k1 k2) as e eqn:Heqe.
revert k1 k2 Heqe.
#4 induct Heqk; try (intros; discriminate).
  (* refl *)
  intros G k Hofk k1 k2 ->.
  invert Hofk.
  intros Hofk1 Hofk2.
  exists k1, k2.
  do2 2 split; auto with static; done.
  
  (* symm *)
  intros G k k' Heqk _ k1 k2 ->.
  so (eqk_pi_invert _#4 Heqk) as (k1' & k2' & Heq & Heqk1 & Heqk2).
  eauto; done.
  
  (* trans *)
  intros G k k' k'' _ IH1 _ IH2 k1 k2 ->.
  so (IH2 _ _ (eq_refl _)) as (k1' & k2' & -> & Heqk1 & Heqk2).
  so (IH1 _ _ (eq_refl _)) as (k1'' & k2'' & -> & Heqk1' & Heqk2').
  exists k1'', k2''.
  do2 2 split; eauto with static; [].
  so (subof_conv_0 _#3 Heqk1) as Hsubof.
  so (subst_eqk _#5 Hsubof Heqk2') as Heqk2''.
  rewrite -> !subst_under_id in Heqk2''.
  eauto with static; done.
  
  (* pi *)
  intros G k1' e1 k2' e2 Heqk1 _ Heqk2 _ k1 k2 Heqe.
  injection Heqe; intros; subst e1 e2.
  exists k1', k2'.
  do2 2 split; auto with static; [].
  apply eqk_symm; [].
  so (subof_conv_0 _#3 (eqk_symm Heqk1)) as Hsubof.
  so (subst_eqk _#5 Hsubof Heqk2) as Heqk2'.
  rewrite -> !subst_under_id in Heqk2'.
  assumption.
Qed.


Lemma pi_injective :
  forall G k1 k1' k2 k2',
    eqk G (kd_pi k1 k2) (kd_pi k1' k2')
    -> eqk G k1 k1' /\ eqk (G; cl_cn k1) k2 k2'.
Proof.
intros G k1 k1' k2 k2' Heqk.
so (eqk_pi_invert _#4 Heqk) as (e1 & e2 & Heq & Heqk1 & Heqk2).
injection Heq; intros; subst e1 e2.
auto; done.
Qed.


Lemma eqk_sigma_invert :
  forall G k1 k2 k',
    eqk G (kd_sigma k1 k2) k'
    -> exists k1' k2', k' = kd_sigma k1' k2' /\ eqk G k1 k1' /\ eqk (G; cl_cn k1) k2 k2'

with eqk_sigma_invert' :
  forall G k1 k2 k',
    eqk G k' (kd_sigma k1 k2)
    -> exists k1' k2', k' = kd_sigma k1' k2' /\ eqk G k1 k1' /\ eqk (G; cl_cn k1) k2 k2'.
Proof.
intros G k1 k2 k' Heqk.
remember (kd_sigma k1 k2) as e eqn:Heqe.
revert k1 k2 Heqe.
#4 induct Heqk; try (intros; discriminate).
  (* refl *)
  intros G k Hofk k1 k2 ->.
  invert Hofk.
  intros Hofk1 Hofk2.
  exists k1, k2.
  do2 2 split; auto with static; done.

  (* symm *)
  intros G k k' Heqk _ k1 k2 ->.
  so (eqk_sigma_invert' _#4 Heqk) as (k1' & k2' & Heq & Heqk1 & Heqk2).
  eauto; done.
  
  (* trans *)
  intros G k k' k'' _ IH1 _ IH2 k1 k2 ->.
  so (IH1 _ _ (eq_refl _)) as (k1' & k2' & -> & Heqk1 & Heqk2).
  so (IH2 _ _ (eq_refl _)) as (k1'' & k2'' & -> & Heqk1' & Heqk2').
  exists k1'', k2''.
  do2 2 split; eauto with static; [].
  so (subof_conv_0 _#3 Heqk1) as Hsubof.
  so (subst_eqk _#5 Hsubof Heqk2') as Heqk2''.
  rewrite -> !subst_under_id in Heqk2''.
  eauto with static; done.
  
  (* sigma *)
  intros G e1 k1' e2 k2' Heqk1 _ Heqk2 _ k1 k2 Heqe.
  injection Heqe; intros; subst e1 e2.
  eauto; done.

intros G k1 k2 k' Heqk.
remember (kd_sigma k1 k2) as e eqn:Heqe.
revert k1 k2 Heqe.
#4 induct Heqk; try (intros; discriminate).
  (* refl *)
  intros G k Hofk k1 k2 ->.
  invert Hofk.
  intros Hofk1 Hofk2.
  exists k1, k2.
  do2 2 split; auto with static; done.
  
  (* symm *)
  intros G k k' Heqk _ k1 k2 ->.
  so (eqk_sigma_invert _#4 Heqk) as (k1' & k2' & Heq & Heqk1 & Heqk2).
  eauto; done.
  
  (* trans *)
  intros G k k' k'' _ IH1 _ IH2 k1 k2 ->.
  so (IH2 _ _ (eq_refl _)) as (k1' & k2' & -> & Heqk1 & Heqk2).
  so (IH1 _ _ (eq_refl _)) as (k1'' & k2'' & -> & Heqk1' & Heqk2').
  exists k1'', k2''.
  do2 2 split; eauto with static; [].
  so (subof_conv_0 _#3 Heqk1) as Hsubof.
  so (subst_eqk _#5 Hsubof Heqk2') as Heqk2''.
  rewrite -> !subst_under_id in Heqk2''.
  eauto with static; done.
  
  (* sigma *)
  intros G k1' e1 k2' e2 Heqk1 _ Heqk2 _ k1 k2 Heqe.
  injection Heqe; intros; subst e1 e2.
  exists k1', k2'.
  do2 2 split; auto with static; [].
  apply eqk_symm; [].
  so (subof_conv_0 _#3 (eqk_symm Heqk1)) as Hsubof.
  so (subst_eqk _#5 Hsubof Heqk2) as Heqk2'.
  rewrite -> !subst_under_id in Heqk2'.
  assumption.
Qed.


Lemma sigma_injective :
  forall G k1 k1' k2 k2',
    eqk G (kd_sigma k1 k2) (kd_sigma k1' k2')
    -> eqk G k1 k1' /\ eqk (G; cl_cn k1) k2 k2'.
Proof.
intros G k1 k1' k2 k2' Heqk.
so (eqk_sigma_invert _#4 Heqk) as (e1 & e2 & Heq & Heqk1 & Heqk2).
injection Heq; intros; subst e1 e2.
auto; done.
Qed.


Lemma subk_sing_invert' :
  forall G c k,
    subk G k (kd_sing c)
    -> exists c', k = kd_sing c' /\ eqc G c c' kd_type.
Proof.
intros G c k Hsubk.
remember (kd_sing c) as e eqn:Heqe.
revert c Heqe.
#2 induct Hsubk; try (intros; discriminate).

(* refl *)
intros G k k' Heqk c ->.
apply eqk_sing_invert'; auto; done.

(* trans *)
intros G k1 k2 k3 _ IH1 _ IH2 c3 ->.
so (IH2 c3 (eq_refl _)) as (c2 & -> & Heqc23).
so (IH1 c2 (eq_refl _)) as (c1 & -> & Heqc12).
exists c1.
split; eauto with static; done.
Qed.


Lemma subk_type_invert' :
  forall G k,
    wf G
    -> subk G k kd_type
    -> k = kd_type \/ exists c, k = kd_sing c /\ ofc G c kd_type.
Proof.
intros G k Hwf Hsubk.
remember kd_type as e eqn:Heqe.
revert Heqe Hwf.
#3 induct Hsubk; try (intros; discriminate).

(* refl *)
intros G k k' Heqk -> _.
left.
eapply eqk_type_invert; eauto using eqk_symm; done.

(* trans *)
intros G k1 k2 k3 Hsubk IH1 _ IH2 -> Hwf.
so (IH2 (eq_refl _) Hwf) as [-> | (c & -> & Hof)].
  apply IH1; auto; done.

  right.
  so (subk_sing_invert' _#3 Hsubk) as (c' & -> & Heqc).
  exists c'.
  split; auto; [].
  exact (eqc_reg _#4 Hwf Heqc ander andel).

(* sing_t *)
intros G c Hof _ Hwf.  
right.
exists c.
auto; done.
Qed.


Lemma subk_pi_invert' :
  forall G k k1' k2',
    subk G k (kd_pi k1' k2')
    -> exists k1 k2, k = kd_pi k1 k2 /\ subk G k1' k1 /\ subk (G; cl_cn k1') k2 k2'.
Proof.
intros G k k1' k2' Hsubk.
remember (kd_pi k1' k2') as e eqn:Heqe.
revert k1' k2' Heqe.
#3 induct Hsubk; try (intros; discriminate).

(* refl *)
intros G k k' Heqk k1 k2  ->.
so (eqk_pi_invert' _#4 Heqk) as (k1' & k2' & Heq & Heqk1 & Heqk2).
exists k1', k2'.
do2 2 split; eauto using subk_refl, eqk_symm; done.

(* trans *)
intros G k k' k'' _ IH1 _ IH2 k1 k2 ->.
so (IH2 _ _ (eq_refl _)) as (k1' & k2' & -> & Hsubk1 & Hsubk2).
so (IH1 _ _ (eq_refl _)) as (k1'' & k2'' & -> & Hsubk1' & Hsubk2').
exists k1'', k2''.
do2 2 split; eauto with static; [].
so (subst_subk _#5 (subof_subsume_0 _#3 Hsubk1) Hsubk2') as Hsubk2''.
rewrite -> !subst_under_id in Hsubk2''; [].
eapply subk_trans; eauto; done.

(* pi *)
intros G k1' e1 k2' e2 Heqk1 _ Heqk2 _ _ k1 k2 Heqe.
injection Heqe; intros; subst e1 e2.
exists k1', k2'.
do2 2 split; auto with static; done.
Qed.


Lemma subk_sigma_invert' :
  forall G k k1' k2',
    subk G k (kd_sigma k1' k2')
    -> exists k1 k2, k = kd_sigma k1 k2 /\ subk G k1 k1' /\ subk (G; cl_cn k1) k2 k2'.
Proof.
intros G k k1' k2' Hsubk.
remember (kd_sigma k1' k2') as e eqn:Heqe.
revert k1' k2' Heqe.
#3 induct Hsubk; try (intros; discriminate).

(* refl *)
intros G k k' Heqk k1 k2  ->.
so (eqk_sigma_invert' _#4 Heqk) as (k1' & k2' & Heq & Heqk1 & Heqk2).
exists k1', k2'.
do2 2 split; eauto using subk_refl, eqk_symm; [].
apply subk_refl; [].
apply eqk_symm; [].
so (subst_eqk _#5 (subof_conv_0 _#3 (eqk_symm Heqk1)) Heqk2) as Heqk2'.
rewrite -> !subst_under_id in Heqk2'; [].
assumption.

(* trans *)
intros G k k' k'' _ IH1 _ IH2 k1 k2 ->.
so (IH2 _ _ (eq_refl _)) as (k1' & k2' & -> & Hsubk1 & Hsubk2).
so (IH1 _ _ (eq_refl _)) as (k1'' & k2'' & -> & Hsubk1' & Hsubk2').
exists k1'', k2''.
do2 2 split; eauto with static; [].
so (subst_subk _#5 (subof_subsume_0 _#3 Hsubk1') Hsubk2) as Hsubk2''.
rewrite -> !subst_under_id in Hsubk2''; [].
eapply subk_trans; eauto; done.

(* sigma *)
intros G k1' e1 k2' e2 Heqk1 _ Heqk2 _ _ k1 k2 Heqe.
injection Heqe; intros; subst e1 e2.
exists k1', k2'.
do2 2 split; auto with static; done.
Qed.


Lemma pi_injective_subk :
  forall G k1 k1' k2 k2',
    subk G (kd_pi k1 k2) (kd_pi k1' k2')
    -> subk G k1' k1 /\ subk (G; cl_cn k1') k2 k2'.
Proof.
intros G k1 k1' k2 k2' Hsubk.
so (subk_pi_invert' _#4 Hsubk) as (k1'' & k2'' & Heq & Hsubk1 & Hsubk2).
injection Heq; intros; subst; auto.
Qed.


Lemma sigma_injective_subk :
  forall G k1 k1' k2 k2',
    subk G (kd_sigma k1 k2) (kd_sigma k1' k2')
    -> subk G k1 k1' /\ subk (G; cl_cn k1) k2 k2'.
Proof.
intros G k1 k1' k2 k2' Hsubk.
so (subk_sigma_invert' _#4 Hsubk) as (k1'' & k2'' & Heq & Hsubk1 & Hsubk2).
injection Heq; intros; subst; auto.
Qed.


Lemma can_fsts_hsg :
  forall T,
    hsg T
    -> exists k, fsts T k.
Proof.
intros T Hhsg.
induct Hhsg;
intros;
repeat (match goal with
        | H : exists _:_, _ |- _ => destruct H
        end);
eauto using fsts.
Qed.


Lemma eqs_one_inverts :
  forall G T1 T2,
    eqs G T1 T2
    -> (T1 = sg_one <-> T2 = sg_one).
Proof.
intros G T1 T2 H.
#3 induct H; try (intros; split; discriminate).

(* refl *)
intros G T Hofs.
reflexivity.

(* symm *)
intros G T T' _ (IH & IH').
split; auto; done.

(* trans *)
intros G T1 T2 T3 _ (IH1 & IH1') _ (IH2 & IH2').
split; auto; done.
Qed.


Lemma eqs_stat_inverts :
  forall G T1 T2,
    eqs G T1 T2
    -> (forall k, T1 = sg_stat k -> exists k', T2 = sg_stat k' /\ eqk G k k')
       /\
       (forall k, T2 = sg_stat k -> exists k', T1 = sg_stat k' /\ eqk G k k').
Proof.
intros G T1 T2 H.
#4 induct H; try (intros; split; discriminate).

(* refl *)
intros G T Hofs.
split.
  intros k ->.
  exists k; split; auto; [].
  invert Hofs; [].
  auto with static; done.

  intros k ->.
  exists k; split; auto; [].
  invert Hofs; [].
  auto with static; done.

(* symm *)
intros G T T' _ (IH & IH').
split; auto; done.

(* trans *)
intros G T1 T2 T3 _ (IH1 & IH1') _ (IH2 & IH2').
split.
  intros k Heq.
  so (IH1 _ Heq) as (k' & Heq' & Heqc).
  so (IH2 _ Heq') as (k'' & Heq'' & Heqc').
  eauto using eqk_trans; done.

  intros k Heq.
  so (IH2' _ Heq) as (k' & Heq' & Heqc).
  so (IH1' _ Heq') as (k'' & Heq'' & Heqc').
  eauto using eqk_trans; done.

(* stat *)
intros G k1 k2 Heqc.
split.
  intros k Heq.
  injection Heq.
  intros; subst; eauto; done.

  intros k Heq.
  injection Heq.
  intros; subst; eauto with static; done.
Qed.


Lemma eqs_dyn_inverts :
  forall G T1 T2,
    eqs G T1 T2
    -> (forall t, T1 = sg_dyn t -> exists t', T2 = sg_dyn t' /\ eqc G t t' kd_type)
       /\
       (forall t, T2 = sg_dyn t -> exists t', T1 = sg_dyn t' /\ eqc G t t' kd_type).
Proof.
intros G T1 T2 H.
#4 induct H; try (intros; split; discriminate).

(* refl *)
intros G T Hofs.
split.
  intros t ->.
  exists t; split; auto; [].
  invert Hofs; [].
  auto with static; done.

  intros t ->.
  exists t; split; auto; [].
  invert Hofs; [].
  auto with static; done.

(* symm *)
intros G T T' _ (IH & IH').
split; auto; done.

(* trans *)
intros G T1 T2 T3 _ (IH1 & IH1') _ (IH2 & IH2').
split.
  intros t Heq.
  so (IH1 _ Heq) as (t' & Heq' & Heqc).
  so (IH2 _ Heq') as (t'' & Heq'' & Heqc').
  eauto using eqc_trans; done.

  intros t Heq.
  so (IH2' _ Heq) as (t' & Heq' & Heqc).
  so (IH1' _ Heq') as (t'' & Heq'' & Heqc').
  eauto using eqc_trans; done.

(* dyn *)
intros G t1 t2 Heqc.
split.
  intros t Heq.
  injection Heq.
  intros; subst; eauto; done.

  intros t Heq.
  injection Heq.
  intros; subst; eauto with static; done.
Qed.


Lemma dyn_injective :
  forall G t t',
    eqs G (sg_dyn t) (sg_dyn t')
    -> eqc G t t' kd_type.
Proof.
intros G t t' Heqs.
so (eqs_dyn_inverts _#3 Heqs andel _ (eq_refl _)) as (t'' & Heq & Heqc).
injection Heq; intros; subst t''.
assumption.
Qed.


Lemma eqs_pigen_inverts :
  forall G T T',
    wf G
    -> eqs G T T'
    -> (forall T1 T2, T = sg_pigen T1 T2 -> exists k1 T1' T2', T' = sg_pigen T1' T2' /\ fsts T1 k1 /\ eqs G T1 T1' /\ eqs (G; cl_cn k1) T2 T2')
       /\
       (forall T1 T2, T' = sg_pigen T1 T2 -> exists k1 T1' T2', T = sg_pigen T1' T2' /\ fsts T1 k1 /\ eqs G T1 T1' /\ eqs (G; cl_cn k1) T2 T2').
Proof.
intros G T T' Hwf H.
revert Hwf.
#4 induct H; try (intros; split; discriminate).

(* refl *)
intros G T Hofs _.
split.
  intros T1 T2 ->.
  invert Hofs; [].
  intros k1 Hof1 Hfst Hof2.
  exists k1, T1, T2.
  do2 3 split; auto with static; done.

  intros T1 T2 ->.
  invert Hofs; [].
  intros k1 Hof1 Hfst Hof2.
  exists k1, T1, T2.
  do2 3 split; auto with static; done.

(* symm *)
intros G T T' _ IH Hwf.
so (IH Hwf) as ().
split; assumption.

(* trans *)
intros G T T' T'' _ IH1 _ IH2 Hwf.
split.
  intros T1 T2 Heq.
  so (IH1 Hwf andel _ _ Heq) as (k & T1' & T2' & Heq' & Hfst & Heq1 & Heq2).
  so (IH2 Hwf andel _ _ Heq') as (k' & T1'' & T2'' & Heq'' & Hfst' & Heq1' & Heq2').
  so (subst_eqs _#5 (subof_conv_0 _#3 (eqs_fst _#5 Hwf Heq1 Hfst Hfst')) Heq2') as Heq2''.
  simp_sub in Heq2''.
  do 3 eexists; do2 3 split; eauto using eqs_trans; done.

  intros T1 T2 Heq.
  so (IH2 Hwf ander _ _ Heq) as (k & T1' & T2' & Heq' & Hfst & Heq1 & Heq2).
  so (IH1 Hwf ander _ _ Heq') as (k' & T1'' & T2'' & Heq'' & Hfst' & Heq1' & Heq2').
  so (subst_eqs _#5 (subof_conv_0 _#3 (eqs_fst _#5 Hwf Heq1 Hfst Hfst')) Heq2') as Heq2''.
  simp_sub in Heq2''.
  do 3 eexists; do2 3 split; eauto using eqs_trans; done.

(* quant *)
intros G T1 T1' k1 T2 T2' Heq1 _ Hfst1 Heq2 _ Hwf.
split.
  intros e1 e2 Heq.
  injection Heq; intros; subst e1 e2.
  do 3 eexists; do2 3 split; eauto; done.

  intros e1 e2 Heq.
  injection Heq; intros; subst e1 e2.
  so (can_fsts _ _ (eqs_reg _#3 Hwf Heq1 ander)) as (k1' & Hfst1').
  exists k1', T1, T2.
  do2 3 split; auto with static; [].
  so (subst_eqs _#5 (subof_conv_0 _#3 (eqk_symm (eqs_fst _#5 Hwf Heq1 Hfst1 Hfst1'))) Heq2) as Heq2'.
  simp_sub in Heq2'.
  apply eqs_symm; assumption.
Qed.


Lemma eqs_piapp_inverts :
  forall G T T',
    wf G
    -> eqs G T T'
    -> (forall T1 T2, T = sg_piapp T1 T2 -> exists k1 T1' T2', T' = sg_piapp T1' T2' /\ fsts T1 k1 /\ eqs G T1 T1' /\ eqs (G; cl_cn k1) T2 T2')
       /\
       (forall T1 T2, T' = sg_piapp T1 T2 -> exists k1 T1' T2', T = sg_piapp T1' T2' /\ fsts T1 k1 /\ eqs G T1 T1' /\ eqs (G; cl_cn k1) T2 T2').
Proof.
intros G T T' Hwf H.
revert Hwf.
#4 induct H; try (intros; split; discriminate).

(* refl *)
intros G T Hofs _.
split.
  intros T1 T2 ->.
  invert Hofs; [].
  intros k1 Hof1 Hfst Hof2.
  exists k1, T1, T2.
  do2 3 split; auto with static; done.

  intros T1 T2 ->.
  invert Hofs; [].
  intros k1 Hof1 Hfst Hof2.
  exists k1, T1, T2.
  do2 3 split; auto with static; done.

(* symm *)
intros G T T' _ IH Hwf.
so (IH Hwf) as ().
split; assumption.

(* trans *)
intros G T T' T'' _ IH1 _ IH2 Hwf.
split.
  intros T1 T2 Heq.
  so (IH1 Hwf andel _ _ Heq) as (k & T1' & T2' & Heq' & Hfst & Heq1 & Heq2).
  so (IH2 Hwf andel _ _ Heq') as (k' & T1'' & T2'' & Heq'' & Hfst' & Heq1' & Heq2').
  so (subst_eqs _#5 (subof_conv_0 _#3 (eqs_fst _#5 Hwf Heq1 Hfst Hfst')) Heq2') as Heq2''.
  simp_sub in Heq2''.
  do 3 eexists; do2 3 split; eauto using eqs_trans; done.

  intros T1 T2 Heq.
  so (IH2 Hwf ander _ _ Heq) as (k & T1' & T2' & Heq' & Hfst & Heq1 & Heq2).
  so (IH1 Hwf ander _ _ Heq') as (k' & T1'' & T2'' & Heq'' & Hfst' & Heq1' & Heq2').
  so (subst_eqs _#5 (subof_conv_0 _#3 (eqs_fst _#5 Hwf Heq1 Hfst Hfst')) Heq2') as Heq2''.
  simp_sub in Heq2''.
  do 3 eexists; do2 3 split; eauto using eqs_trans; done.

(* quant *)
intros G T1 T1' k1 T2 T2' Heq1 _ Hfst1 Heq2 _ Hwf.
split.
  intros e1 e2 Heq.
  injection Heq; intros; subst e1 e2.
  do 3 eexists; do2 3 split; eauto; done.

  intros e1 e2 Heq.
  injection Heq; intros; subst e1 e2.
  so (can_fsts _ _ (eqs_reg _#3 Hwf Heq1 ander)) as (k1' & Hfst1').
  exists k1', T1, T2.
  do2 3 split; auto with static; [].
  so (subst_eqs _#5 (subof_conv_0 _#3 (eqk_symm (eqs_fst _#5 Hwf Heq1 Hfst1 Hfst1'))) Heq2) as Heq2'.
  simp_sub in Heq2'.
  apply eqs_symm; assumption.
Qed.


Lemma eqs_sigma_inverts :
  forall G T T',
    wf G
    -> eqs G T T'
    -> (forall T1 T2, T = sg_sigma T1 T2 -> exists k1 T1' T2', T' = sg_sigma T1' T2' /\ fsts T1 k1 /\ eqs G T1 T1' /\ eqs (G; cl_cn k1) T2 T2')
       /\
       (forall T1 T2, T' = sg_sigma T1 T2 -> exists k1 T1' T2', T = sg_sigma T1' T2' /\ fsts T1 k1 /\ eqs G T1 T1' /\ eqs (G; cl_cn k1) T2 T2').
Proof.
intros G T T' Hwf H.
revert Hwf.
#4 induct H; try (intros; split; discriminate).

(* refl *)
intros G T Hofs _.
split.
  intros T1 T2 ->.
  invert Hofs; [].
  intros k1 Hof1 Hfst Hof2.
  exists k1, T1, T2.
  do2 3 split; auto with static; done.

  intros T1 T2 ->.
  invert Hofs; [].
  intros k1 Hof1 Hfst Hof2.
  exists k1, T1, T2.
  do2 3 split; auto with static; done.

(* symm *)
intros G T T' _ IH Hwf.
so (IH Hwf) as ().
split; assumption.

(* trans *)
intros G T T' T'' _ IH1 _ IH2 Hwf.
split.
  intros T1 T2 Heq.
  so (IH1 Hwf andel _ _ Heq) as (k & T1' & T2' & Heq' & Hfst & Heq1 & Heq2).
  so (IH2 Hwf andel _ _ Heq') as (k' & T1'' & T2'' & Heq'' & Hfst' & Heq1' & Heq2').
  so (subst_eqs _#5 (subof_conv_0 _#3 (eqs_fst _#5 Hwf Heq1 Hfst Hfst')) Heq2') as Heq2''.
  simp_sub in Heq2''.
  do 3 eexists; do2 3 split; eauto using eqs_trans; done.

  intros T1 T2 Heq.
  so (IH2 Hwf ander _ _ Heq) as (k & T1' & T2' & Heq' & Hfst & Heq1 & Heq2).
  so (IH1 Hwf ander _ _ Heq') as (k' & T1'' & T2'' & Heq'' & Hfst' & Heq1' & Heq2').
  so (subst_eqs _#5 (subof_conv_0 _#3 (eqs_fst _#5 Hwf Heq1 Hfst Hfst')) Heq2') as Heq2''.
  simp_sub in Heq2''.
  do 3 eexists; do2 3 split; eauto using eqs_trans; done.

(* quant *)
intros G T1 T1' k1 T2 T2' Heq1 _ Hfst1 Heq2 _ Hwf.
split.
  intros e1 e2 Heq.
  injection Heq; intros; subst e1 e2.
  do 3 eexists; do2 3 split; eauto; done.

  intros e1 e2 Heq.
  injection Heq; intros; subst e1 e2.
  so (can_fsts _ _ (eqs_reg _#3 Hwf Heq1 ander)) as (k1' & Hfst1').
  exists k1', T1, T2.
  do2 3 split; auto with static; [].
  so (subst_eqs _#5 (subof_conv_0 _#3 (eqk_symm (eqs_fst _#5 Hwf Heq1 Hfst1 Hfst1'))) Heq2) as Heq2'.
  simp_sub in Heq2'.
  apply eqs_symm; assumption.
Qed.


Lemma subs_one_invert' :
  forall G T,
    subs G T sg_one
    -> T = sg_one.
Proof.
intros G T H.
remember sg_one as E eqn:Heqe.
revert Heqe.
#2 induct H; try (intros; discriminate).

(* refl *)
intros G T T' Heqs Heqe.
so (eqs_one_inverts _#3 Heqs ander Heqe); [].
subst; reflexivity.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 Heq3.
so (IH2 Heq3) as Heq2.
subst T3.
so (IH1 Heq2) as Heq1.
subst T2 T1.
reflexivity.
Qed.


Lemma subs_stat_invert' :
  forall G k T,
    subs G T (sg_stat k)
    -> exists k', T = sg_stat k' /\ subk G k' k.
Proof.
intros G k T H.
remember (sg_stat k) as E eqn:Heqe.
revert k Heqe.
#3 induct H; try (intros; discriminate).

(* refl *)
intros G T T' Heqs k Heqe.
so (eqs_stat_inverts _#3 Heqs ander _ Heqe) as (k' & Heq & Heqk).
exists k'.
split; auto with static; done.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 t3 Heq3.
so (IH2 _ Heq3) as (t2 & Heq2 & Heqc).
so (IH1 _ Heq2) as (t1 & Heq1 & Heqc').
eauto using subk_trans; done.

(* stat *)
intros G k k' H k0 Heqe.
injection Heqe.
intros <-.
eauto; done.
Qed.


Lemma subs_dyn_invert :
  forall G t T,
    subs G (sg_dyn t) T
    -> exists t', T = sg_dyn t' /\ eqc G t t' kd_type.
Proof.
intros G t T H.
remember (sg_dyn t) as E eqn:Heqe.
revert t Heqe.
#2 induct H; try (intros; discriminate).

(* refl *)
intros G T T' Heqs t Heqe.
exact (eqs_dyn_inverts _#3 Heqs andel _ Heqe).

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 t1 Heq1.
so (IH1 _ Heq1) as (t2 & Heq2 & Heqc).
so (IH2 _ Heq2) as (t3 & Heq3 & Heqc').
eauto using eqc_trans; done.
Qed.


Lemma subs_dyn_invert' :
  forall G t T,
    subs G T (sg_dyn t)
    -> exists t', T = sg_dyn t' /\ eqc G t t' kd_type.
Proof.
intros G t T H.
remember (sg_dyn t) as E eqn:Heqe.
revert t Heqe.
#2 induct H; try (intros; discriminate).

(* refl *)
intros G T T' Heqs t Heqe.
exact (eqs_dyn_inverts _#3 Heqs ander _ Heqe).

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 t3 Heq3.
so (IH2 _ Heq3) as (t2 & Heq2 & Heqc).
so (IH1 _ Heq2) as (t1 & Heq1 & Heqc').
eauto using eqc_trans; done.
Qed.


Lemma subs_pigen_invert' :
  forall G T1 T2 T,
    wf G
    -> subs G T (sg_pigen T1 T2)
    -> exists T1' T2' k1,
         T = sg_pigen T1' T2'
         /\ fsts T1 k1
         /\ subs G T1 T1'
         /\ subs (G; cl_cn k1) T2' T2.
Proof.
intros G T1 T2 T Hwf H.
remember (sg_pigen T1 T2) as E eqn:Heqe.
revert T1 T2 Hwf Heqe.
#3 induct H; try (intros; discriminate).

(* refl *)
intros G T T' Heqs T1 T2 Hwf ->.
so (eqs_pigen_inverts _#3 Hwf Heqs ander _ _ (eq_refl _)) as (k1 & T1' & T2' & -> & Hfst1 & Heq1 & Heq2).
do 3 eexists; do2 3 split; eauto with static; done.

(* trans *)
intros G T T' T'' _ IH1 _ IH2 T1 T2 Hwf Heq.
so (IH2 _ _ Hwf Heq) as (? & ? & k & Heq' & Hfst & Hsub1 & Hsub2).
so (IH1 _ _ Hwf Heq') as (? & ? & k' & Heq'' & Hfst' & Hsub1' & Hsub2').
so (subst_subs _#5 (subof_subsume_0 _#3 (subs_fst _#5 Hwf Hsub1 Hfst Hfst')) Hsub2') as Hsub2''.
simp_sub in Hsub2''.
do 3 eexists; do2 3 split; eauto with static; done.

(* quant *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 _ HfstT1' Hsubs2 _ HfstT1 HofT2 E1 E2 Hwf Heqe.
injection Heqe.
intros; subst E1 E2.
do 3 eexists; do2 3 split; eauto; done.
Qed.


Lemma subs_pigen_invert'' :
  forall G T1 T2 T1' T2',
    wf G
    -> subs G (sg_pigen T1 T2) (sg_pigen T1' T2')
    -> exists k1',
         subs G T1' T1
         /\ fsts T1' k1'
         /\ subs (G; cl_cn k1') T2 T2'.
Proof.
intros G T1 T2 T1' T2' Hwf Hsubs.
so (subs_pigen_invert' _#4 Hwf Hsubs) as (U1 & U2 & k1' & Heq & Hfst & Hsubs1 & Hsubs2).
injection Heq; intros; subst U1 U2.
eauto.
Qed.


Lemma subs_piapp_invert' :
  forall G T1 T2 T,
    wf G
    -> subs G T (sg_piapp T1 T2)
    -> exists T1' T2' k1,
         T = sg_piapp T1' T2'
         /\ fsts T1 k1
         /\ subs G T1 T1'
         /\ subs (G; cl_cn k1) T2' T2.
Proof.
intros G T1 T2 T Hwf H.
remember (sg_piapp T1 T2) as E eqn:Heqe.
revert T1 T2 Hwf Heqe.
#3 induct H; try (intros; discriminate).

(* refl *)
intros G T T' Heqs T1 T2 Hwf ->.
so (eqs_piapp_inverts _#3 Hwf Heqs ander _ _ (eq_refl _)) as (k1 & T1' & T2' & -> & Hfst1 & Heq1 & Heq2).
do 3 eexists; do2 3 split; eauto with static; done.

(* trans *)
intros G T T' T'' _ IH1 _ IH2 T1 T2 Hwf Heq.
so (IH2 _ _ Hwf Heq) as (? & ? & k & Heq' & Hfst & Hsub1 & Hsub2).
so (IH1 _ _ Hwf Heq') as (? & ? & k' & Heq'' & Hfst' & Hsub1' & Hsub2').
so (subst_subs _#5 (subof_subsume_0 _#3 (subs_fst _#5 Hwf Hsub1 Hfst Hfst')) Hsub2') as Hsub2''.
simp_sub in Hsub2''.
do 3 eexists; do2 3 split; eauto with static; done.

(* quant *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 _ HfstT1' Hsubs2 _ HfstT1 HofT2 E1 E2 Hwf Heqe.
injection Heqe.
intros; subst E1 E2.
do 3 eexists; do2 3 split; eauto; done.
Qed.


Lemma subs_piapp_invert'' :
  forall G T1 T2 T1' T2',
    wf G
    -> subs G (sg_piapp T1 T2) (sg_piapp T1' T2')
    -> exists k1',
         subs G T1' T1
         /\ fsts T1' k1'
         /\ subs (G; cl_cn k1') T2 T2'.
Proof.
intros G T1 T2 T1' T2' Hwf Hsubs.
so (subs_piapp_invert' _#4 Hwf Hsubs) as (U1 & U2 & k1' & Heq & Hfst & Hsubs1 & Hsubs2).
injection Heq; intros; subst U1 U2.
eauto.
Qed.


Lemma subs_sigma_invert' :
  forall G T1 T2 T,
    wf G
    -> subs G T (sg_sigma T1 T2)
    -> exists T1' T2' k1',
         T = sg_sigma T1' T2'
         /\ fsts T1' k1'
         /\ subs G T1' T1
         /\ subs (G; cl_cn k1') T2' T2.
Proof.
intros G T1 T2 T Hwf H.
remember (sg_sigma T1 T2) as E eqn:Heqe.
revert T1 T2 Hwf Heqe.
#3 induct H; try (intros; discriminate).

(* refl *)
intros G T T' Heqs T1 T2 Hwf ->.
so (eqs_sigma_inverts _#3 Hwf Heqs ander _ _ (eq_refl _)) as (k1 & T1' & T2' & -> & Hfst1 & Heq1 & Heq2).
so (can_fsts _ _ (eqs_reg _#3 Hwf Heq1 ander)) as (k1' & Hfst1').
so (subst_eqs _#5 (subof_conv_0 _#3 (eqk_symm (eqs_fst _#5 Hwf Heq1 Hfst1 Hfst1'))) Heq2) as Heq2'.
simp_sub in Heq2'.
exists T1', T2', k1'.
do2 3 split; eauto with static; done.

(* trans *)
intros G T T' T'' _ IH1 _ IH2 T1 T2 Hwf Heq.
so (IH2 _ _ Hwf Heq) as (T1' & T2' & k & Heq' & Hfst & Hsub1 & Hsub2).
so (IH1 _ _ Hwf Heq') as (T1'' & T2'' & k' & Heq'' & Hfst' & Hsub1' & Hsub2').
so (subst_subs _#5 (subof_subsume_0 _#3 (subs_fst _#5 Hwf Hsub1' Hfst' Hfst)) Hsub2) as Hsub2''.
simp_sub in Hsub2''.
do 3 eexists; do2 3 split; eauto with static; done.

(* quant *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 _ HfstT1' Hsubs2 _ HfstT1 HofT2 E1 E2 Hwf Heqe.
injection Heqe.
intros; subst E1 E2.
do 3 eexists; do2 3 split; eauto; done.
Qed.
