
(* Principal types/signatures for terms/modules.  Principal kinds for signatures are in SubjectReduction. *)

Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.
Require Import Declarative.
Require Import Singleton.
Require Import Algorithm.
Require Import SubjectReduction.
Require Import Complete.
Require Import Sound.


(* Should we check that the rhs is well-formed? *)
Inductive ctxsub : context -> context -> Prop :=
| ctxsub_nil
    : ctxsub nil nil

| ctxsub_cn {G G' k k'}
    : ctxsub G G'
      -> subk G k k'
      -> ctxsub (G; cl_cn k) (G'; cl_cn k')

| ctxsub_tm {G G' t t'}
    : ctxsub G G'
      -> eqc G t t' kd_type
      -> ctxsub (G; cl_tm t) (G'; cl_tm t')

| ctxsub_md {G G' T T' k k'}
    : ctxsub G G'
      -> subs G T T'
      -> fsts T k
      -> fsts T' k'
      -> ctxsub (G; cl_cn k; cl_md (shift1 T) (var 0)) (G'; cl_cn k'; cl_md (shift1 T') (var 0))
.


Lemma ctxsub_refl :
  forall G, wf G -> ctxsub G G.
Proof.
intros G Hwf.
induct Hwf; eauto using ctxsub with static.
Qed.


Lemma ctxsub_wf :
  forall G G', ctxsub G G' -> wf G.
Proof.
intros G G' Hsub.
induct Hsub.

(* nil *)
apply wf_nil; done.

(* cn *)
intros G G' k k' _ Hwf Hsubk.
so (subk_reg _#3 Hwf Hsubk) as (Hof & _).
apply wf_cn; auto; done.

(* tm *)
intros G G' t t' _ Hwf Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hof & _).
apply wf_tm; auto; done.

(* md *)
intros G G' T T' k k' _ Hwf Hsubs Hfst _.
so (subs_reg _#3 Hwf Hsubs) as (Hof & _).
apply wf_md; auto; done.
Qed.


Lemma ctxsub_subof :
  forall G G',
    ctxsub G' G
    -> subof dof G' (under (length G) id) G.
Proof.
intros G G' Hctxsub.
induct Hctxsub.

(* nil *)
simpl; auto with static; done.

(* cn *)
intros G G' k k' _ IH Heqk.
simpl length.
rewrite -> under_succ.
#2 apply subof_dot.
  so (subof_weaken_dof _ (nil; cl_cn k) _ _ IH) as Hsubof.
  simpl in Hsubof.
  assumption.

  simp_sub.
  apply dof_cn; [].
  rewrite -> subst_compose.
  rewrite -> subst_under_id.
  #2 apply (@ofc_subsume _ (shift1 k)).
    apply ofc_var; auto using index_0; done.

    eapply subst_subk; eauto using weaken1; done.

(* tm *)
intros G G' t t' _ IH Heqc.
simpl length.
rewrite -> under_succ.
#2 apply subof_dot.
  so (subof_weaken_dof _ (nil; cl_tm t) _ _ IH) as Hsubof.
  simpl in Hsubof.
  assumption.

  simp_sub.
  apply dof_tm; [].
  rewrite -> subst_compose.
  rewrite -> subst_under_id.
  #2 apply (@oft_equiv _ (shift1 t)).
    apply oft_var; auto using index_0; done.

    replsub kd_type with (shift1 kd_type).
    eapply subst_eqc; eauto using weaken1; done.

(* md *)
intros G G' T T' k k' Hctxsub IH Hsubs Hfst Hfst'.
simpl length.
rewrite -> under_succ.
#2 apply subof_dot.
  rewrite -> under_succ.
  simp_sub.
  #2 apply subof_dot.
    so (subof_weaken_dof _ (nil; cl_cn k; cl_md (shift1 T) (var 0)) _ _ IH) as Hsubof.
    simpl in Hsubof.
    assumption.

    simp_sub.
    apply dof_cn; [].
    rewrite -> subst_compose.
    rewrite -> subst_under_id.
    #2 apply (@ofc_subsume _ (subst (sh 2) k)).
      apply ofc_var; eauto with sequence; done.
      
      so (subs_fst _#5 (ctxsub_wf _ _ Hctxsub) Hsubs Hfst Hfst').
      eapply subst_subk; eauto using weaken2; done.

  simp_sub.
  rewrite -> subst_compose; [].
  rewrite -> subst_under_id; [].
  #2 apply dof_md.
    #2 apply (@ofm_subsume _ _ (subst (sh 2) T)).
      replsub (subst (sh 2) T) with (shift1 (shift1 T)).
      eapply ofm_var; eauto with sequence; done.

      eapply subst_subs; eauto using weaken2; done.

    replsub (var 1) with (shift1 (var 0)).
    eapply fstm_var; eauto with sequence; done.
Qed.



Lemma ctxsub_cons_invert :
  forall G G0' C',
    ctxsub G (G0'; C')
    -> exists G0 C,
         G = G0; C
         /\ ctxsub G0 G0'.
Proof.
intros G G0' C' H.
invert H; try (intros; subst; do 2 eexists; split; eauto; done); [].
intros G0 G' T T' k k' Hsub Hsubs Hfst Hfst' <- <- <-.
exists (G0; cl_cn k).
eexists.
split; eauto; [].
so (subs_fst _#5 (ctxsub_wf _ _ Hsub) Hsubs Hfst Hfst').
apply ctxsub_cn; eauto; done.
Qed.


Definition purity_leq (p p' : purity) := p = pure \/ p' = impure.


Definition purity_join (p1 p2 : purity) :=
  (match p1 with
   | pure => p2
   | impure => impure
   end).


Lemma ofm_forget_leq :
  forall G p p' M T,
    ofm G p M T
    -> purity_leq p p'
    -> ofm G p' M T.
Proof.
intros G p p' M T Hof Hleq.
#2 destruct p; destruct p'; auto.
  apply ofm_forget; assumption.

  invert Hleq; done.
Qed.


Lemma purity_join_leq1 :
  forall p1 p2, purity_leq p1 (purity_join p1 p2).
Proof.
intros p1 p2.
destruct p1; destruct p2; simpl; [left | right ..]; auto.
Qed.


Lemma purity_join_leq2 :
  forall p1 p2, purity_leq p2 (purity_join p1 p2).
Proof.
intros p1 p2.
destruct p1; destruct p2; simpl; [left | right ..]; auto.
Qed.


(* There is a discrepancy between this definition and the definition of
   principal kinds: in checking contexts, the principal kind judgement
   just uses ofc, while here we synthesize a principal type/signature and
   check equality/subsignature.  This is because we want to show that
   print/prins implies oftp/ofmp.  Maybe we should change prink to be
   consistent.
*)
Inductive print : context -> (* tm *) term -> (* cn *) term -> Prop :=
| print_var {G i t}
    : index i G (cl_tm t)
      -> print G (var i) (subst (sh (S i)) t)

| print_star {G}
    : print G tm_star cn_unit

| print_lam {G t1 t2 e}
    : ofc G t1 kd_type
      -> print (G; cl_tm t1) e t2
      -> print G (tm_lam t1 e) (cn_arrow t1 (subst1 nonsense t2))

| print_app {G t t1 t2 t1' e1 e2}
    : print G e1 t
      -> headnorm G t (cn_arrow t1 t2)
      -> print G e2 t1'
      -> eqc G t1' t1 kd_type
      -> print G (tm_app e1 e2) t2

| print_pair {G t1 t2 e1 e2}
    : print G e1 t1
      -> print G e2 t2
      -> print G (tm_pair e1 e2) (cn_prod t1 t2)

| print_pi1 {G t t1 t2 e}
    : print G e t
      -> headnorm G t (cn_prod t1 t2)
      -> print G (tm_pi1 e) t1

| print_pi2 {G t t1 t2 e}
    : print G e t
      -> headnorm G t (cn_prod t1 t2)
      -> print G (tm_pi2 e) t2

| print_plam {G k t e}
    : ofk G k
      -> print (G; cl_cn k) e t
      -> print G (tm_plam k e) (cn_all k t)

| print_papp {G u k t e c}
    : print G e u
      -> headnorm G u (cn_all k t)
      -> ofc G c k
      -> print G (tm_papp e c) (subst1 c t)

| print_pack {G k t c e u}
    : ofc G c k
      -> print G e u
      -> eqc G u (subst1 c t) kd_type
      -> ofc (G; cl_cn k) t kd_type
      -> print G (tm_pack c e k t) (cn_exists k t)

| print_unpack {G u k t t' e1 e2}
    : print G e1 u
      -> headnorm G u (cn_exists k t)
      -> print (G; cl_cn k; cl_tm t) e2 t'
      -> eqc (G; cl_cn k) (subst (dot nonsense id) t') (subst (dot nonsense (dot (shift1 (inh k)) sh1)) t') kd_type
      -> print G (tm_unpack e1 e2) (subst (dot nonsense (dot (inh k) id)) t')

| print_fix {G t e u}
    : print G e u
      -> eqc G u (cn_arrow (cn_arrow cn_unit t) t) kd_type
      -> print G (tm_fix t e) t

| print_lett {G t1 t2 e1 e2}
    : print G e1 t1
      -> print (G; cl_tm t1) e2 t2
      -> print G (tm_lett e1 e2) (subst1 nonsense t2)

| print_letm {G T k t M e p}
    : prins G p M T
      -> fsts T k
      -> print (G; cl_cn k; cl_md (shift1 T) (var 0)) e t
      -> eqc (G; cl_cn k) (subst (dot nonsense id) t) (subst (dot nonsense (dot (shift1 (inh k)) sh1)) t) kd_type
      -> print G (tm_letm M e) (subst (dot nonsense (dot (inh k) id)) t)

| print_ext {G t M p}
    : prins G p M (sg_dyn t)
      -> print G (tm_ext M) t


with prins : context -> purity -> (* cn *) term -> (* kd *) term -> Prop :=
| prins_var {G i T c}
    : index i G (cl_md T c)
      -> prins G pure (var i) (singles (subst (sh (S i)) c) (subst (sh (S i)) T))

| prins_star {G}
    : prins G pure md_star sg_one

| prins_stat {G c k}
    : prink G c k
      -> prins G pure (md_stat c) (sg_stat k)

| prins_dyn {G e t}
    : print G e t
      -> prins G pure (md_dyn e) (sg_dyn t)

| prins_lamgn {G T1 T2 k1 M p}
    : ofs G T1
      -> fsts T1 k1
      -> prins (G; cl_cn k1; cl_md (shift1 T1) (var 0)) p M (shift1 T2)
      -> prins G pure (md_lamgn T1 M) (sg_pigen T1 T2)

| prins_appgn {G T1 T2 T1' M1 M2 c2 p}
    : prins G p M1 (sg_pigen T1 T2)
      -> prins G pure M2 T1'
      -> subs G T1' T1
      -> fstm G M2 c2
      -> prins G impure (md_appgn M1 M2) (subst1 c2 T2)

| prins_lamap {G T1 T2 k1 M}
    : ofs G T1
      -> fsts T1 k1
      -> prins (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure M (shift1 T2)
      -> prins G pure (md_lamap T1 M) (sg_piapp T1 T2)

| prins_appap {G p T1 T2 T1' M1 M2 c2}
    : prins G p M1 (sg_piapp T1 T2)
      -> prins G pure M2 T1'
      -> subs G T1' T1
      -> fstm G M2 c2
      -> prins G p (md_appap M1 M2) (subst1 c2 T2)

| prins_pair {G p1 p2 T1 T2 M1 M2}
    : prins G p1 M1 T1
      -> prins G p2 M2 T2
      -> prins G (purity_join p1 p2) (md_pair M1 M2) (sg_sigma T1 (shift1 T2))

| prins_pi1 {G T1 T2 M}
    : prins G pure M (sg_sigma T1 T2)
      -> prins G pure (md_pi1 M) T1

| prins_pi2 {G T1 T2 M c}
    : prins G pure M (sg_sigma T1 T2)
      -> fstm G M c
      -> prins G pure (md_pi2 M) (subst1 (cn_pi1 c) T2)

| prins_unpack {G p u k t T T' e M}
    : print G e u
      -> headnorm G u (cn_exists k t)
      -> prins (G; cl_cn k; cl_tm t) p M T'
      -> subs (G; cl_cn k; cl_tm t) T' (subst (sh 2) T)
      -> ofs G T
      -> prins G impure (md_unpack e M T) T

| prins_lett {G p t T e M}
    : print G e t
      -> prins (G; cl_tm t) p M T
      -> prins G p (md_lett e M) (subst1 nonsense T)

| prins_letm {G T1 T2 T2' k1 M1 M2 p1 p2}
    : prins G p1 M1 T1
      -> fsts T1 k1
      -> prins (G; cl_cn k1; cl_md (shift1 T1) (var 0)) p2 M2 T2'
      -> subs (G; cl_cn k1; cl_md (shift1 T1) (var 0)) T2' (subst (sh 2) T2)
      -> ofs G T2
      -> prins G impure (md_letm M1 M2 T2) T2

| prins_seal {G T T' M p}
    : prins G p M T'
      -> subs G T' T
      -> prins G impure (md_seal M T) T
.


Scheme print_mut_ind := Minimality for print Sort Prop
with   prins_mut_ind := Minimality for prins Sort Prop.
Combined Scheme prin_ind from print_mut_ind, prins_mut_ind.



Lemma weaken_prin_gen :
  (forall i j G1 G2 G3 e t,
     length G2 = i
     -> length G3 = j
     -> Forall iscl G3
     -> print (G1;; G3) e t
     -> print (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) e) (subst (under j (sh i)) t))
  /\
  (forall i j G1 G2 G3 p M T,
     length G2 = i
     -> length G3 = j
     -> Forall iscl G3
     -> prins (G1;; G3) p M T
     -> prins (G1;; G2;; subst_ctx (sh i) G3) p (subst (under j (sh i)) M) (subst (under j (sh i)) T)).
Proof.
exploit
  (prin_ind
     (fun G e t =>
        forall i j G1 G2 G3,
          G = G1;; G3
          -> length G2 = i
          -> length G3 = j
          -> Forall iscl G3
          -> print (G1;; G2;; subst_ctx (sh i) G3) (subst (under j (sh i)) e) (subst (under j (sh i)) t))
     (fun G p M T =>
        forall i j G1 G2 G3,
          G = G1;; G3
          -> length G2 = i
          -> length G3 = j
          -> Forall iscl G3
          -> prins (G1;; G2;; subst_ctx (sh i) G3) p (subst (under j (sh i)) M) (subst (under j (sh i)) T))) as Hind.

(* var *)
intros G n t Hindex i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (Compare_dec.lt_dec n j) as [Hlt | Hnlt].
  rewrite -> project_under_lt; auto; [].
  so (index_app_lt _#5 (eqconv (eqsymm Hlen3) Hlt) Hindex) as Hindex'.
  so (index_subst_ctx _#3 (sh i) Hindex') as Hindex''.
  rewrite -> Hlen3 in Hindex''.
  so (index_app_left _ _ (G1;; G2) _ _ Hindex'') as Hindex'''.
  simp_sub in Hindex'''.
  so (print_var Hindex''') as Hprint.
  simp_sub in Hprint.
  rewrite -> compose_sh_under_leq; [| omega].
  repl (j - (1 + n)) with (j - n - 1) by omega.
  assumption.

  assert (n >= j) as Hgeq by omega; clear Hnlt.
  rewrite -> project_under_ge; auto; [].
  simp_sub.
  repl (j + i + (n - j)) with (n + i) by omega.
  so (index_app_ge _#5 (eqconv (eqsymm Hlen3) Hgeq) Hindex) as Hindex'.
  so (index_app_right _ (G2;; subst_ctx (sh i) G3) _#3 Hindex') as Hindex''.
  rewrite -> app_length in Hindex''; [].
  rewrite -> subst_ctx_length in Hindex''; [].
  rewrite -> Hlen2 in Hindex''; [].
  rewrite -> Hlen3 in Hindex''; [].
  rewrite <- app_assoc in Hindex''; [].
  repl (n - j + (j + i)) with (n + i) in Hindex'' by omega.
  so (print_var Hindex'') as Hprint.
  rewrite -> compose_sh_under_geq; auto; [].
  simp_sub.
  repl (1 + n - j + i + j) with (S (n + i)) by omega.
  assumption.

(* star *)
intros G i j G1 G2 G3 -> H0 H1 H2.     
simp_sub.
apply print_star; done.

(* lam *)
intros G t1 t2 e Hoft1 _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
replsub (subst (dot nonsense (under j (sh i))) t2) with (subst1 nonsense (subst (under (S j) (sh i)) t2)).
#2 apply print_lam.
  eapply subst_ofc_type; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_tm (subst (under j (sh i)) t1)) with (subst_ctx (sh i) (G3; cl_tm t1)) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH; auto.
    simpl; auto; done.

    apply Forall_cons; auto with static; done.

(* app *)
intros G t t1 t2 t1' e1 e2 _ IH1 Hnorm _ IH2 Heqc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint1.
so (IH2 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint2.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
eapply print_app; eauto; [].
eapply subst_eqc_type; eauto; [].
subst i j.
apply subof_under_sof; auto; [].
apply subof_sh; [].
apply app_truncate; done.

(* pair *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint1.
so (IH2 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint2.
eapply print_pair; eauto; done.

(* pi1 *)
intros G t t1 t2 e _ IH Hnorm i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
so (IH _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
eapply print_pi1; eauto; done.

(* pi2 *)
intros G t t1 t2 e _ IH Hnorm i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
so (IH _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
eapply print_pi2; eauto; done.

(* plam *)
intros G k t e Hofk _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
#2 apply print_plam.
  eapply subst_ofk; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k)) with (subst_ctx (sh i) (G3; cl_cn k)) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH; auto.
    simpl; auto; done.

    apply Forall_cons; auto with static; done.

(* papp *)
intros G u k t e c _ IH Hnorm Hofc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
rewrite -> split_dot; rewrite -> subst_compose; [].
eapply print_papp; eauto; [].
eapply subst_ofc; eauto; [].
subst i j.
apply subof_under_sof; auto; [].
apply subof_sh; [].
apply app_truncate; done.

(* pack *)
intros G k t c e u Hofc _ IH Heqc Hoft i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
assert (subof sof (G1;; G2;; subst_ctx (sh i) G3) (under j (sh i)) (G1;; G3)) as Hsubof.
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.
#2 eapply print_pack; eauto using subst_ofc.
  simp_sub.
  so (subst_eqc _#6 Hsubof Heqc) as Heqc'.
  simp_sub in Heqc'.
  assumption.

  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k)) with (subst_ctx (sh i) (G3; cl_cn k)) by (simpl; simp_sub; subst; reflexivity).
  eapply subst_ofc_type; eauto; [].
  subst i j.
  repl (G1;; G3; cl_cn k) with (G1;; (G3; cl_cn k)) by (simpl; reflexivity).
  #2 apply subof_under_sof; auto.
    apply Forall_cons; auto with static; done.

    apply subof_sh; [].
    apply app_truncate; done.

(* unpack *)
intros G u k t t' e1 e2 _ IH1 Hnorm _ IH2 Heqc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint1.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
rewrite -> shift_under_inh; [].
replsub (subst (dot nonsense (dot (inh (subst (under j (sh i)) k)) (under j (sh i)))) t') with (subst (dot nonsense (dot (inh (subst (under j (sh i)) k)) id)) (subst (under (S (S j)) (sh i)) t')).
#2 eapply print_unpack; eauto.
  simp_sub.
  rewrite <- !under_succ2; [].
  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k); cl_tm (subst (under (S j) (sh i)) t)) with (subst_ctx (sh i) (G3; cl_cn k; cl_tm t)) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH2; auto.
    simpl; auto; done.

    apply Forall_cons; auto with static; done.

  simp_sub. 
  #2 sor (refine (subst_eqc (G1;; G2;; subst_ctx (sh i) (G3; cl_cn k)) _ (under (S j) (sh i)) _#4 Heqc)) as Heqc'.
    subst i j.
    repl (G1;; G3; cl_cn k) with (G1;; (G3; cl_cn k)) by (simpl; reflexivity).
    #2 apply subof_under_sof; auto.
      apply Forall_cons; auto with static; done.
  
      apply subof_sh; [].
      apply app_truncate; done.
  simpl in Heqc'.
  simp_sub in Heqc'.
  replsub (subst (compose (under j (sh i)) sh1) (inh k)) with (subst (sh 1) (subst (under j (sh i)) (inh k))) in Heqc'.
  rewrite -> shift_under_inh in Heqc'.
  subst i j; assumption.

(* fix *)
intros G t e u _ IH Heqc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
eapply print_fix; eauto; [].
replsub (cn_arrow (cn_arrow cn_unit (subst (under j (sh i)) t)) (subst (under j (sh i)) t)) with (subst (under j (sh i)) (cn_arrow (cn_arrow cn_unit t) t)).
eapply subst_eqc_type; eauto; [].
subst i j.
apply subof_under_sof; auto; [].
apply subof_sh; [].
apply app_truncate; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
replsub (subst (dot nonsense (under j (sh i))) t2) with (subst1 nonsense (subst (under (S j) (sh i)) t2)).
eapply print_lett; eauto; [].
rewrite <- under_succ; [].
rewrite -> app_comm_cons; [].
repl (subst_ctx (sh i) G3; cl_tm (subst (under j (sh i)) t1)) with (subst_ctx (sh i) (G3; cl_tm t1)) by (simpl; simp_sub; subst; reflexivity).
#2 apply IH2; auto.
  simpl; auto; done.

  apply Forall_cons; auto with static; done.

(* letm *)
intros G T k t M e p _ IH1 Hfst _ IH2 Heqc i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprin1.
rewrite -> shift_under_inh; [].
replsub (subst (dot nonsense (dot (inh (subst (under j (sh i)) k)) (under j (sh i)))) t) with (subst (dot nonsense (dot (inh (subst (under j (sh i)) k)) id)) (subst (under (S (S j)) (sh i)) t)).
#2 eapply print_letm; eauto using subst_fsts.
  simp_sub.
  rewrite <- !under_succ2; [].
  replsub (subst (compose (under j (sh i)) sh1) T) with (subst (under (S j) (sh i)) (shift1 T)).
  rewrite -> app_comm_cons; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k); cl_md (subst (under (S j) (sh i)) (shift1 T)) (var 0)) with (subst_ctx (sh i) (G3; cl_cn k; cl_md (shift1 T) (var 0))) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH2; auto.
    simpl; auto; done.

    apply Forall_cons; auto with static; done.

  simp_sub.
  #2 sor (refine (subst_eqc (G1;; G2;; subst_ctx (sh i) (G3; cl_cn k)) _ (under (S j) (sh i)) _#4 Heqc)) as Heqc'.
    subst i j.
    repl (G1;; G3; cl_cn k) with (G1;; (G3; cl_cn k)) by (simpl; reflexivity).
    #2 apply subof_under_sof; auto.
      apply Forall_cons; auto with static; done.
  
      apply subof_sh; [].
      apply app_truncate; done.
  simpl in Heqc'.
  simp_sub in Heqc'.
  replsub (subst (compose (under j (sh i)) sh1) (inh k)) with (subst (sh 1) (subst (under j (sh i)) (inh k))) in Heqc'.
  rewrite -> shift_under_inh in Heqc'.
  subst i j; assumption.

(* ext *)
intros G t M p _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
eapply print_ext; eauto; done.

(* var *)
intros G n T c Hindex i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (Compare_dec.lt_dec n j) as [Hlt | Hnlt].
  rewrite -> project_under_lt; auto; [].
  so (index_app_lt _#5 (eqconv (eqsymm Hlen3) Hlt) Hindex) as Hindex'.
  so (index_subst_ctx _#3 (sh i) Hindex') as Hindex''.
  rewrite -> Hlen3 in Hindex''.
  so (index_app_left _ _ (G1;; G2) _ _ Hindex'') as Hindex'''.
  simp_sub in Hindex'''.
  so (prins_var Hindex''') as Hprins.
  simp_sub in Hprins.
  rewrite -> shift_under_singles; [].
  simp_sub.
  rewrite -> compose_sh_under_leq; [| omega].
  repl (j - (1 + n)) with (j - n - 1) by omega.
  assumption.

  assert (n >= j) as Hgeq by omega; clear Hnlt.
  rewrite -> project_under_ge; auto; [].
  simp_sub.
  repl (j + i + (n - j)) with (n + i) by omega.
  so (index_app_ge _#5 (eqconv (eqsymm Hlen3) Hgeq) Hindex) as Hindex'.
  so (index_app_right _ (G2;; subst_ctx (sh i) G3) _#3 Hindex') as Hindex''.
  rewrite -> app_length in Hindex''; [].
  rewrite -> subst_ctx_length in Hindex''; [].
  rewrite -> Hlen2 in Hindex''; [].
  rewrite -> Hlen3 in Hindex''; [].
  rewrite <- app_assoc in Hindex''; [].
  repl (n - j + (j + i)) with (n + i) in Hindex'' by omega.
  so (prins_var Hindex'') as Hprins.
  rewrite -> shift_under_singles; [].
  simp_sub.
  rewrite -> compose_sh_under_geq; auto; [].
  simp_sub.
  repl (1 + n - j + i + j) with (S (n + i)) by omega.
  assumption.

(* star *)
intros G i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
apply prins_star; done.

(* stat *)
intros G c k Hprink i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
apply prins_stat; [].
apply weaken_prink_gen; auto; done.

(* dyn *)
intros G e t _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
apply prins_dyn; auto; done.

(* lamgn *)
intros G T1 T2 k1 M p HofT1 Hfst _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
#2 eapply prins_lamgn; eauto using subst_fsts.
  eapply subst_ofs; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  simp_sub.
  #2 exploit (IH i (S (S j)) G1 G2 (G3; cl_cn k1; cl_md (shift1 T1) (var 0))) as H; auto with static.
    simpl.
    omega.
  simpl in H.
  simp_sub in H.
  subst i j; eassumption.

(* appgn *)
intros G T1 T2 T1' M1 M2 c2 p _ IH1 _ IH2 Hsubs Hfstm i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
rewrite -> split_dot; rewrite -> subst_compose; [].
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprin1.
simp_sub in Hprin1.
#2 eapply prins_appgn; eauto.
  eapply subst_subs; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  eapply subst_fstm; eauto; [].
  subst i j.
  apply subof_under_tof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

(* lamap *)
intros G T1 T2 k1 M HofT1 Hfst _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
#2 eapply prins_lamap; eauto using subst_fsts.
  eapply subst_ofs; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  simp_sub.
  #2 exploit (IH i (S (S j)) G1 G2 (G3; cl_cn k1; cl_md (shift1 T1) (var 0))) as H; auto with static.
    simpl.
    omega.
  simpl in H.
  simp_sub in H.
  subst i j; eassumption.

(* appgn *)
intros G p T1 T2 T1' M1 M2 c2 _ IH1 _ IH2 Hsubs Hfstm i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
rewrite -> split_dot; rewrite -> subst_compose; [].
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprin1.
simp_sub in Hprin1.
#2 eapply prins_appap; eauto.
  eapply subst_subs; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

  eapply subst_fstm; eauto; [].
  subst i j.
  apply subof_under_tof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

(* pair *)
intros G p1 p2 T1 T2 M1 M2 _ IH1 _ IH2 i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
rewrite -> subst_compose; [].
eapply prins_pair; eauto; done.

(* pi1 *)
intros G T1 T2 M _ IH i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
eapply prins_pi1; eauto; done.

(* pi2 *)
intros G T1 T2 M c _ IH Hfstm i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
rewrite -> split_dot; rewrite -> subst_compose; [].
so (IH _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as H.
simp_sub in H.
eapply prins_pi2; eauto; [].
eapply subst_fstm; eauto; [].
subst i j.
apply subof_under_tof; auto; [].
apply subof_sh; [].
apply app_truncate; done.

(* unpack *)
intros G p u k t T T' e M _ IH1 Hnorm _ IH2 Hsubs HofT i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprint1.
so (weaken_headnorm_gen _#7 Hlen2 Hlen3 Hnorm) as Hnorm'.
simp_sub in Hnorm'.
#3 eapply (@prins_unpack _#6 (subst (under (2 + j) (sh i)) T')); eauto.
  simp_sub.
  rewrite <- !under_succ2; [].
  rewrite <- under_succ; [].
  rewrite -> app_comm_cons; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k); cl_tm (subst (under (S j) (sh i)) t)) with (subst_ctx (sh i) (G3; cl_cn k; cl_tm t)) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH2; auto.
    simpl; auto; done.

    apply Forall_cons; auto with static; done.

  simp_sub.
  #2 sor (refine (subst_subs (G1;; G2;; subst_ctx (sh i) (G3; cl_cn k; cl_tm t)) _ (under (2 + j) (sh i)) _#3 Hsubs)) as Hsubs'.
    subst i j.
    repl (G1;; G3; cl_cn k; cl_tm t) with (G1;; (G3; cl_cn k; cl_tm t)) by (simpl; reflexivity).
    #2 apply subof_under_sof; auto.
      apply Forall_cons; auto with static; done.
  
      apply subof_sh; [].
      apply app_truncate; done.
  simp_sub in Hsubs'.
  rewrite -> compose_sh_under in Hsubs'.
  simpl subst_ctx in Hsubs'.
  simp_sub in Hsubs'.
  subst i j; assumption.

  eapply subst_ofs; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

(* lett *)
intros G p t T e M _ IH1 _ IH2 i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
replsub (subst (dot nonsense (under j (sh i))) T) with (subst1 nonsense (subst (under (S j) (sh i)) T)).
eapply prins_lett; eauto; [].
rewrite <- under_succ; [].
rewrite -> app_comm_cons; [].
repl (subst_ctx (sh i) G3; cl_tm (subst (under j (sh i)) t)) with (subst_ctx (sh i) (G3; cl_tm t)) by (simpl; simp_sub; subst; reflexivity).
#2 apply IH2; auto.
  simpl; auto; done.

  apply Forall_cons; auto with static; done.

(* letm *)
intros G T1 T2 T2' k1 M1 M2 p1 p2 _ IH1 Hfst _ IH2 Hsubs HofHs i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
simp_sub.
so (IH1 _#5 (eq_refl _) Hlen2 Hlen3 Hallcl) as Hprin1.
#3 eapply prins_letm; eauto using subst_fsts.
  simp_sub.
  rewrite <- !under_succ2; [].
  replsub (subst (compose (under j (sh i)) sh1) T1) with (subst (under (S j) (sh i)) (shift1 T1)).
  rewrite -> app_comm_cons; [].
  rewrite -> app_comm_cons; [].
  repl (subst_ctx (sh i) G3; cl_cn (subst (under j (sh i)) k1); cl_md (subst (under (S j) (sh i)) (shift1 T1)) (var 0)) with (subst_ctx (sh i) (G3; cl_cn k1; cl_md (shift1 T1) (var 0))) by (simpl; simp_sub; subst; reflexivity).
  #2 apply IH2; auto.
    simpl; auto; done.

    apply Forall_cons; auto with static; done.

  simp_sub.
  #2 sor (refine (subst_subs (G1;; G2;; subst_ctx (sh i) (G3; cl_cn k1; cl_md (shift1 T1) (var 0))) _ (under (2 + j) (sh i)) _#3 Hsubs)) as Hsubs'.
    subst i j.
    repl (G1;; G3; cl_cn k1; cl_md (shift1 T1) (var 0)) with (G1;; (G3; cl_cn k1; cl_md (shift1 T1) (var 0))) by (simpl; reflexivity).
    #2 apply subof_under_sof; auto.
      apply Forall_cons; auto with static; done.
  
      apply subof_sh; [].
      apply app_truncate; done.
  simp_sub in Hsubs'.
  rewrite -> compose_sh_under in Hsubs'.
  simpl subst_ctx in Hsubs'.
  simp_sub in Hsubs'.
  rewrite <- under_succ2; [].
  subst i j; assumption.

  eapply subst_ofs; eauto; [].
  subst i j.
  apply subof_under_sof; auto; [].
  apply subof_sh; [].
  apply app_truncate; done.

(* seal *)
intros G T T' M p _ IH Hsubs i j G1 G2 G3 -> Hlen2 Hlen3 Hallcl.
eapply prins_seal; eauto; [].
eapply subst_subs; eauto; [].
subst i j.
apply subof_under_sof; auto; [].
apply subof_sh; [].
apply app_truncate; done.

(* epilogue *)
destruct Hind.
split; intros; eauto; done.
Qed.


Lemma weaken1_prins :
  forall G p M T A,
    prins G p M T
    -> prins (G; A) p (shift1 M) (shift1 T).
Proof.
intros G p M T A Hprins.
so (weaken_prin_gen ander 1 0 G (nil; A) nil _ _ _ (eq_refl _) (eq_refl _) (Forall_nil _) Hprins) as H.
simp_sub in H.
assumption.
Qed.


Lemma prin_fun :
  (forall G e t t',
     print G e t
     -> print G e t'
     -> t = t')
  /\
  (forall G M p p' T T',
     prins G p M T
     -> prins G p' M T'
     -> p = p' /\ T = T').
Proof.
exploit (prin_ind
           (fun G e t =>
              forall t',
                print G e t' -> t = t')
           (fun G p M T =>
              forall p' T',
                prins G p' M T'
                -> p = p' /\ T = T')) as Hind.

(* var *)
intros G i t Hindex t' Hprint.
invert Hprint; [].
intros t'' Hindex' <-.
f_equal; [].
so (index_fun _#5 Hindex Hindex') as H.
injection H.
auto; done.

(* star *)
intros G t' H.
invert H; [].
intros; subst; auto; done.

(* lam *)
intros G t1 t2 e Hoft1 _ IH t' Hprint.
invert Hprint; [].
intros t2' _ Hprint' <-.
f_equal; [].
f_equal; [].
auto; done.

(* app *)
intros G t t1 t2 t1' e1 e2 _ IH1 Hnorm _ _ _ u2 Hprint.
invert Hprint; [].
intros u u1 _ Hprint' Hnorm' _ _.
so (IH1 _ Hprint'); subst u.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq; auto; done.

(* pair *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 u Hprint.
invert Hprint; [].
intros u1 u2 Hprint1 Hprint2 <-.
so (IH1 _ Hprint1).
so (IH2 _ Hprint2).
f_equal; auto; done.

(* pi1 *)
intros G t t1 t2 e _ IH Hnorm u1 Hprint.
invert Hprint; [].
intros u u2 Hprint' Hnorm'.
so (IH _ Hprint'); subst u.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq; auto; done.

(* pi2 *)
intros G t t1 t2 e _ IH Hnorm u2 Hprint.
invert Hprint; [].
intros u u1 Hprint' Hnorm'.
so (IH _ Hprint'); subst u.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq; auto; done.

(* plam *)
intros G k t e _ _ IH u Hprint.
invert Hprint; [].
intros t' _ Hprint' <-.
f_equal; auto; done.

(* papp *)
intros G u k t e c _ IH Hnorm _ v Hprint.
invert Hprint; [].
intros u' k' t' Hprint' Hnorm' _ <-.
f_equal; [].
so (IH _ Hprint'); subst u'.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq; auto; done.

(* pack *)
intros G k t c e _ _ _ _ _ _ u Hprint.
invert Hprint; auto; done.

(* unpack *)
intros G u k t1 t2 e1 e2 _ IH1 Hnorm _ IH2 _ v Hprint.
invert Hprint; [].
intros u' k' t1' t2' Hprint1 Hnorm' Hprint2 _ <-.
so (IH1 _ Hprint1); subst u'.
so (headnorm_fun _#4 Hnorm Hnorm') as Heq.
injection Heq.
intros <- <-.
so (IH2 _ Hprint2); subst t2'.
reflexivity.

(* fix *)
intros G t e _ _ _ _ u Hprint.
invert Hprint; [].
auto; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 u Hprint.
invert Hprint; [].
intros t1' t2' Hprint1 Hprint2 <-.
so (IH1 _ Hprint1); subst t1'.
so (IH2 _ Hprint2); subst t2'.
reflexivity.

(* letm *)
intros G T k t M e p _ IH1 Hfst _ IH2 _ u Hprint.
invert Hprint; [].
intros T' k' t' p' Hprint1 Hfst' Hprint2 _ <-.
so (IH1 _ _ Hprint1) as (<- & <-).
so (fsts_fun _#3 Hfst Hfst'); subst k'.
so (IH2 _ Hprint2); subst t'.
reflexivity.

(* ext *)
intros G t M p _ IH u Hprint.
invert Hprint; [].
intros p' Hprint'.
so (IH _ _ Hprint') as (_ & Heq).
injection Heq; auto; done.

(* var *)
intros G i T c Hindex p' U Hprint.
invert Hprint; [].
intros T' c' Hindex' <- <-.
split; trivial; [].
so (index_fun _#5 Hindex Hindex') as Heq.
injection Heq.
intros <- <-.
reflexivity.

(* star *)
intros G p' T' Hprins.
invert Hprins; [].
intros <- <-.
auto; done.

(* stat *)
intros G c k Hprink p' T' Hprins.
invert Hprins; [].
intros k' Hprink' <- <-.
split; trivial; [].
so (prink_fun _#4 Hprink Hprink'); subst k'.
reflexivity.

(* dyn *)
intros G e t _ IH p' T' Hprins.
invert Hprins; [].
intros t' Hprint <- <-.
so (IH _ Hprint); subst t'.
auto; done.

(* lamgn *)
intros G T1 T2 k1 M p _ Hfst _ IH p' T' Hprins.
invert Hprins; [].
intros T2' k1' p'' _ Hfst' Hprins' <- <-.
so (fsts_fun _#3 Hfst Hfst'); subst k1'.
so (IH _ _ Hprins') as (_ & Heq).
f_apply (subst1 nonsense) in Heq as Heq'.
simp_sub in Heq'.
subst T2'.
auto; done.

(* appgn *)
intros G T1 T2 T1' M1 M2 c2 p _ IH _ _ _ Hfstm p' U Hprins.
invert Hprins; [].
intros U1 U2 U1' c2' p'' Hprins' _ _ Hfstm' <- <-.
so (IH _ _ Hprins') as (_ & Heq).
injection Heq.
intros <- <-.
so (fstm_fun _#4 Hfstm Hfstm'); subst c2'.
auto; done.

(* lamap *)
intros G T1 T2 k1 M _ Hfst _ IH p' T' Hprins.
invert Hprins; [].
intros T2' k1' _ Hfst' Hprins' <- <-.
so (fsts_fun _#3 Hfst Hfst'); subst k1'.
so (IH _ _ Hprins') as (_ & Heq).
f_apply (subst1 nonsense) in Heq as Heq'.
simp_sub in Heq'.
subst T2'.
auto; done.

(* appap *)
intros G p T1 T2 T1' M1 M2 c2 _ IH _ _ _ Hfstm p' U Hprins.
invert Hprins; [].
intros U1 U2 U1' c2' Hprins' _ _ Hfstm' <-.
so (IH _ _ Hprins') as (Heqp & Heq).
injection Heq.
intros <- <-.
so (fstm_fun _#4 Hfstm Hfstm'); subst c2'.
auto; done.

(* pair *)
intros G p1 p2 T1 T2 M1 M2 _ IH1 _ IH2 p' T' Hprins.
invert Hprins; [].
intros p1' p2' T1' T2' Hprins1 Hprins2 <- <-.
so (IH1 _ _ Hprins1) as (<- & <-).
so (IH2 _ _ Hprins2) as (<- & <-).
auto; done.

(* pi1 *)
intros G T1 T2 M _ IH p' T' Hprins.
invert Hprins; [].
intros T2' Hprins' <-.
so (IH _ _ Hprins') as (_ & Heq).
injection Heq.
intros <- <-.
auto; done.

(* pi2 *)
intros G T1 T2 M c _ IH Hfstm p' T' Hprins.
invert Hprins; [].
intros T1' T2' c' Hprins' Hfstm' <- <-.
so (IH _ _ Hprins') as (_ & Heq).
injection Heq.
intros <- <-.
so (fstm_fun _#4 Hfstm Hfstm'); subst c'.
auto; done.

(* unpack *)
intros G p u k t T T' e M _ _ _ _ _ _ _ p' U Hprins.
invert Hprins; [].
auto; done.

(* lett *)
intros G p t T e M _ IH1 _ IH2 p' U Hprins.
invert Hprins; [].
intros t' T' Hprin1 Hprin2 <-.
so (IH1 _ Hprin1); subst t'.
so (IH2 _ _ Hprin2) as (<- & <-).
auto; done.

(* letm *)
intros G T1 T2 T2' k1 M1 M2 p1 p2 _ _ _ _ _ _ _ p' U Hprins.
invert Hprins; [].
auto; done.

(* seal *)
intros G T T' M p _ _ _ p' U Hprins.
invert Hprins; [].
auto; done.

(* epilogue *)
destruct Hind.
split; eauto; done.
Qed.


Lemma print_fun :
  forall G e t t',
    print G e t
    -> print G e t'
    -> t = t'.
Proof.
exact (prin_fun andel).
Qed.


Lemma prins_fun :
  forall G M p p' T T',
    prins G p M T
    -> prins G p' M T'
    -> p = p' /\ T = T'.
Proof.
exact (prin_fun ander).
Qed.


Inductive oftp : context -> (* tm *) term -> (* cn *) term -> Prop :=
| oftp_var {G i t}
    : index i G (cl_tm t)
      -> oftp G (var i) (subst (sh (S i)) t)

| oftp_star {G}
    : oftp G tm_star cn_unit

| oftp_lam {G t1 t2 e}
    : ofc G t1 kd_type
      -> oftp (G; cl_tm t1) e (shift1 t2)
      -> oftp G (tm_lam t1 e) (cn_arrow t1 t2)

| oftp_app {G t1 t2 e1 e2}
    : oftp G e1 (cn_arrow t1 t2)
      -> oftp G e2 t1
      -> oftp G (tm_app e1 e2) t2

| oftp_pair {G t1 t2 e1 e2}
    : oftp G e1 t1
      -> oftp G e2 t2
      -> oftp G (tm_pair e1 e2) (cn_prod t1 t2)

| oftp_pi1 {G t1 t2 e}
    : oftp G e (cn_prod t1 t2)
      -> oftp G (tm_pi1 e) t1

| oftp_pi2 {G t1 t2 e}
    : oftp G e (cn_prod t1 t2)
      -> oftp G (tm_pi2 e) t2

| oftp_plam {G k t e}
    : ofk G k
      -> oftp (G; cl_cn k) e t
      -> oftp G (tm_plam k e) (cn_all k t)

| oftp_papp {G k t e c}
    : oftp G e (cn_all k t)
      -> ofc G c k
      -> oftp G (tm_papp e c) (subst1 c t)

| oftp_pack {G k t c e}
    : ofc G c k
      -> oftp G e (subst1 c t)
      -> ofc (G; cl_cn k) t kd_type
      -> oftp G (tm_pack c e k t) (cn_exists k t)

| oftp_unpack {G k t t' e1 e2}
    : oftp G e1 (cn_exists k t)
      -> oftp (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t')
      -> ofc G t' kd_type
      -> oftp G (tm_unpack e1 e2) t'

| oftp_fix {G t e}
    : oftp G e (cn_arrow (cn_arrow cn_unit t) t)
      -> oftp G (tm_fix t e) t

| oftp_lett {G t1 t2 e1 e2}
    : oftp G e1 t1
      -> oftp (G; cl_tm t1) e2 (shift1 t2)
      -> oftp G (tm_lett e1 e2) t2

| oftp_equiv {G t t' e}
    : oftp G e t
      -> eqc G t t' kd_type
      -> oftp G e t'

| oftp_letm {G T k t M e}
    : ofmp G impure M T
      -> fsts T k
      -> oftp (G; cl_cn k; cl_md (shift1 T) (var 0)) e (subst (sh 2) t)
      -> ofc G t kd_type
      -> oftp G (tm_letm M e) t

| oftp_ext {G t M}
    : ofmp G impure M (sg_dyn t)
      -> oftp G (tm_ext M) t


with ofmp : context -> purity -> (* cn *) term -> (* kd *) term -> Prop :=
| ofmp_var {G i T c}
    : index i G (cl_md T c)
      -> ofmp G pure (var i) (singles (subst (sh (S i)) c) (subst (sh (S i)) T))

| ofmp_star {G}
    : ofmp G pure md_star sg_one

| ofmp_stat {G c k}
    : ofc G c k
      -> ofmp G pure (md_stat c) (sg_stat (single c k))

| ofmp_dyn {G e t}
    : oftp G e t
      -> ofmp G pure (md_dyn e) (sg_dyn t)

| ofmp_lamgn {G T1 T2 k1 M}
    : ofs G T1
      -> fsts T1 k1
      -> ofmp (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M (shift1 T2)
      -> ofmp G pure (md_lamgn T1 M) (sg_pigen T1 T2)

| ofmp_appgn {G T1 T2 M1 M2 c2}
    : ofmp G impure M1 (sg_pigen T1 T2)
      -> ofmp G pure M2 T1
      -> fstm G M2 c2
      -> ofmp G impure (md_appgn M1 M2) (subst1 c2 T2)

| ofmp_lamap {G T1 T2 k1 M}
    : ofs G T1
      -> fsts T1 k1
      -> ofmp (G; cl_cn k1; cl_md (shift1 T1) (var 0)) pure M (shift1 T2)
      -> ofmp G pure (md_lamap T1 M) (sg_piapp T1 T2)

| ofmp_appap {G p T1 T2 M1 M2 c2}
    : ofmp G p M1 (sg_piapp T1 T2)
      -> ofmp G pure M2 T1
      -> fstm G M2 c2
      -> ofmp G p (md_appap M1 M2) (subst1 c2 T2)

| ofmp_pair {G p T1 T2 M1 M2}
    : ofmp G p M1 T1
      -> ofmp G p M2 T2
      -> ofmp G p (md_pair M1 M2) (sg_sigma T1 (shift1 T2))

| ofmp_pi1 {G T1 T2 M}
    : ofmp G pure M (sg_sigma T1 T2)
      -> ofmp G pure (md_pi1 M) T1

| ofmp_pi2 {G T1 T2 M c}
    : ofmp G pure M (sg_sigma T1 T2)
      -> fstm G M c
      -> ofmp G pure (md_pi2 M) (subst1 (cn_pi1 c) T2)

| ofmp_unpack {G k t T e M}
    : oftp G e (cn_exists k t)
      -> ofmp (G; cl_cn k; cl_tm t) impure M (subst (sh 2) T)
      -> ofs G T
      -> ofmp G impure (md_unpack e M T) T

| ofmp_lett {G p t T e M}
    : oftp G e t
      -> ofmp (G; cl_tm t) p M (shift1 T)
      -> ofmp G p (md_lett e M) T

| ofmp_letm {G T1 T2 k1 M1 M2}
    : ofmp G impure M1 T1
      -> fsts T1 k1
      -> ofmp (G; cl_cn k1; cl_md (shift1 T1) (var 0)) impure M2 (subst (sh 2) T2)
      -> ofs G T2
      -> ofmp G impure (md_letm M1 M2 T2) T2

| ofmp_seal {G T M}
    : ofmp G impure M T
      -> ofmp G impure (md_seal M T) T

| ofmp_forget {G T M}
    : ofmp G pure M T
      -> ofmp G impure M T

| ofmp_subsume {G p T T' M}
    : ofmp G p M T
      -> subs G T T'
      -> ofmp G p M T'
.


Scheme oftp_mut_ind := Minimality for oftp Sort Prop
with   ofmp_mut_ind := Minimality for ofmp Sort Prop.
Combined Scheme dynamic_principal_ind from oftp_mut_ind, ofmp_mut_ind.


Lemma ofxp_ofx :
  (forall G e t,
     wf G
     -> oftp G e t
     -> oft G e t)
  /\
  (forall G p M T,
     wf G
     -> ofmp G p M T
     -> ofm G p M T).
Proof.
exploit
  (dynamic_principal_ind
     (fun G e t => wf G -> oft G e t)
     (fun G p M T => wf G -> ofm G p M T)) as Hind;
try (intros; eauto with static; done).

(* unpack *)
intros G k t t' e1 e2 _ IH1 _ IH2 Hoft' Hwf.
so (IH1 Hwf) as Hof1.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hof1)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
eauto with static; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 Hwf.
so (oft_reg _#3 Hwf (IH1 Hwf)).
eauto with static; done.

(* letm *)
intros G T k t M e _ IH1 Hfst _ Ih2 Hoft Hwf.
so (ofm_reg _#4 Hwf (IH1 Hwf)).
eauto with static; done.

(* var *)
intros G i T c Hindex Hwf.
eapply singles_ofm; eauto with static sequence; done.

(* stat *)
intros G c k Hof Hwf.
apply ofm_stat; [].
eapply single_ofc; eauto; done.

(* unpack *)
intros G k t T e M _ IH1 _ IH2 HofT Hwf.
so (IH1 Hwf) as Hof1.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hof1)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
eauto with static; done.

(* lett *)
intros G p t T e M _ IH1 _ IH2 Hwf.
so (oft_reg _#3 Hwf (IH1 Hwf)).
eauto with static; done.

(* letm *)
intros G T1 T2 k1 M1 M2 _ IH1 Hfst _ IH2 Hofs Hwf.
so (ofm_reg _#4 Hwf (IH1 Hwf)).
eauto with static; done.

(* epilogue *)
destruct Hind.
split; auto; done.
Qed.


Lemma oftp_oft :
  forall G e t,
    wf G
    -> oftp G e t
    -> oft G e t.
Proof.
exact (ofxp_ofx andel).
Qed.


Lemma ofmp_ofm :
  forall G p M T,
    wf G
    -> ofmp G p M T
    -> ofm G p M T.
Proof.
exact (ofxp_ofx ander).
Qed.


Lemma prinx_ofxp :
  (forall G e t,
     wf G
     -> print G e t
     -> oftp G e t)
  /\
  (forall G p M T,
     wf G
     -> prins G p M T
     -> ofmp G p M T).
Proof.
#20 exploit (prin_ind
           (fun G e t => wf G -> oftp G e t)
           (fun G p M T => wf G -> ofmp G p M T)) as Hind;
try (intros; eauto using oftp, ofmp, ofc_reg; done).

(* lam *)
intros G t1 t2 e Hofc _ IH Hwf.
so (wf_tm Hwf Hofc) as Hwf'.
so (IH Hwf') as Hofe.
so (oft_reg _#3 Hwf' (oftp_oft _#3 Hwf' Hofe)) as Hoft2.
apply oftp_lam; auto; [].
simp_sub.
erewrite <- subord_tm_cn; eauto; done.

(* app *)
intros G t t1 t2 t1' e1 e2 _ IH1 Hnorm _ IH2 Heqc2 Hwf.
so (IH1 Hwf) as Hofe1.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe1)) as Hoft.
so (headnorm_sound _#4 Hwf Hoft Hnorm) as Heqc1.
eapply oftp_app; eauto using oftp_equiv; done.

(* pi1 *)
intros G t t1 t2 e _ IH Hnorm Hwf.
so (IH Hwf) as Hofe.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hoft.
so (headnorm_sound _#4 Hwf Hoft Hnorm) as Heqc.
eapply oftp_pi1; eauto using oftp_equiv; done.

(* pi2 *)
intros G t t1 t2 e _ IH Hnorm Hwf.
so (IH Hwf) as Hofe.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hoft.
so (headnorm_sound _#4 Hwf Hoft Hnorm) as Heqc.
eapply oftp_pi2; eauto using oftp_equiv; done.

(* plam *)
intros G k t e Hofk _ IH Hwf.
so (IH (wf_cn Hwf Hofk)) as Hofe.
apply oftp_plam; auto; done.

(* papp *)
intros G u k t e c _ IH1 Hnorm Hofc Hwf.
so (IH1 Hwf) as Hofe.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hofu.
so (headnorm_sound _#4 Hwf Hofu Hnorm) as Heqc.
eapply oftp_papp; eauto using oftp_equiv; done.

(* unpack *)
intros G u k t t' e1 e2 _ IH1 Hnorm _ IH2 Heqinh Hwf.
so (IH1 Hwf) as Hofe1.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe1)) as Hofu.
so (headnorm_sound _#4 Hwf Hofu Hnorm) as Hequ.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Hequ ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear l Hofcr.
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
so (IH2 Hwf') as Hofe2.
so (oft_reg _#3 Hwf' (oftp_oft _#3 Hwf' Hofe2)) as Hoft'.
#2 eapply oftp_unpack; eauto using oftp_equiv.
  simp_sub.
  eapply oftp_equiv; eauto; [].
  repl t' with (subst (dot nonsense sh1) t') at 1.
    symmetry.
    eapply subord_tm_cn; eauto; done.
  so (subst_eqc _#6 (weaken1 _ _ (cl_tm t)) Heqinh) as Heqinh'.
  simp_sub in Heqinh'.
  assumption.

  eapply subst_ofc_type; eauto; [].
  #2 apply subof_dot.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply sof_cn; [].
    apply inh_ofc; auto; done.

    simp_sub.
    apply sof_tm; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hofe1.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe1)) as Hoft1.
so (wf_tm Hwf Hoft1) as Hwf'.
so (IH2 Hwf') as Hofe2.
so (oft_reg _#3 Hwf' (oftp_oft _#3 Hwf' Hofe2)) as Hoft2.
eapply oftp_lett; eauto; [].
simp_sub.
erewrite <- subord_tm_cn; eauto; done.

(* letm *)
intros G T k t M e p _ IH1 Hfst _ IH2 Heqc Hwf.
so (IH1 Hwf) as HofM.
so (ofm_reg _#4 Hwf (ofmp_ofm _#4 Hwf HofM)) as HofT.
so (wf_md Hwf HofT Hfst) as Hwf'.
so (IH2 Hwf') as Hofe.
so (oft_reg _#3 Hwf' (oftp_oft _#3 Hwf' Hofe)) as Hoft.
so (ofs_fst _#3 HofT Hfst) as Hofk.
#3 eapply oftp_letm; eauto.
  #2 destruct p.
    apply ofmp_forget; assumption.

    assumption.

  simp_sub.
  eapply oftp_equiv; eauto; [].
  repl t with (subst (dot nonsense sh1) t) at 1.
    symmetry.
    eapply subord_md_cn; eauto; done.

    so (subst_eqc _#6 (weaken1 _ _ (cl_md (shift1 T) (var 0))) Heqc) as Heqc'.
    simp_sub in Heqc'.
    assumption.
  eapply subst_ofc_type; eauto; [].
  #2 apply subof_dot.
    apply subof_dot; auto using subof_id; [].
    simp_sub.
    apply sof_cn; [].
    apply inh_ofc; auto; done.

    simp_sub.
    apply sof_md; done.

(* ext *)
intros G t M p _ IH Hwf.
apply oftp_ext; [].
so (IH Hwf) as HofM.
#2 destruct p.
  apply ofmp_forget; assumption.

  assumption.

(* stat *)
intros G c k Hprink Hwf.
#2 apply (@ofmp_subsume _ _ (sg_stat (single c k))).
  apply ofmp_stat; auto using prink_sound; done.

  apply subs_stat; [].
  apply single_subk; auto using prink_sound; done.

(* lamgn *)
intros G T1 T2 k1 M p HofT1 Hfst _ IH Hwf.
so (IH (wf_md Hwf HofT1 Hfst)) as HofM.
eapply ofmp_lamgn; eauto; [].
destruct p; [apply ofmp_forget |]; assumption.

(* appgn *)
intros G T1 T2 T1' M1 M2 c2 p _ IH1 _ IH2 Hsubs HfstM2 Hwf.
so (IH1 Hwf) as HofM1.
so (IH2 Hwf) as HofM2.
#2 apply (@ofmp_appgn _ T1); auto.
  destruct p; [apply ofmp_forget |]; assumption.

  eapply ofmp_subsume; eauto; done.

(* lamap *)
intros G T1 T2 k1 M HofT1 Hfst _ IH Hwf.
so (IH (wf_md Hwf HofT1 Hfst)) as HofM.
eapply ofmp_lamap; eauto; done.

(* pair *)
intros G p1 p2 T1 T2 M1 M2 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as HofM1.
so (IH2 Hwf) as HofM2.
#2 apply ofmp_pair.
  destruct p1; destruct p2; auto using ofmp_forget; done.

  destruct p1; destruct p2; auto using ofmp_forget; done.

(* unpack *)
intros G p u k t T T' e M _ IH1 Hnorm _ IH2 Hsubs HofT Hwf.
so (IH1 Hwf) as Hofe1.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe1)) as Hofu.
so (headnorm_sound _#4 Hwf Hofu Hnorm) as Hequ.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Hequ ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _; clear l Hofcr.
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
so (IH2 Hwf') as HofM.
eapply ofmp_unpack; eauto using oftp_equiv; [].
#2 apply (@ofmp_subsume _ _ T').
  destruct p; [apply ofmp_forget |]; assumption.

  assumption.

(* lett *)
intros G p t T e M _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hofe.
so (oft_reg _#3 Hwf (oftp_oft _#3 Hwf Hofe)) as Hoft.
so (wf_tm Hwf Hoft) as Hwf'.
so (IH2 Hwf') as HofM.
so (ofm_reg _#4 Hwf' (ofmp_ofm _#4 Hwf' HofM)) as HofT.
eapply ofmp_lett; eauto; [].
simp_sub.
erewrite <- subord_tm_sg; eauto; done.

(* letm *)
intros G T1 T2 T2' k1 M1 M2 p1 p2 _ IH1 Hfst _ IH2 Hsubs HofT2 Hwf.
so (IH1 Hwf) as HofM1.
so (ofm_reg _#4 Hwf (ofmp_ofm _#4 Hwf HofM1)) as HofT1.
so (wf_md Hwf HofT1 Hfst) as Hwf'.
so (IH2 Hwf') as HofM2.
#2 eapply ofmp_letm; eauto.
  destruct p1; [apply ofmp_forget |]; assumption.
 
  #2 apply (@ofmp_subsume _ _ T2').
    destruct p2; [apply ofmp_forget |]; assumption.

    assumption.

(* seal *)
intros G T T' M p _ IH Hsubs Hwf.
so (IH Hwf) as HofM.
so (subs_reg _#3 Hwf Hsubs ander).
apply ofmp_seal; auto; [].
eapply ofmp_subsume; eauto; [].
destruct p; [apply ofmp_forget |]; assumption.

(* epilogue *)
destruct Hind.
split; auto; done.
Qed.


Lemma print_oftp :
  forall G e t,
    wf G
    -> print G e t
    -> oftp G e t.
Proof.
exact (prinx_ofxp andel).
Qed.


Lemma prins_ofmp :
  forall G p M T,
    wf G
    -> prins G p M T
    -> ofmp G p M T.
Proof.
exact (prinx_ofxp ander).
Qed.


Lemma print_sound :
  forall G e t,
    wf G
    -> print G e t
    -> oft G e t.
Proof.
intros.
auto using print_oftp, oftp_oft.
Qed.


Lemma prins_sound :
  forall G p M T,
    wf G
    -> prins G p M T
    -> ofm G p M T.
Proof.
intros.
auto using prins_ofmp, ofmp_ofm.
Qed.


Lemma prins_prink_fst :
  forall G p M T c k,
    wf G
    -> fstm G M c
    -> prins G p M T
    -> prink G c k
    -> fsts T k.
Proof.
intros G p M T c k Hwf HfstM Hprins Hprink.
revert c k Hwf HfstM Hprink.
induct Hprins.

(* var *)
intros G i T c Hindex d l Hwf HfstM Hprink.
so (fstm_fun _#4 HfstM (fstm_var Hindex)); subst d.
assert (exists T' k, T = shift1 T' /\ hsg T' /\ c = var 0 /\ fsts T' k /\ index (S i) G (cl_cn k)) as H.
  clear HfstM Hprink.
  remember (cl_md T c) as C eqn:HeqC.
  revert Hwf HeqC.
  #2 induct Hindex.
    intros C G Hwf ->.
    invert Hwf; [].
    intros G' k T' _ Hof Hfst <- <- <-.
    exists T', k.
    do2 4 split; eauto using ofs_hsg with sequence; done.

    intros i D G C _ IH Hwf HeqC.
    so (IH (wf_tail _ _ Hwf) HeqC) as (T' & k & HeqT & Hhsg & Heqc & Hfsts & Hindex).
    exists T', k.
    do2 4 split; auto.
    apply index_S; auto; done.
destruct H as (T' & k & -> & Hhsg & -> & Hfst & Hindex').
simp_sub.
simp_sub in Hprink.
so (prink_fun _#4 Hprink (prink_var Hindex')); subst l.
apply singles_fst; [].
eapply subst_fsts; eauto; done.

(* star *)
intros G c k Hwf HfstM Hprink.
invert HfstM; [].
intros <-.
so (prink_fun _#4 Hprink prink_star); subst k.
apply fsts_one; done.

(* stat *)
intros G c k Hprink c' k' Hwf HfstM Hprink'.
invert HfstM; [].
intros <-.
so (prink_fun _#4 Hprink Hprink'); subst k'.
apply fsts_stat; done.

(* dyn *)
intros G e t _ c k Hwf HfstM Hprink.
invert HfstM; [].
intros <-.
so (prink_fun _#4 Hprink prink_star); subst k.
apply fsts_dyn; done.

(* lamgn *)
intros G T1 T2 k1 M p _ _ _ _ c k Hwf HfstM Hprink.
invert HfstM; [].
intros <-.
so (prink_fun _#4 Hprink prink_star); subst k.
apply fsts_pigen; done.

(* appgn *)
intros G T1 T2 T1' M1 M2 c2 p _ _ _ _ _ _ c k Hwf HfstM Hprink.
invert HfstM; done.

(* lamap *)
intros G T1 T2 k1 M Hofs Hfst _ IH d k Hwf Hfstapp Hprink.
invert Hfstapp; [].
intros k1' c Hfst' HfstM <-.
so (fsts_fun _#3 Hfst Hfst'); subst k1'; clear Hfst'.
invert Hprink; [].
intros k2 Hofk1 Hprink' <-.
apply fsts_piapp; trivial; [].
#3 sor (refine (IH (shift1 c) (shift1 k2) _ HfstM _)) as Hfst2.
  apply wf_md; auto; done.

  eapply weaken1_prink; eauto; done.

  so (subst_fsts _ _ (dot nonsense id) Hfst2) as Hfst2'.
  simp_sub in Hfst2'.
  assumption.

(* appap *)
intros G p T1 T2 T1' M1 M2 c2 _ IH1 _ IH2 Hsub HfstM2 c k Hwf Hfstapp Hprink.
invert Hfstapp; [].
intros c1 c2' HfstM1 HfstM2' <-.
so (fstm_fun _#4 HfstM2 HfstM2'); subst c2'; clear HfstM2'.
invert Hprink; [].
intros k1 k2 Hprink1 Hofc2 <-.
so (IH1 _ _ Hwf HfstM1 Hprink1) as Hfstpi.
invert Hfstpi; [].
intros Hfst1 Hfst2.
apply subst_fsts; auto; done.

(* pair *)
intros G p1 p2 T1 T2 M1 M2 _ IH1 _ IH2 c k Hwf HfstM Hprink.  
invert HfstM; [].
intros c1 c2 HfstM1 HfstM2 <-.
invert Hprink; [].
intros k1 k2 Hprink1 Hprink2 <-.
so (IH1 _ _ Hwf HfstM1 Hprink1) as Hfst1.
so (IH2 _ _ Hwf HfstM2 Hprink2) as Hfst2.
apply fsts_sigma; eauto using subst_fsts; done.

(* pi1 *)
intros G T1 T2 M _ IH d k Hwf Hfstpi Hprink.
invert Hfstpi; [].
intros c HfstM <-.
invert Hprink; [].
intros k2 Hprink'.
so (IH _ _ Hwf HfstM Hprink') as Hfsts.
invert Hfsts; auto; done.

(* pi2 *)
intros G T1 T2 M c _ IH HfstM d k Hwf Hfstpi Hprink.
invert Hfstpi; [].
intros c' HfstM' <-.
so (fstm_fun _#4 HfstM HfstM'); subst c'.
invert Hprink; [].
intros k1 k2 Hprink' <-.
so (IH _ _ Hwf HfstM Hprink') as Hfsts.
invert Hfsts; eauto using subst_fsts; done.

(* unpack *)
intros G p u k t T T' e M _ _ _ _ _ _ c k0 Hwf HfstM Hprink.
invert HfstM; done.

(* lett *)
intros G p t T e M Hprint _ IH c k Hwf Hfstlet Hprink.
invert Hfstlet; [].
intros HfstM.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprint)) as Hoft.
so (subst_fstm _#5 (subof_tm_utm_tof _ t) HfstM) as HfstM'.
simp_sub in HfstM'.
so (IH _ _ (wf_tm Hwf Hoft) HfstM' (weaken1_prink _#4 Hprink)) as Hfsts.
so (subst_fsts _ _ (dot nonsense id) Hfsts) as Hfsts'.
simp_sub in Hfsts'.
assumption.

(* letm *)
intros G T1 T2 T2' k1 M1 M2 p1 p2 _ _ _ _ _ _ _ c k Hwf HfstM Hprink.
invert HfstM; done.

(* seal *)
intros G T T' M p _ _ _ c k Hwf HfstM Hprink.
invert HfstM; done.
Qed.


Lemma arrow_headnorm :
  forall G t t1 t2,
    wf G
    -> eqc G t (cn_arrow t1 t2) kd_type
    -> exists t1' t2',
         headnorm G t (cn_arrow t1' t2')
         /\ eqc G t1' t1 kd_type
         /\ eqc G t2' t2 kd_type.
Proof.
intros G t t1 t2 Hwf Heqc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hoft1 Hoft2 _.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
so (headnorm_fun _#4 Hnorm' (headnorm_done natk_arrow)); subst p'.
invert Hpeq; [].
intros t1' t2' Haeq1 Haeq2 <-.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf (eqc_reg _#4 Hwf Heqc andel) Hnorm) ander andel)) as (l' & Hofcr').
invert Hofcr'; [].
intros Hoft1' Hoft2' _.
exists t1', t2'.
do2 2 split.
  assumption.

  eapply aeq_sound; eauto using ctxeq_refl with static; done.

  eapply aeq_sound; eauto using ctxeq_refl with static; done.
Qed.


Lemma prod_headnorm :
  forall G t t1 t2,
    wf G
    -> eqc G t (cn_prod t1 t2) kd_type
    -> exists t1' t2',
         headnorm G t (cn_prod t1' t2')
         /\ eqc G t1' t1 kd_type
         /\ eqc G t2' t2 kd_type.
Proof.
intros G t t1 t2 Hwf Heqc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hoft1 Hoft2 _.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
so (headnorm_fun _#4 Hnorm' (headnorm_done natk_prod)); subst p'.
invert Hpeq; [].
intros t1' t2' Haeq1 Haeq2 <-.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf (eqc_reg _#4 Hwf Heqc andel) Hnorm) ander andel)) as (l' & Hofcr').
invert Hofcr'; [].
intros Hoft1' Hoft2' _.
exists t1', t2'.
do2 2 split.
  assumption.

  eapply aeq_sound; eauto using ctxeq_refl with static; done.

  eapply aeq_sound; eauto using ctxeq_refl with static; done.
Qed.


Lemma all_headnorm :
  forall G u k t,
    wf G
    -> eqc G u (cn_all k t) kd_type
    -> exists k' t',
         headnorm G u (cn_all k' t')
         /\ eqk G k' k
         /\ eqc (G; cl_cn k) t' t kd_type.
Proof.
intros G u k t Hwf Heqc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
so (headnorm_fun _#4 Hnorm' (headnorm_done natk_all)); subst p'.
invert Hpeq; [].
intros k' t' Haeq1 Haeq2 <-.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf (eqc_reg _#4 Hwf Heqc andel) Hnorm) ander andel)) as (l' & Hofcr').
invert Hofcr'; [].
intros Hofk' Hoft' _.
exists k', t'.
do2 2 split.
  assumption.

  eapply aeqk_sound; eauto using ctxeq_refl with static; done.

  apply eqc_symm; [].
  eapply aeq_sound; eauto using aeq_symm with static; [].
  apply ctxeq_cn; auto using ctxeq_refl; [].
  apply eqk_symm; [].
  eapply aeqk_sound; eauto using ctxeq_refl with static; done.
Qed.


Lemma exists_headnorm :
  forall G u k t,
    wf G
    -> eqc G u (cn_exists k t) kd_type
    -> exists k' t',
         headnorm G u (cn_exists k' t')
         /\ eqk G k' k
         /\ eqc (G; cl_cn k) t' t kd_type.
Proof.
intros G u k t Hwf Heqc.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (aeq_complete _#4 Hwf Heqc) as Haeq.
invert Haeq; [].
intros p p' Hnorm Hnorm' Hpeq.
so (headnorm_fun _#4 Hnorm' (headnorm_done natk_exists)); subst p'.
invert Hpeq; [].
intros k' t' Haeq1 Haeq2 <-.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf (eqc_reg _#4 Hwf Heqc andel) Hnorm) ander andel)) as (l' & Hofcr').
invert Hofcr'; [].
intros Hofk' Hoft' _.
exists k', t'.
do2 2 split.
  assumption.

  eapply aeqk_sound; eauto using ctxeq_refl with static; done.

  apply eqc_symm; [].
  eapply aeq_sound; eauto using aeq_symm with static; [].
  apply ctxeq_cn; auto using ctxeq_refl; [].
  apply eqk_symm; [].
  eapply aeqk_sound; eauto using ctxeq_refl with static; done.
Qed.


(* no dependent sigmas *)
Inductive ndss : (* sg *) term -> Prop :=
| ndss_one
    : ndss sg_one

| ndss_stat {k}
    : ndss (sg_stat k)

| ndss_dyn {t}
    : ndss (sg_dyn t)

| ndss_pigen {T1 T2}
    : ndss (sg_pigen T1 T2)

| ndss_piapp {T1 T2}
    : ndss T2
      -> ndss (sg_piapp T1 T2)

| ndss_sigma {T1 T2}
    : ndss T1
      -> ndss T2
      -> ndss (sg_sigma T1 (shift1 T2))
.


Lemma subst_ndss :
  forall s T,
    ndss T
    -> ndss (subst s T).
Proof.
intros s T Hnds.
revert s.
induct Hnds; try (intros; simp_sub; eauto using ndss; done); [].

(* sigma *)
intros T1 T2 _ IH1 _ IH2 s.
simp_sub.
rewrite -> subst_compose; [].
apply ndss_sigma; auto; done.
Qed.


Lemma single_ndss :
  forall c T, ndss (singles c T).
Proof.
intros c T.
revert c.
induct T; try (intros; simpl; eauto using ndss; done); [].
(* sigma *)
intros T1 IH1 T2 IH2 c.
simpl.
replsub (dot (cn_pi1 (shift1 c)) sh1) with (compose (dot (cn_pi1 c) id) sh1).
rewrite -> subst_compose; [].
apply ndss_sigma; auto using subst_ndss; done.
Qed.


Lemma prins_ndss :
  forall G M T,
    prins G pure M T
    -> ndss T.
Proof.
intros G M T H.
remember pure as p eqn:Heqp.
revert Heqp.
#7 induct H;
try (intros; eauto using ndss; discriminate).

(* var *)
intros.
apply single_ndss; done.

(* lamap *)
intros G T1 T2 k1 M _ _ _ IH _.
so (IH (eq_refl _)) as H.
so (subst_ndss (dot nonsense id) _ H) as H'.
simp_sub in H'.
apply ndss_piapp; auto; done.

(* appap *)
intros G p T1 T2 T1' M1 M2 c2 _ IH _ _ _ _ ->.
so (IH (eq_refl _)) as H.
invert H; [].
intros H'.
eapply subst_ndss; eauto; done.

(* pair *)
intros G p1 p2 T1 T2 M1 M2 _ IH1 _ IH2 Heq.
#2 destruct p1.
  #2 destruct p2.
    apply ndss_sigma; auto; done.

    discriminate Heq.

  discriminate Heq.

(* pi1 *)
intros G T1 T2 M _ IH Heqp.
so (IH (eq_refl _)) as H.
invert H; [].
auto; done.

(* pi2 *)
intros G T1 T2 M c _ IH _ _.
so (IH (eq_refl _)) as H.
invert H; [].
intros T2' _ H' <-.
simp_sub.
assumption.

(* lett *)
intros G p t T e M _ _ IH ->.
so (IH (eq_refl _)) as H.
eapply subst_ndss; eauto; done.
Qed.


Lemma prin_complete_gen :
  (forall G G' e t,
     ctxsub G G'
     -> oft G' e t
     -> exists t',
          print G e t'
          /\ eqc G t' t kd_type)
  /\
  (forall G G' p M T,
     ctxsub G G'
     -> ofm G' p M T
     -> exists p' T',
          prins G p' M T'
          /\ purity_leq p' p
          /\ subs G T' T).
Proof.
exploit
  (dynamic_ind
     (fun G' e t =>
        forall G,
          ctxsub G G'
          -> exists t',
               print G e t'
               /\ eqc G t' t kd_type)
     (fun G' p M T =>
        forall G,
          ctxsub G G'
          -> exists p' T',
               prins G p' M T'
               /\ purity_leq p' p
               /\ subs G T' T)) as Hind.
     
(* var *)
intros G i t Hindex G' Hsub.
assert (exists t', index i G' (cl_tm t') /\ eqc G' (subst (sh (S i)) t') (subst (sh (S i)) t) kd_type) as (t' & Hindex' & Heqc).
  remember (cl_tm t) as C eqn:HeqC.
  revert G' Hsub HeqC.
  #2 induct Hindex.
    intros ? G0 G' Hsub ->.
    invert Hsub; [].
    intros G0' t' _ Heqc <-.
    eexists; split; eauto with sequence; [].
    replsub kd_type with (shift1 kd_type).
    eapply weaken1_eqc; eauto; done.

    intros i D G0 ? _ IH G Hsub ->.
    so (ctxsub_cons_invert _#3 Hsub) as (G0' & C & -> & Hsub').    
    so (IH _ Hsub' (eq_refl _)) as (t' & Hindex & Heqc).
    exists t'.
    split.
      apply index_S; auto; done.

      so (subst_eqc _#6 (weaken1 _ _ C) Heqc) as Heqc'.
      simp_sub in Heqc'.
      simp_sub.
      assumption.
eauto using print_var; done.

(* unit *)
intros G G' Hsub.
eauto using print with static; done.

(* lam *)
intros G t1 t2 e Hoft1 _ IH G' Hsub.
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) Hoft1) as Hoft1'.
simp_sub in Hoft1'.
#2 exploit (IH (G'; cl_tm t1)) as (t2' & Hprint & Heqc).
  apply ctxsub_tm; auto with static; done.
  
  eexists; split; eauto using print; [].
  so (subst_eqc _#6 (unshift_tm _ _) Heqc) as Heqc'.
  simp_sub in Heqc'.
  eauto with static; done.

(* app *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 G' Hsub.
so (IH1 _ Hsub) as (t & Hprint & Heqc).
so (arrow_headnorm _#4 (ctxsub_wf _ _ Hsub) Heqc) as (t1' & t2' & Hnorm & Heqc1 & Heqc2).
so (IH2 _ Hsub) as (u & Hprint2 & Hequ).
exists t2'.
eauto using print with dynamic static; done.

(* pair *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 G' Hsub.
so (IH1 _ Hsub) as (t1' & Hprint1 & Heqc1).
so (IH2 _ Hsub) as (t2' & Hprint2 & Heqc2).
exists (cn_prod t1' t2').
eauto using print with static; done.

(* pi1 *)
intros G t1 t2 e _ IH G' Hsub.
so (IH _ Hsub) as (t & Hprint & Heqc).
so (prod_headnorm _#4 (ctxsub_wf _ _ Hsub) Heqc) as (t1' & t2' & Hnorm & Heqc1 & Heqc2).
exists t1'.
eauto using print with dynamic static; done.

(* pi2 *)
intros G t1 t2 e _ IH G' Hsub.
so (IH _ Hsub) as (t & Hprint & Heqc).
so (prod_headnorm _#4 (ctxsub_wf _ _ Hsub) Heqc) as (t1' & t2' & Hnorm & Heqc1 & Heqc2).
exists t2'.
eauto using print with dynamic static; done.

(* plam *)
intros G k t e Hofk _ IH G' Hsub.
so (subst_ofk _#4 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) Hofk) as Hofk'.
simp_sub in Hofk'.
#2 exploit (IH (G'; cl_cn k)) as (t' & Hprint & Heqc).
  apply ctxsub_cn; auto with static; done.
exists (cn_all k t').
split; eauto using print with static; done.

(* papp *)
intros G k t e c _ IH Hofc G' Hsub.
so (IH _ Hsub) as (u & Hprint & Heqc).
so (all_headnorm _#4 (ctxsub_wf _ _ Hsub) Heqc) as (k' & t' & Hnorm & Heqk & Heqc').
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) Hofc) as Hofc'.
simp_sub in Hofc'.
exists (subst1 c t').
split.
  eauto using print with static; done.

  replsub kd_type with (subst1 c kd_type).
  eapply subst1_eqc; eauto; done.

(* pack *)
intros G k t c e Hofc _ IH Hoft G' Hsub.
exists (cn_exists k t).
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) Hofc) as Hofc'.
simp_sub in Hofc'.
so (IH _ Hsub) as (u & Hprint & Heqc).
so (ofc_reg _#3 (ctxsub_wf _ _ Hsub) Hofc') as Hofk.
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ (ctxsub_cn Hsub (subk_refl (eqk_refl Hofk))))) Hoft) as Hoft'.
simp_sub in Hoft'.
split.
  eapply print_pack; eauto; done.

  apply eqc_refl; [].
  apply ofc_exists; auto; done.

(* unpack *)
intros G k t s e1 e2 _ IH1 _ IH2 Hofs G' Hsub.
so (IH1 _ Hsub) as (u1 & Hprint1 & Heqc1).
so (exists_headnorm _#4 (ctxsub_wf _ _ Hsub) Heqc1) as (k' & t' & Hnorm & Heqk & Heqt).
#2 exploit (IH2 (G'; cl_cn k'; cl_tm t')) as (u2 & Hprint2 & Heqc2).
  #2 apply ctxsub_tm.
    apply ctxsub_cn; eauto with static; done.

    so (subst_eqc _#6 (subof_conv_0 _#3 Heqk) Heqt) as Heqt'.
    simp_sub in Heqt'.
    assumption.
exists (subst (dot nonsense (dot (inh k') id)) u2).
so (eqk_reg _#3 (ctxsub_wf _ _ Hsub) Heqk) as (Hofk' & _).
so (subst_eqc _#6 (unshift_tm _ _) Heqc2) as Heqc2'.
simp_sub in Heqc2'.
so (subst1_eqc _#6 (inh_ofc _ _ Hofk') Heqc2') as Heqc2''.
simp_sub in Heqc2''.
split; auto; [].
eapply print_unpack; eauto; [].
eapply eqc_trans; eauto; [].
apply eqc_symm; [].
so (weaken1_eqc _ (cl_cn k') _#3 Heqc2'') as Heqc2'''.
simp_sub in Heqc2'''.
assumption.

(* fix *)
intros G t e _ IH G' Hsub.
so (IH _ Hsub) as (u & Hprint & Heqc).
so (ofc_ofcr _#3 (eqc_reg _#4 (ctxsub_wf _ _ Hsub) Heqc ander andel)) as (l & Hofcr).
invert Hofcr; [].
intros _ Hoft _.
eauto using print with static; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 G' Hsub.
so (IH1 _ Hsub) as (t1' & Hprint1 & Heqc1).
so (IH2 _ (ctxsub_tm Hsub Heqc1)) as (t2' & Hprint2 & Heqc2).
exists (subst1 nonsense t2').
split; eauto using print; [].
so (subst_eqc _#6 (unshift_tm _ _) Heqc2) as Heqc2'.
simp_sub in Heqc2'.
assumption.

(* equiv *)
intros G t t' e _ IH Heqc G' Hsub.
so (IH _ Hsub) as (u & Hprint & Heqcu).
so (subst_eqc _#6 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) Heqc) as Heqc'.
simp_sub in Heqc'.
eauto using print with static; done.

(* letm *)
intros G T k t M e _ IH1 HfstT _ IH2 Hoft G' Hsub.
so (IH1 _ Hsub) as (p & T' & Hprins & Hpleq & Hsubs).
so (can_fsts_hsg _ (subs_hsg _#3 Hsubs andel)) as (k' & HfstT').
#2 exploit (IH2 (G'; cl_cn k'; cl_md (shift1 T') (var 0))) as (t' & Hprint & Heqc).
  apply ctxsub_md; auto; done.
so (subs_reg _#3 (ctxsub_wf _ _ Hsub) Hsubs) as (HofT' & _).
so (ofs_fst _#3 HofT' HfstT') as Hofk'.
so (subst_eqc _#6 (unshift_md _#3) Heqc) as Heqc'.
simp_sub in Heqc'.
so (subst1_eqc _#6 (inh_ofc _ _ Hofk') Heqc') as Heqc''.
simp_sub in Heqc''.
exists (subst (dot nonsense (dot (inh k') id)) t').
split; auto; [].
eapply print_letm; eauto; [].
eapply eqc_trans; eauto; [].
apply eqc_symm; [].
so (weaken1_eqc _ (cl_cn k') _#3 Heqc'') as Heqc'''.
simp_sub in Heqc'''.
assumption.

(* ext *)
intros G t M _ IH G' Hsub.
so (IH _ Hsub) as (p & T & Hprins & Hpleq & Hsubs).
so (subs_dyn_invert' _#3 Hsubs) as (t' & -> & Heqc).
eauto using print with static; done.

(* var *)
intros G i T c Hindex G' Hsub.
assert (exists T', index i G' (cl_md T' (var 0)) /\ subs G' (subst (sh (S i)) T') (subst (sh (S i)) T)) as (T' & Hindex' & Hsubs).
  remember (cl_md T c) as C eqn:HeqC.
  revert G' Hsub HeqC.
  #2 induct Hindex.
    intros ? G0 G' Hsub ->.
    invert Hsub; [].
    intros G1' G1 U' U k' k Hsub' Hsubs Hfst' Hfst <- <- <- <-.
    exists (shift1 U').
    split.
      apply index_0; done.

      simp_sub.
      eapply subst_subs; eauto using weaken2; done.

    intros i D G0 ? _ IH G Hsub ->.
    so (ctxsub_cons_invert _#3 Hsub) as (G0' & C & -> & Hsub').    
    so (IH _ Hsub' (eq_refl _)) as (t' & Hindex & Hsubs).
    exists t'.
    split.
      apply index_S; auto; done.

      so (subst_subs _#5 (weaken1 _ _ C) Hsubs) as Hsubs'.
      simp_sub in Hsubs'.
      simp_sub.
      assumption.
#2 do 2 eexists; do2 2 split; eauto using prins_var.
  left; reflexivity.

  simp_sub.
  simp_sub in Hsubs.
  eapply subs_trans; eauto; [].
  so (subs_reg _#3 (ctxsub_wf _ _ Hsub) Hsubs andel) as Hof.
  so (can_fsts _ _ Hof) as (k & Hfst).
  eapply singles_subs; eauto using ctxsub_wf; [].
  #2 apply (ofm_fst _ pure (var i) (subst (sh (S i)) T')); eauto using ctxsub_wf.
    eapply ofm_var; eauto; done.

    replsub (var (1 + i)) with (subst (sh (S i)) (var 0)).
    eapply fstm_var; eauto; done.

(* star *)
intros G G' Hsub.
exists pure, sg_one.
do2 2 split; eauto using prins with static; [].
left; reflexivity.

(* stat *)
intros G c k Hof G' Hsub.
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) Hof) as Hof'.
simp_sub in Hof'.
so (prink_complete _#3 (ctxsub_wf _ _ Hsub) Hof') as (k' & Hprink & Hsubk).
exists pure, (sg_stat k').
do2 2 split; eauto using prins with static; [].
left; reflexivity.

(* dyn *)
intros G e t _ IH G' Hsub.
so (IH _ Hsub) as (t' & Hprint & Heqc).
exists pure, (sg_dyn t').
do2 2 split; eauto using prins with static; [].
left; reflexivity.

(* lamgn *)
intros G T1 T2 k1 M HofT1 Hfst1 _ IH G' Hsub.
so (subst_ofs _#4 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) HofT1) as HofT1'.
simp_sub in HofT1'.
so (ctxsub_md Hsub (subs_refl (eqs_refl HofT1')) Hfst1 Hfst1) as Hsub'.
so (IH _ Hsub') as (p & T2' & Hprins & _ & Hsubs).
exists pure, (sg_pigen T1 (subst1 nonsense T2')).
do2 2 split.
  eapply prins_lamgn; eauto; [].
  simp_sub.
  so (subord_md_sg _#4 (subs_reg _#3 (ctxsub_wf _ _ Hsub') Hsubs andel)) as Heq.
  rewrite <- Heq; [].
  eassumption.

  left; reflexivity.

  #2 eapply subs_pigen; eauto with static.
    so (subst_subs _#5 (unshift_md _#3) Hsubs) as Hsubs'.
    simp_sub in Hsubs'.
    assumption.

    so (subst_ofs _#4 (unshift_md _#3) (subs_reg _#3 (ctxsub_wf _ _ Hsub') Hsubs andel)).
    assumption.

(* appgn *)
intros G T1 T2 M1 M2 c2 _ IH1 _ IH2 HfstM2 G' Hsub.
so (IH1 _ Hsub) as (p1 & T & Hprins1 & _ & Hsubspi).
so (subs_pigen_invert' _#4 (ctxsub_wf _ _ Hsub) Hsubspi) as (T1' & T2' & k1 & -> & Hfst1 & Hsubs1 & Hsubs2).
so (IH2 _ Hsub) as (p2 & U & Hprins2 & Hpleq & HsubsU).
destruct Hpleq as [-> | Heq]; [| discriminate Heq].
so (subst_fstm _#5 (subof_dof_tof _#3 (ctxsub_subof _ _ Hsub)) HfstM2) as HfstM2'.
simp_sub in HfstM2'.
exists impure, (subst1 c2 T2').
do2 2 split.
  eapply prins_appgn; eauto with static; done.

  right; reflexivity.

  eapply subst1_subs; eauto; [].
  so (ofm_fst_o _#5 (ctxsub_wf _ _ Hsub) (prins_sound _#4 (ctxsub_wf _ _ Hsub) Hprins2) HfstM2') as (k & HfstU & Hofc2).
  so (subs_fst _#5 (ctxsub_wf _ _ Hsub) HsubsU HfstU Hfst1) as Hsubk.
  eapply ofc_subsume; eauto; done.

(* lamap *)
intros G T1 T2 k1 M HofT1 Hfst1 _ IH G' Hsub.
so (subst_ofs _#4 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) HofT1) as HofT1'.
simp_sub in HofT1'.
so (ctxsub_md Hsub (subs_refl (eqs_refl HofT1')) Hfst1 Hfst1) as Hsub'.
so (IH _ Hsub') as (p & T2' & Hprins & Hpleq & Hsubs).
destruct Hpleq as [-> | Heq]; [| discriminate Heq].
exists pure, (sg_piapp T1 (subst1 nonsense T2')).
do2 2 split.
  eapply prins_lamap; eauto; [].
  simp_sub.
  so (subord_md_sg _#4 (subs_reg _#3 (ctxsub_wf _ _ Hsub') Hsubs andel)) as Heq.
  rewrite <- Heq; [].
  eassumption.

  left; reflexivity.

  #2 eapply subs_piapp; eauto with static.
    so (subst_subs _#5 (unshift_md _#3) Hsubs) as Hsubs'.
    simp_sub in Hsubs'.
    assumption.

    so (subst_ofs _#4 (unshift_md _#3) (subs_reg _#3 (ctxsub_wf _ _ Hsub') Hsubs andel)).
    assumption.

(* appap *)
intros G p T1 T2 M1 M2 c2 _ IH1 _ IH2 HfstM2 G' Hsub.
so (IH1 _ Hsub) as (p1 & T & Hprins1 & Hpleq & Hsubspi).
so (subs_piapp_invert' _#4 (ctxsub_wf _ _ Hsub) Hsubspi) as (T1' & T2' & k1 & -> & Hfst1 & Hsubs1 & Hsubs2).
so (IH2 _ Hsub) as (p2 & U & Hprins2 & Hpleq' & HsubsU).
destruct Hpleq' as [-> | Heq]; [| discriminate Heq].
so (subst_fstm _#5 (subof_dof_tof _#3 (ctxsub_subof _ _ Hsub)) HfstM2) as HfstM2'.
simp_sub in HfstM2'.
exists p1, (subst1 c2 T2').
do2 2 split.
  eapply prins_appap; eauto with static; done.

  assumption.

  eapply subst1_subs; eauto; [].
  so (ofm_fst_o _#5 (ctxsub_wf _ _ Hsub) (prins_sound _#4 (ctxsub_wf _ _ Hsub) Hprins2) HfstM2') as (k & HfstU & Hofc2).
  so (subs_fst _#5 (ctxsub_wf _ _ Hsub) HsubsU HfstU Hfst1) as Hsubk.
  eapply ofc_subsume; eauto; done.

(* pair *)
intros G p T1 T2 M1 M2 _ IH1 _ IH2 G' Hsub.
so (IH1 _ Hsub) as (p1 & T1' & Hprin1 & Hpleq1 & Hsubs1).
so (IH2 _ Hsub) as (p2 & T2' & Hprin2 & Hpleq2 & Hsubs2).
exists (purity_join p1 p2), (sg_sigma T1' (shift1 T2')).
do2 2 split.
  apply prins_pair; auto; done.

  destruct Hpleq1 as [-> | ->].
    destruct Hpleq2 as [-> | ->].
      left; reflexivity.

      right; reflexivity.

    right; reflexivity.

  so (can_fsts_hsg _ (subs_hsg _#3 Hsubs1 ander)) as (k1 & Hfst1).
  so (can_fsts_hsg _ (subs_hsg _#3 Hsubs1 andel)) as (k1' & Hfst1').
  #2 eapply subs_sigma; eauto.
    eapply subst_subs; eauto using weaken1; done.

    so (subs_reg _#3 (ctxsub_wf _ _ Hsub) Hsubs2 ander) as Hof2.
    eapply subst_ofs; eauto using weaken1; done.

(* pi1 *)
intros G T1 T2 M _ IH G' Hsub.
so (IH _ Hsub) as (p & T & Hprin & Hpleq & Hsubs).
destruct Hpleq as [-> | Heq]; [| discriminate].
so (subs_sigma_invert' _#4 (ctxsub_wf _ _ Hsub) Hsubs) as (T1' & T2' & k1' & -> & HfstT1' & Hsubs1 & Hsubs2).
exists pure, T1'.
do2 2 split; eauto using prins with static; [].
left; reflexivity.

(* pi2 *)
intros G T1 T2 M c _ IH HfstM G' Hsub.
so (IH _ Hsub) as (p & T & Hprin & Hpleq & Hsubs).
destruct Hpleq as [-> | Heq]; [| discriminate].
so (subs_sigma_invert' _#4 (ctxsub_wf _ _ Hsub) Hsubs) as (T1' & T2' & k1' & -> & HfstT1' & Hsubs1 & Hsubs2).
so (subst_fstm _#5 (subof_dof_tof _#3 (ctxsub_subof _ _ Hsub)) HfstM) as HfstM'.
simp_sub in HfstM'.
exists pure, (subst1 (cn_pi1 c) T2').
#2 do2 2 split; eauto using prins.
  left; reflexivity.

  eapply subst1_subs; eauto; [].
  so (prins_sound _#4 (ctxsub_wf _ _ Hsub) Hprin) as Hof'.
  exact (ofm_fst _#6 (ctxsub_wf _ _ Hsub) (ofm_pi1 Hof') (fstm_pi1 HfstM') HfstT1').

(* unpack *)
intros G k t T e M _ IH1 _ IH2 HofT G' Hsub.
so (IH1 _ Hsub) as (u & Hprint & Heqc).
so (exists_headnorm _#4 (ctxsub_wf _ _ Hsub) Heqc) as (k' & t' & Hnorm & Heqk & Heqt).
#2 exploit (IH2 (G'; cl_cn k'; cl_tm t')) as (p & T' & Hprins & _ & Hsubs).
  #2 apply ctxsub_tm.
    apply ctxsub_cn; eauto with static; done.

    so (subst_eqc _#6 (subof_conv_0 _#3 Heqk) Heqt) as Heqt'.
    simp_sub in Heqt'.
    assumption.
so (subst_ofs _#4 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) HofT) as HofT'.
simp_sub in HofT'.
exists impure, T.  
do2 2 split; eauto using prins with static; [].
right; reflexivity.

(* lett *)
intros G p t T e M _ IH1 _ IH2 G' Hsub.
so (IH1 _ Hsub) as (t' & Hprint & Heqc).
so (IH2 _ (ctxsub_tm Hsub Heqc)) as (p' & T' & Hprins & Hpleq & Hsubs).
exists p', (subst1 nonsense T').
do2 2 split; eauto using prins with static; [].
so (subst_subs _#5 (unshift_tm _ _) Hsubs) as Hsubs'.
simp_sub in Hsubs'.
assumption.

(* letm *)
intros G T1 T2 k1 M1 M2 _ IH1 HfstT1 _ IH2 HofT2 G' Hsub.
so (IH1 _ Hsub) as (p1 & T1' & Hprin1 & _ & Hsubs1).
so (can_fsts_hsg _ (subs_hsg _#3 Hsubs1 andel)) as (k1' & HfstT1').
#2 exploit (IH2 (G'; cl_cn k1'; cl_md (shift1 T1') (var 0))) as (p2 & T2' & Hprin2 & _ & Hsubs2).
  apply ctxsub_md; auto; done.
so (subs_reg _#3 (ctxsub_wf _ _ Hsub) Hsubs1) as (HofT1' & _).
so (ofs_fst _#3 HofT1' HfstT1') as Hofk1'.
so (subst_ofs _#4 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) HofT2) as HofT2'.
simp_sub in HofT2'.
exists impure, T2.
do2 2 split; eauto using prins with static; [].
right; reflexivity.

(* seal *)
intros G T M _ IH G' Hsub.
so (IH _ Hsub) as (p & T' & Hprin & _ & Hsubs).
exists impure, T.
so (subs_reg _#3 (ctxsub_wf _ _ Hsub) Hsubs) as (_ & HofT).
do2 2 split; eauto using prins with static; [].
right; reflexivity.

(* extstat *)
intros G k k' M c _ IH HfstM Hofc G' Hsub.
so (IH _ Hsub) as (p & T & Hprins & Hpleq & Hsubs).
destruct Hpleq as [-> | Heq]; [| discriminate Heq].
so (subs_stat_invert' _#3 Hsubs) as (l & -> & Hsubk).
exists pure, (sg_stat l).
#2 do2 2 split; auto.
  left; reflexivity.
apply subs_stat; [].
so (subst_ofc _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) Hofc) as Hofc'.
simp_sub in Hofc'.
so (subst_fstm _#5 (subof_dof_tof _#3 (ctxsub_subof _ _ Hsub)) HfstM) as HfstM'.
simp_sub in HfstM'.
so (prink_complete _#3 (ctxsub_wf _ _ Hsub) Hofc') as (l' & Hprink & Hsubk').
so (prins_prink_fst _#6 (ctxsub_wf _ _ Hsub) HfstM' Hprins Hprink) as Hfsts.
invert Hfsts; [].
intros <-.
assumption.

(* extpi *)
intros G T1 T2 T3 k1 M _ IH1 HfstT1 _ IH2 G' Hsub.
so (IH1 _ Hsub) as (p & T & Hprins & Hpleq & Hsubs).
destruct Hpleq as [-> | Heq]; [| discriminate Heq].
exists pure, T.
#2 do2 2 split; auto.
  left; reflexivity.
so (subs_piapp_invert' _#4 (ctxsub_wf _ _ Hsub) Hsubs) as (T1' & T2' & l & -> & HfstT1' & Hsubs1 & _).
so (fsts_fun _#3 HfstT1 HfstT1'); subst l; clear HfstT1'.
so (subs_reg _#3 (ctxsub_wf _ _ Hsub) Hsubs) as (Hofapp & _).
invert Hofapp; [].
intros k1' HofT1' HfstT1' HofT2'.
eapply subs_piapp; eauto; [].
so (subs_reg _#3 (ctxsub_wf _ _ Hsub) Hsubs1) as (HofT1 & _).
#2 exploit (IH2 (G'; cl_cn k1; cl_md (shift1 T1) (var 0))) as (p & U & Hprins' & Hpleq & Hsubs2).
  apply ctxsub_md; auto with static; done.
cut (U = shift1 T2').
  intros ->.
  so (subst_subs _#5 (unshift_md _#3) Hsubs2) as Hsubs2'.
  simp_sub in Hsubs2'.
  assumption.
eapply (prins_fun _ _ p pure); eauto; [].
#5 exploit (@prins_appap (G'; cl_cn k1; cl_md (shift1 T1) (var 0)) pure (subst (sh 2) T1') (subst (dot (var 0) (sh 3)) T2') (singles (shift1 (var 0)) (shift1 (shift1 T1))) (subst (sh 2) M) (var 0) (var 1)) as H.
  so (weaken1_prins _#4 (cl_md (shift1 T1) (var 0)) (weaken1_prins _#4 (cl_cn k1) Hprins)) as Hprinsx.
  simp_sub in Hprinsx.
  assumption.

  apply prins_var; eauto using index_0; done.

  simp_sub.
  eapply subs_trans; eauto using subst_subs, weaken2; [].
  eapply singles_subs; eauto using subst_fsts, subst_ofs, weaken2, ctxsub_wf with static sequence; done.

  replsub (var 1) with (shift1 (var 0)).
  eapply fstm_var; eauto with sequence; done.

  simp_sub in H.
  rewrite -> subst_dot1_sh in H; [].
  assumption.

(* extsigma *)
intros G T1 T2 M _ IH1 _ IH2 G' Hsub.
so (IH1 _ Hsub) as (p1 & U1 & Hprins1 & _ & Hsubs1).
so (IH2 _ Hsub) as (p2 & U2x & Hprins2 & _ & Hsubs2).
invert Hprins1; [].
intros U2 Hprins <-.
invert Hprins2; [].
intros U1' U2' c Hprins' HfstM <- <-.
so (prins_fun _#6 Hprins Hprins') as (_ & Heq).
injection Heq.
intros <- <-.
so (prins_ndss _#3 Hprins) as Hndss.
invert Hndss; [].
intros U2' _ _ <-.
simp_sub in Hsubs2.
exists pure, (sg_sigma U1 (shift1 U2')).
#2 do2 2 split; auto.
  left; reflexivity.
so (subs_reg _#3 (ctxsub_wf _ _ Hsub) Hsubs1) as (HofU1 & HofT1).
so (can_fsts _ _ HofU1) as (l & HfstU1).
so (can_fsts _ _ HofT1) as (k & HfstT1).
so (subs_reg _#3 (ctxsub_wf _ _ Hsub) Hsubs2) as (_ & HofU2').
#2 eapply subs_sigma; eauto.
  eapply subst_subs; eauto using weaken1; done.

  eapply subst_ofs; eauto using weaken1; done.

(* forget *)
intros G T M _ IH G' Hsub.  
so (IH _ Hsub) as (p & T' & Hprins & Hpleq & Hsubs).
do 2 eexists; do2 2 split; eauto; [].
right; reflexivity.

(* subsume *)
intros G p T T' M _ IH Hsubs G' Hsub.
so (IH _ Hsub) as (p' & T'' & Hprins & Hpleq & Hsubs').
do 2 eexists; do2 2 split; eauto; [].
so (subst_subs _#5 (subof_dof_sof _#3 (ctxsub_subof _ _ Hsub)) Hsubs) as Hsubs''.
simp_sub in Hsubs''.
eapply subs_trans; eauto; done.

(* epilogue *)
destruct Hind.
split; intros; eauto; done.
Qed.


Lemma print_complete :
  forall G e t,
    wf G
    -> oft G e t
    -> exists t',
         print G e t'
         /\ eqc G t' t kd_type.
Proof.
intros G e t Hwf Hof.
exact (prin_complete_gen andel _#4 (ctxsub_refl _ Hwf) Hof).
Qed.


Lemma prins_complete :
  forall G p M T,
    wf G
    -> ofm G p M T
    -> exists p' T',
         prins G p' M T'
         /\ purity_leq p' p
         /\ subs G T' T.
Proof.
intros G p M T Hwf Hof.
exact (prin_complete_gen ander _#5 (ctxsub_refl _ Hwf) Hof).
Qed.


Lemma oft_oftp :
  forall G e t,
    wf G
    -> oft G e t
    -> oftp G e t.
Proof.
intros G e t Hwf Hof.
so (print_complete _#3 Hwf Hof) as (t' & Hprin & Heqc).
so (print_oftp _#3 Hwf Hprin) as Hof'.
eapply oftp_equiv; eauto; done.
Qed.


Lemma ofm_ofmp :
  forall G p M T,
    wf G
    -> ofm G p M T
    -> ofmp G p M T.
Proof.
intros G p M T Hwf Hof.
so (prins_complete _#4 Hwf Hof) as (p' & T' & Hprin & Hpleq & Hsubs).
so (prins_ofmp _#4 Hwf Hprin) as Hof'.
so (ofmp_subsume Hof' Hsubs) as Hof''.
#2 destruct p; destruct p'; try assumption.
  destruct Hpleq; discriminate.

  apply ofmp_forget; assumption.
Qed.


Lemma oft_unique :
  forall G e t t',
    wf G
    -> oft G e t
    -> oft G e t'
    -> eqc G t t' kd_type.
Proof.
intros G e t t' Hwf Hof Hof'.
so (print_complete _#3 Hwf Hof) as (u & Hprint & Heqc).
so (print_complete _#3 Hwf Hof') as (u' & Hprint' & Heqc').
so (print_fun _#4 Hprint Hprint'); subst u'.
eauto using eqc_symm, eqc_trans.
Qed.
