
(* The standardization theorem for the CBV untyped lambda-calculus. *)

Require Import Tactics.
Require Import Relation.
Require Import Syntax.
Require Import Subst.
Require Decidable.
Require Compare_dec.


Inductive hlm : (* lm *) term -> Prop :=
| hlm_var {i}
    : hlm (var i)

| hlm_lam {m}
    : hlm m
      -> hlm (lm_lam m)

| hlm_app {m1 m2}
    : hlm m1
      -> hlm m2
      -> hlm (lm_app m1 m2)
.


Inductive val : (* lm *) term -> Prop :=
| val_var {i}
    : val (var i)

| val_lam {m}
    : val (lm_lam m)
.


Inductive step : (* lm *) term -> (* lm *) term -> Prop :=
| step_app1 {m1 m1' m2}
    : step m1 m1'
      -> step (lm_app m1 m2) (lm_app m1' m2)

| step_app2 {m1 m2 m2'}
    : val m1
      -> step m2 m2'
      -> step (lm_app m1 m2) (lm_app m1 m2')

| step_beta {m1 m2}
    : val m2
      -> step (lm_app (lm_lam m1) m2) (subst1 m2 m1)
.


Inductive reduce : (* lm *) term -> (* lm *) term -> Prop :=
| reduce_var {i}
    : reduce (var i) (var i)

| reduce_lam {m m'}
    : reduce m m'
      -> reduce (lm_lam m) (lm_lam m')

| reduce_app {m1 m1' m2 m2'}
    : reduce m1 m1'
      -> reduce m2 m2'
      -> reduce (lm_app m1 m2) (lm_app m1' m2')

| reduce_beta {m1 m1' m2 m2'}
    : val m2
      -> reduce m1 m1'
      -> reduce m2 m2'
      -> reduce (lm_app (lm_lam m1) m2) (subst1 m2' m1')
.


Inductive ireduce : (* lm *) term -> (* lm *) term -> Prop :=
| ireduce_var {i}
    : ireduce (var i) (var i)

| ireduce_lam {m m'}
    : reduce m m'
      -> ireduce (lm_lam m) (lm_lam m')

| ireduce_app1 {m1 m1' m2 m2'}
    : ~ val m1
      -> ireduce m1 m1'
      -> reduce m2 m2'
      -> ireduce (lm_app m1 m2) (lm_app m1' m2')

| ireduce_app2 {m1 m1' m2 m2'}
    : ireduce m1 m1'
      -> ireduce m2 m2'
      -> ireduce (lm_app m1 m2) (lm_app m1' m2')
.


Inductive equiv : (* lm *) term -> (* lm *) term -> Prop :=
| equiv_symm {m m'}
    : equiv m m'
      -> equiv m' m

| equiv_trans {m1 m2 m3}
    : equiv m1 m2
      -> equiv m2 m3
      -> equiv m1 m3

| equiv_var {i}
    : equiv (var i) (var i)

| equiv_lam {m m'}
    : equiv m m'
      -> equiv (lm_lam m) (lm_lam m')

| equiv_app {m1 m1' m2 m2'}
    : equiv m1 m1'
      -> equiv m2 m2'
      -> equiv (lm_app m1 m2) (lm_app m1' m2')

| equiv_beta {m1 m2}
    : hlm m1
      -> hlm m2
      -> val m2
      -> equiv (lm_app (lm_lam m1) m2) (subst1 m2 m1)
.



Hint Constructors hlm val step reduce ireduce equiv : lambda.


Definition halt (m : term) := exists v, star step m v /\ val v.


Lemma shift_under_hlm :
  forall i j m, hlm m -> hlm (subst (under i (sh j)) m).
Proof.
intros i j m Hhlm.
revert i.
induct Hhlm.

(* var *)
intros n i.
simp_sub.
so (Compare_dec.dec_lt n i) as [Hlt | Hge].
  rewrite -> !project_under_lt; auto; [].
  apply hlm_var; done.

  rewrite -> !project_under_ge; try omega; [].
  simp_sub.
  apply hlm_var; done.

(* lam *)
intros m _ IH i.
simp_sub.
apply hlm_lam; [].
so (IH (S i)) as H.
simp_sub in H.
assumption.

(* app *)
eauto using hlm.
Qed.


Lemma shift_hlm :
  forall i m, hlm m -> hlm (subst (sh i) m).
Proof.
intros i m Hhlm.
so (shift_under_hlm 0 i _ Hhlm) as H.
simp_sub in H.
assumption.
Qed.


Lemma subst_under_hlm :
  forall i m n,
    hlm m
    -> hlm n
    -> hlm (subst (under i (dot m id)) n).
Proof.
intros i m n Hhlmm Hhlmn.
revert i.
induct Hhlmn.

(* var *)
intros j i.
simp_sub.
so (Compare_dec.lt_eq_lt_dec j i) as [[Hlt | Heq]| Hgt].
  rewrite -> !project_under_lt; auto; [].
  apply hlm_var; done.

  subst i.
  rewrite -> !project_under_ge; try omega; [].
  repl (j - j) with 0 by omega.
  simp_sub.
  apply shift_hlm; auto; done.

  rewrite -> !project_under_ge; try omega; [].
  repl (j - i) with (S (j - i - 1)) by omega.
  simp_sub.
  apply hlm_var; done.

(* lam *)
intros n _ IH i.
simp_sub.
apply hlm_lam; [].
so (IH (S i)) as H.
simp_sub in H.
assumption.

(* app *)
eauto using hlm.
Qed.


Lemma subst1_hlm :
  forall m n,
    hlm m
    -> hlm n
    -> hlm (subst1 m n).
Proof.
intros m n Hm Hn.
so (subst_under_hlm 0 _ _ Hm Hn) as H.
simp_sub in H.
assumption.
Qed.


Lemma val_decidable :
  forall m, Decidable.decidable (val m).
Proof.
intros m.
unfold Decidable.decidable.
destruct m; solve [left; auto using val | right; intro H; invert H].
Qed.


Lemma shift_under_val :
  forall i j m, val m -> val (subst (under i (sh j)) m).
Proof.
intros i j m Hval.
induct Hval; try (intros; simp_sub; eauto using val; done); [].
(* var *)
intros n.
simp_sub.
so (Compare_dec.dec_lt n i) as [Hlt | Hge].
  rewrite -> !project_under_lt; auto; [].
  apply val_var; done.

  rewrite -> !project_under_ge; try omega; [].
  simp_sub.
  apply val_var; done.
Qed.


Lemma shift_val :
  forall i m, val m -> val (subst (sh i) m).
Proof.
intros i m Hval.
so (shift_under_val 0 i _ Hval) as H.
simp_sub in H.
assumption.
Qed.


Lemma subst_under_val :
  forall i m n,
    val m
    -> val n
    -> val (subst (under i (dot m id)) n).
Proof.
intros i m n Hvalm Hvaln.
revert i.
induct Hvaln; try (intros; simp_sub; eauto using val; done); [].
(* var *)
intros j i.
simp_sub.
so (Compare_dec.lt_eq_lt_dec j i) as [[Hlt | Heq]| Hgt].
  rewrite -> !project_under_lt; auto; [].
  apply val_var; done.

  subst i.
  rewrite -> !project_under_ge; try omega; [].
  repl (j - j) with 0 by omega.
  simp_sub.
  apply shift_val; auto; done.

  rewrite -> !project_under_ge; try omega; [].
  repl (j - i) with (S (j - i - 1)) by omega.
  simp_sub.
  apply val_var; done.
Qed.


Lemma subst_val :
  forall m n,
    val m
    -> val n
    -> val (subst1 m n).
Proof.
intros m n Hvalm Hvaln.
so (subst_under_val 0 _ _ Hvalm Hvaln) as H.
simp_sub in H.
assumption.
Qed.


Lemma subst_val_inv :
  forall s m, val (subst s m) -> val m.
Proof.
intros s m Hval.
destruct m; try (simp_sub in Hval; invert Hval; done); auto using val.
Qed.


Lemma shift_step :
  forall i n n',
    step n n'
    -> step (subst (sh i) n) (subst (sh i) n').
Proof.
intros i n n' Hstep.
induct Hstep.

(* app1 *)
intros m1 m1' m2 _ IH.
simp_sub.
apply step_app1; auto; done.

(* app2 *)
intros m1 m2 m2' Hval' _ IH.
simp_sub.
apply step_app2; auto; [].
apply shift_val; auto; done.

(* beta *)
intros m1 m2 Hval'.
simp_sub.
exploit (@step_beta (subst (dot (var 0) (sh (S i))) m1) (subst (sh i) m2)) as H.
  apply shift_val; auto; done.

  simp_sub in H.
  assumption.
Qed.


Lemma subst_step :
  forall m n n',
    val m
    -> step n n'
    -> step (subst1 m n) (subst1 m n').
Proof.
intros m n n' Hval Hstep.
induct Hstep.

(* app1 *)
intros m1 m1' m2 _ IH.
simp_sub.
apply step_app1; auto; done.

(* app2 *)
intros m1 m2 m2' Hval' _ IH.
simp_sub.
apply step_app2; auto; [].
apply subst_val; auto; done.

(* beta *)
intros m1 m2 Hval'.
simp_sub.
exploit (@step_beta (subst (dot (var 0) (dot (shift1 m) sh1)) m1) (subst1 m m2)) as H.
  apply subst_val; auto; done.

  simp_sub in H.
  assumption.
Qed.


Lemma step_val_contra :
  forall m m',
    step m m'
    -> val m
    -> False.
Proof.
intros m m' Hstep Hval.
invert Hval; intros; subst; invert Hstep; done.
Qed.


Lemma step_determ :
  forall m m1 m2,
    step m m1
    -> step m m2
    -> m1 = m2.
Proof.
intros m n n' Hstep Hstep'.
revert n' Hstep'.
induct Hstep.

(* app1 *)
intros m1 m1' m2 Hstep IH n' Hstep'.
#3 invert Hstep'.
  intros m1'' Hstep'' <-.
  f_equal; [].
  auto; done.

  intros _ Hval _ _.
  exfalso.
  exact (step_val_contra _ _ Hstep Hval).

  intros n _ <- <-.
  invert Hstep; done.

(* app2 *)
intros m1 m2 m2' Hval1 Hstep IH n' Hstep'.
#3 invert Hstep'.
  intros m1' Hstep1 _.
  exfalso.
  exact (step_val_contra _ _ Hstep1 Hval1).

  intros m2'' _ Hstep'' <-.
  f_equal; [].
  auto; done.

  intros _ Hval2 _ _.
  exfalso.
  exact (step_val_contra _ _ Hstep Hval2).

(* beta *)
intros m1 m2 Hval2 n' Hstep'.
#3 invert Hstep'.  
  intros m1' Hstep1 _.
  exfalso.
  invert Hstep1; done.

  intros m2' _ Hstep2 _.
  exfalso.
  exact (step_val_contra _ _ Hstep2 Hval2).

  intros _ <-.
  reflexivity.
Qed.


Lemma reduce_hlm :
  forall m m',
    reduce m m'
    -> hlm m /\ hlm m'.
Proof.
intros m m' Hred.
induct Hred.

(* var *)
auto using hlm; done.

(* lam *)
intros m m' _ (H & H').
auto using hlm; done.

(* app *)
intros m1 m1' m2 m2' _ (H1 & H1') _ (H2 & H2').
auto using hlm; done.

(* beta *)
intros m1 m1' m2 m2' _ _ (H1 & H1') _ (H2 & H2').
split.
  auto using hlm; done.

  apply subst1_hlm; auto; done.
Qed.


Lemma hlm_reduce :
  forall m,
    hlm m
    -> reduce m m.
Proof.
intros m Hlm.
induct Hlm; intros; auto using reduce.
Qed.


Hint Resolve hlm_reduce : lambda.


Lemma shift_under_reduce :
  forall j i m m',
    reduce m m'
    -> reduce (subst (under j (sh i)) m) (subst (under j (sh i)) m').
Proof.
intros j i m m' Hred.
revert j.
induct Hred.

(* var *)
intros n j.
simp_sub.
so (Compare_dec.dec_lt n j) as [Hlt | Hge].
  rewrite -> !project_under_lt; auto; [].
  apply reduce_var; done.

  rewrite -> !project_under_ge; try omega; [].
  simp_sub.
  apply reduce_var; done.

(* lam *)
intros m m' _ IH j.
simp_sub.
apply reduce_lam; [].
so (IH (S j)) as H.
simp_sub in H.
assumption.

(* app *)
intros m1 m1' m2 m2' _ IH1 _ IH2 j.
simp_sub.
apply reduce_app; auto; done.

(* beta *)
intros m1 m1' m2 m2' Hval _ IH1 _ IH2 j.
simp_sub.
so (shift_under_val j i _ Hval) as Hval'.
so (IH1 (S j)) as Hred1.
simp_sub in Hred1.
so (IH2 j) as Hred2.
so (reduce_beta Hval' Hred1 Hred2) as Hred.
simp_sub in Hred.
assumption.
Qed.


Lemma shift_reduce :
  forall i m m',
    reduce m m'
    -> reduce (subst (sh i) m) (subst (sh i) m').
Proof.
intros i m m' H.
so (shift_under_reduce 0 i _ _ H) as H'.
simp_sub in H'.
assumption.
Qed.


Lemma step_reduce :
  forall m m', hlm m -> step m m' -> reduce m m'.
Proof.
intros m m' Hlm H.
revert Hlm.
induct H.

(* app1 *)
intros m1 m1' m2 _ IH Hlm.
invert Hlm.
intros Hlm1 Hlm2.
apply reduce_app; eauto; [].
apply hlm_reduce; auto; done.

(* app2 *)
intros m1 m2 m2' _ _ IH Hlm.
invert Hlm.
intros Hlm1 Hlm2.
apply reduce_app; eauto; [].
apply hlm_reduce; auto; done.

(* beta *)
intros m1 m2 Hval Hlm.
invert Hlm.
intros Hlm1 Hlm2.
invert Hlm1.
intro Hlm1'.
apply reduce_beta; auto using hlm_reduce; done.
Qed.


Lemma steps_reduces :
  forall m n,
    hlm m
    -> star step m n
    -> star reduce m n.
Proof.
intros m n Hlm Hsteps.
revert Hlm.
induct Hsteps.

(* refl *)
intros m _.
apply star_refl; done.

(* step *)
intros m p n Hmp _ IH Hlm.
so (step_reduce _ _ Hlm Hmp) as Hmp'.
so (reduce_hlm _ _ Hmp') as (_ & Hlm').
eapply star_step; eauto; done.
Qed.


Lemma equiv_hlm :
  forall m n,
    equiv m n
    -> hlm m /\ hlm n.
Proof.
intros m n H.
induct H.

(* symm *)
intros m m' _ (H & H').
auto; done.

(* trans *)
intros m1 m2 m3 _ (H & _) _ (_ & H').
auto; done.

(* var *)
intros i.
auto using hlm; done.

(* lam *)
intros m m' _ (H & H').
auto using hlm; done.

(* app *)
intros m1 n1 m2 n2 _ (H1 & H1') _ (H2 & H2').
auto using hlm; done.

(* beta *)
intros m1 m2 H1 H2 _.
split.
  auto using hlm; done.

  apply subst1_hlm; auto; done.
Qed.


Lemma shift_under_ireduce :
  forall j i m m',
    ireduce m m'
    -> ireduce (subst (under j (sh i)) m) (subst (under j (sh i)) m').
Proof.
intros j i m m' Hred.
revert j.
induct Hred.

(* var *)
intros n j.
simp_sub.
so (Compare_dec.dec_lt n j) as [Hlt | Hge].
  rewrite -> !project_under_lt; auto; [].
  apply ireduce_var; done.

  rewrite -> !project_under_ge; try omega; [].
  simp_sub.
  apply ireduce_var; done.

(* lam *)
intros m m' Hred j.
simp_sub.
apply ireduce_lam; [].
so (shift_under_reduce (S j) i _ _ Hred) as H.
simp_sub in H.
assumption.

(* app1 *)
intros m1 m1' m2 m2' Hnval _ IH1 Hred2 j.
simp_sub.
apply ireduce_app1; auto using shift_under_reduce; [].
contradict Hnval.
eapply subst_val_inv; eauto; done.

(* beta *)
intros m1 m1' m2 m2' _ IH1 _ IH2 j.
simp_sub.
apply ireduce_app2; auto; done.
Qed.


Lemma subst_under_reduce :
  forall i m m' n n',
    val m
    -> reduce m m'
    -> reduce n n'
    -> reduce (subst (under i (dot m id)) n) (subst (under i (dot m' id)) n').
Proof.
intros i m m' n n' Hvalm Hredm Hredn.
revert i.
induct Hredn.

(* var *)
intros j i.
simp_sub.
so (Compare_dec.lt_eq_lt_dec j i) as [[Hlt | Heq]| Hgt].
  rewrite -> !project_under_lt; auto; [].
  apply reduce_var; done.

  subst i.
  rewrite -> !project_under_ge; try omega; [].
  repl (j - j) with 0 by omega.
  simp_sub.
  apply shift_reduce; auto; done.

  rewrite -> !project_under_ge; try omega; [].
  repl (j - i) with (S (j - i - 1)) by omega.
  simp_sub.
  apply reduce_var; done.

(* lam *)
intros n n' _ IH i.
simp_sub.
apply reduce_lam; [].
so (IH (S i)) as H.
simp_sub in H.
assumption.

(* app *)
intros n1 n1' n2 n2' _ IH1 _ IH2 i.
simp_sub.
apply reduce_app; auto; done.

(* beta *)
intros n1 n1' n2 n2' Hvaln2 _ IH1 _ IH2 i.
simp_sub.
so (subst_under_val i m n2 Hvalm Hvaln2) as Hval.
so (IH1 (S i)) as Hred1.
simp_sub in Hred1.
so (IH2 i) as Hred2.
so (reduce_beta Hval Hred1 Hred2) as Hred.
simp_sub in Hred.
assumption.
Qed.


Lemma subst_reduce :
  forall m m' n n',
    val m
    -> reduce m m'
    -> reduce n n'
    -> reduce (subst1 m n) (subst1 m' n').
Proof.
intros m m' n n' Hval Hredm Hredn.
so (subst_under_reduce 0 _#4 Hval Hredm Hredn) as H.
simp_sub in H.
assumption.
Qed.


Lemma reduce_val :
  forall m m',
    val m
    -> reduce m m'
    -> val m'.
Proof.
intros m m' Hval Hred.
invert Hred; intros; subst; invert Hval; intros; auto using val; done.
Qed.


Lemma reduces_val :
  forall m m',
    val m
    -> star reduce m m'
    -> val m'.
Proof.
intros m m' Hval Hreduces.
revert Hval.
induct Hreduces; eauto using reduce_val.
Qed.


Lemma shift_ireduce :
  forall i m m',
    ireduce m m'
    -> ireduce (subst (sh i) m) (subst (sh i) m').
Proof.
intros i m m' Hred.
so (shift_under_ireduce 0 i _ _ Hred) as H.
simp_sub in H.
assumption.
Qed.


Lemma ireduce_reduce :
  forall m m', ireduce m m' -> reduce m m'.
Proof.
intros m m' H.
induct H; intros; auto using reduce.
Qed.


Lemma ireduce_to_val :
  forall m m', ireduce m m' -> val m' -> val m.
Proof.
intros m m' Hred Hval.
invert Hval; intros; subst; invert Hred; intros; subst; auto using val.
Qed.


Lemma ireduces_to_val :
  forall m m', star ireduce m m' -> val m' -> val m.
Proof.
intros m m' Hred Hval.
revert Hval.
induct Hred; eauto using ireduce_to_val.
Qed.


Lemma diamond :
  forall m p q,
    reduce m p
    -> reduce m q
    -> exists n,
         reduce p n
         /\ reduce q n.
Proof.
intros m p q Hredp Hredq.
revert q Hredq.
induct Hredp.

(* var *)
intros i q Hredq.
invert Hredq.
intros <-.
eauto; done.

(* lam *)
intros m p _ IH Q HredQ.
invert HredQ.
intros q Hredq <-.
so (IH _ Hredq) as (n & Hredp' & Hredq').
eauto using reduce; done.

(* app *)
intros m1 p1 m2 p2 Hredp1 IH1 Hredp2 IH2 q Hredq.
#2 invert Hredq.
  (* app/app *)
  intros q1 q2 Hredq1 Hredq2 <-.
  so (IH1 _ Hredq1) as (n1 & Hredp1' & Hredq1').
  so (IH2 _ Hredq2) as (n2 & Hredp2' & Hredq2').
  eauto using reduce; done.

  (* app/beta *)
  intros m3 q3 q2 Hvalm2 Hredq3 Hredq2 <- <-.
  invert Hredp1.
  intros p3 Hredp3 <-.
  so (IH1 _ (reduce_lam Hredq3)) as (n1 & Hredp1' & Hredq1').
  so (IH2 _ Hredq2) as (n2 & Hredp2' & Hredq2').
  invert Hredp1'.
  intros n3 Hredp3' <-.
  invert Hredq1'.
  intro Hredq3'.
  exists (subst1 n2 n3).
  split.
    eapply reduce_beta; eauto using reduce_val; done.

    eapply subst_reduce; eauto using reduce_val; done.

(* beta *)
intros m3 p3 m2 p2 Hvalm2 Hredp3 IH1 Hredp2 IH2 q Hredq.
#2 invert Hredq.
  (* beta/app *)
  intros q1 q2 Hredq1 Hredq2 <-.
  invert Hredq1.
  intros q3 Hredq3 <-.
  so (IH1 _ Hredq3) as (n3 & Hredp3' & Hredq3').
  so (IH2 _ Hredq2) as (n2 & Hredp2' & Hredq2').
  exists (subst1 n2 n3).
  split.
    eapply subst_reduce; eauto using reduce_val; done.

    eapply reduce_beta; eauto using reduce_val; done.

  (* beta/beta *)
  intros q3 q2 _ Hredq3 Hredq2 <-.
  so (IH1 _ Hredq3) as (n3 & Hredp3' & Hredq3').
  so (IH2 _ Hredq2) as (n2 & Hredp2' & Hredq2').
  exists (subst1 n2 n3).
  split; eapply subst_reduce; eauto using reduce_val; done.
Qed.


Lemma confluence :
  forall m p q,
    star reduce m p
    -> star reduce m q
    -> exists n,
         star reduce p n
         /\ star reduce q n.
Proof.
exact (codiamond_star _#3 (codiamond_transpose _#3 (codiamond_star _#3 diamond))).
Qed.


Lemma church_rosser :
  forall m n,
    equiv m n
    -> exists p, star reduce m p /\ star reduce n p.
Proof.
intros m n H.
induct H.

(* symm *)
intros m m' _ (p & H & H').
exists p.
split; auto; done.

(* trans *)
intros m1 m2 m3 _ (p & H1 & H2) _ (q & H2' & H3).
so (confluence _ _ _ H2 H2') as (r & H4 & H5).
exists r; eauto using star_trans; done.

(* var *)
intros i.
exists (var i).
auto using star_refl; done.

(* lam *)
intros m n _ (p & H & H').
exists (lm_lam p).
split.
  apply (star_map' _ _ lm_lam); auto with lambda; done.

  apply (star_map' _ _ lm_lam); auto with lambda; done.

(* app *)
intros m1 n1 m2 n2 Hequiv1 (p1 & H1 & H1') Hequiv2 (p2 & H2 & H2').
so (equiv_hlm _ _ Hequiv1) as (_ & Hn1).
so (equiv_hlm _ _ Hequiv2) as (Hm2 & Hn2).
assert (hlm p1) as Hp1.
  so (star_plus _#4 H1') as [Heq | H].
    subst n1; assumption.

    destruct (plus_plusr _#4 H) as (? & _ & Hred).
    so (reduce_hlm _ _ Hred) as (_ & ?).
    assumption.
exists (lm_app p1 p2).
split.
  #2 apply (star_trans _#3 (lm_app p1 m2)).
    apply (star_map' _ _ (fun m => lm_app m m2)); auto using hlm_reduce with lambda; done.

    apply (star_map' _ _ (fun m => lm_app p1 m)); auto using hlm_reduce with lambda; done.

  #2 apply (star_trans _#3 (lm_app p1 n2)).
    apply (star_map' _ _ (fun m => lm_app m n2)); auto using hlm_reduce with lambda; done.
    
    apply (star_map' _ _ (fun m => lm_app p1 m)); auto using hlm_reduce with lambda; done.

(* beta *)
intros m1 m2 Hm1 Hm2 Hval.
exists (subst1 m2 m1).
split.
  apply star_one; [].
  apply reduce_beta; auto using hlm_reduce; done.

  apply star_refl; done.
Qed.


Definition sreduce (m n : term) :=
  reduce m n /\ exists p, star step m p /\ ireduce p n.


Lemma sreduce_reduce :
  forall m m', sreduce m m' -> reduce m m'.
Proof.
intros m m' H.
destruct H; auto; done.
Qed.


Lemma shift_sreduce :
  forall i m m',
    sreduce m m'
    -> sreduce (subst (sh i) m) (subst (sh i) m').
Proof.
intros i m m' Hsred.
destruct Hsred as (Hred & p & Hstep & Hired).
split.
  apply shift_reduce; auto; done.

  exists (subst (sh i) p).
  split.
    #2 induct Hstep.
      intro; apply star_refl; done.

      intros.
      eapply star_step; eauto using shift_step; done.

    apply shift_ireduce; auto; done.
Qed.


Lemma sreduce_var :
  forall i, sreduce (var i) (var i).
Proof.
intros i.
split.
  apply reduce_var; done.

  exists (var i).
  split.
    apply star_refl; done.

    apply ireduce_var; done.
Qed.


Lemma sreduce_lam :
  forall m m', reduce m m' -> sreduce (lm_lam m) (lm_lam m').
Proof.
intros m m' Hred.
split.
  apply reduce_lam; auto; done.

  exists (lm_lam m).
  split.
    apply star_refl; done.

    apply ireduce_lam; auto; done.
Qed.


Lemma sreduce_app_nval :
  forall m m' n n',
    sreduce m m'
    -> reduce n n'
    -> ~ val m'
    -> sreduce (lm_app m n) (lm_app m' n').
Proof.
intros m m' n n' Hsredm Hredn Hnval.
destruct Hsredm as (Hredm & p & Hstep & Hiredp).
split.
  apply reduce_app; auto; done.

  exists (lm_app p n).
  split.
    #2 induct Hstep.
      auto using star_refl; done.
      
      intros m1 m2 m3 H12 _ H23.
      eapply star_step; eauto; [].
      apply step_app1; auto; done.

    apply ireduce_app1; auto; [].
    contradict Hnval.
    eapply reduce_val; eauto using ireduce_reduce; done.
Qed.


Lemma sreduce_app :
  forall m m' n n',
    sreduce m m'
    -> sreduce n n'
    -> sreduce (lm_app m n) (lm_app m' n').
Proof.
intros m m' n n' Hsredm Hsredn.
so (val_decidable m') as [Hval | Hnval].
  destruct Hsredm as (Hredm & p & Hstepm & Hiredp).
  destruct Hsredn as (Hredn & q & Hstepn & Hiredq).
  split.
    apply reduce_app; auto; done.
  
    exists (lm_app p q).
    split.
      #2 apply (star_trans _#3 (lm_app p n)).
        #2 induct Hstepm.
          auto using star_refl; done.

          intros.
          eapply star_step; eauto; [].
          apply step_app1; auto; done.

        so (ireduce_to_val _ _ Hiredp Hval) as Hvalp.
        #2 induct Hstepn.
          auto using star_refl; done.

          intros.
          eapply star_step; eauto; [].
          apply step_app2; auto; done.

      apply ireduce_app2; auto; done.

  apply sreduce_app_nval; auto using sreduce_reduce; done.
Qed.


Lemma subst_ireduce_sreduce :
  forall m m' n n',
    val n
    -> ireduce m m'
    -> sreduce n n'
    -> sreduce (subst1 n m) (subst1 n' m').
Proof.
intros m m' n n' Hvaln Hiredm Hsredn.
cut (forall i, sreduce (subst (under i (dot n id)) m) (subst (under i (dot n' id)) m')).
  intro Hind.
  so (Hind 0) as H.
  simp_sub in H.
  assumption.
induct Hiredm.

(* var *)
intros j i.
simp_sub.
so (Compare_dec.lt_eq_lt_dec j i) as [[Hlt | Heq]| Hgt].
  rewrite -> !project_under_lt; auto; [].
  apply sreduce_var; done.

  subst i.
  rewrite -> !project_under_ge; try omega; [].
  repl (j - j) with 0 by omega.
  simp_sub.
  apply shift_sreduce; auto; done.

  rewrite -> !project_under_ge; try omega; [].
  repl (j - i) with (S (j - i - 1)) by omega.
  simp_sub.
  apply sreduce_var; done.

(* lam *)
intros m m' Hredm i.
simp_sub.
so (subst_under_reduce (S i) _#4 Hvaln (sreduce_reduce _ _ Hsredn) Hredm) as Hred.
simp_sub in Hred.
apply sreduce_lam; auto; done.  

(* app1 *)
intros p p' q q' Hnvalp Hredp IH Hredq i.
simp_sub.
#3 apply sreduce_app_nval.
  apply IH; done.

  apply subst_under_reduce; auto using sreduce_reduce; done.

  contradict Hnvalp.
  eapply ireduce_to_val; eauto; [].
  eapply subst_val_inv; eauto; done.

(* app2 *)
intros p p' q q' _ IHp _ IHq i.
simp_sub.
apply sreduce_app; auto; done.
Qed.


Lemma subst_sreduce :
  forall m m' n n',
    val n
    -> sreduce m m'
    -> sreduce n n'
    -> sreduce (subst1 n m) (subst1 n' m').
Proof.
intros m m' n n' Hval Hredm Hredn.
destruct Hredm as (Hredm & p & Hstep & Hredp).
split.
  apply subst_reduce; auto using sreduce_reduce; done.

  so (subst_ireduce_sreduce _#4 Hval Hredp Hredn) as Hrednp.
  destruct Hrednp as (_ & q & Hstep' & Hredq).
  exists q.
  split.
    eapply star_trans; eauto; [].
    #2 induct Hstep.
      intro.
      apply star_refl; done.

      intros.
      eapply star_step; eauto; [].
      apply subst_step; auto; done.

    assumption.
Qed.


Lemma split_reduce :
  forall m n,
    reduce m n
    -> exists p,
         star step m p
         /\ ireduce p n.
Proof.
intros m n Hred.
cut (sreduce m n).
  intro Hsred.
  exact (Hsred ander).
induct Hred.

(* var *)
auto using sreduce_var; done.

(* lam *)
auto using sreduce_lam; done.

(* app *)
auto using sreduce_app; done.

(* beta *)
intros m1 m1' m2 m2' Hvalm2 Hredm1 Hsredm1 Hredm2 Hsredm2.
so (subst_sreduce _#4 Hvalm2 Hsredm1 Hsredm2) as Hsred.
destruct Hsred as (_ & p & Hstep & Hired).
split.
  apply reduce_beta; auto; done.

  exists p.
  split.
    eapply star_step; eauto; [].
    apply step_beta; auto; done.

    assumption.
Qed.


Lemma postpone :
  forall m n p,
    ireduce m n
    -> step n p
    -> exists n',
         step m n'
         /\ reduce n' p.
Proof.
intros m n p Hmn Hnp.
revert m Hmn.
induct Hnp.

(* app1 *)
intros n1 p1 n2 _ IH m Hmn.
assert (exists m1 m2, m = lm_app m1 m2 /\ ireduce m1 n1 /\ reduce m2 n2) as (m1 & m2 & -> & Hredm1 & Hredm2).
  invert Hmn; intros; subst; eexists; eauto using ireduce_reduce; done.
so (IH _ Hredm1) as (n1' & Hstep & Hred).
exists (lm_app n1' m2).
split.
  apply step_app1; auto; done.

  apply reduce_app; auto; done.

(* app2 *)
intros n1 n2 p2 Hvaln1 _ IH m Hmn.
assert (exists m1 m2, m = lm_app m1 m2 /\ ireduce m1 n1 /\ ireduce m2 n2) as (m1 & m2 & -> & Hredm1 & Hredm2).
  #2 invert Hmn.
    intros m1 m2 Hnval Hred _ _.
    destruct Hnval.
    eapply ireduce_to_val; eauto; done.

    intros; subst; eauto; done.
so (IH _ Hredm2) as (n2' & Hstep & Hred).
exists (lm_app m1 n2').
split.
  apply step_app2; auto; [].
  eapply ireduce_to_val; eauto; done.

  apply reduce_app; auto using ireduce_reduce; done.

(* beta *)
intros n1 n2 Hvaln2 m Hmn.
assert (exists m1 m2, m = lm_app (lm_lam m1) m2 /\ reduce m1 n1 /\ ireduce m2 n2) as (m1 & m2 & -> & Hredm1 & Hredm2).
  #2 invert Hmn.
    intros m1 m2 Hnval Hred _ _.
    destruct Hnval.
    eapply ireduce_to_val; eauto using val; done.

    intros m1 m2 Hred1 Hred2 <-.
    invert Hred1; intros; subst; eexists; eauto; done.
exists (subst1 m2 m1).
split.
  apply step_beta; [].
  eapply ireduce_to_val; eauto; done.

  apply subst_reduce; auto using ireduce_reduce; [].
  eapply ireduce_to_val; eauto; done.
Qed.


Lemma postpone' :
  forall m n p,
    ireduce m n
    -> step n p
    -> exists n',
         star step m n'
         /\ ireduce n' p.
Proof.
intros m n p Hmn Hnp.
so (postpone _#3 Hmn Hnp) as (n' & Hstep & Hred).
so (split_reduce _ _ Hred) as (n'' & Hstep' & Hred').
exists n''.
split.
  eapply star_step; eauto; done.

  assumption.
Qed.


Lemma bifurcation :
  forall m n,
    star reduce m n
    -> exists p,
         star step m p
         /\ star ireduce p n.
Proof.
intros m n Hred.
induct Hred.

(* refl *)
intros m.
exists m.
split; apply star_refl; done.

(* step *)
intros m q n Hmq _ (r & Hqr & Hrn).
so (split_reduce _ _ Hmq) as (s & Hms & Hsq).
assert (exists t, star step s t /\ ireduce t r) as (t & Hst & Htr).
  clear m n Hms Hmq Hrn.
  revert s Hsq.
  #2 induct Hqr.
    (* refl *)
    intros q s Hsq.
    exists s.
    auto using star_refl; done.

    (* step *)
    intros q1 q2 r Hstep _ IH s Hsq.
    so (postpone' _#3 Hsq Hstep) as (t & Hstep' & Hred).
    so (IH _ Hred) as (t' & Hstep'' & Hred').
    exists t'.
    eauto using star_trans; done.
exists t.
split.
  eapply star_trans; eauto; done.

  eapply star_step; eauto; done.
Qed.


Theorem adequacy_reduce :
  forall m n,
    star reduce m n
    -> (halt m <-> halt n).
Proof.
intros m n Hred.
so (star_plus _#4 Hred) as [Heq | Hred'].
  subst n; split; auto; done.
so Hred' as (x & Hred'' & _).
so (reduce_hlm _ _ Hred'' andel) as Hlmm.
clear x Hred''.
so (plus_plusr _#4 Hred') as (x & _ & Hred'').
so (reduce_hlm _ _ Hred'' ander) as Hlmn.
clear Hred' x Hred''.
split.
  intro Hhalt.
  destruct Hhalt as (v & Hsteps & Hval).
  so (steps_reduces _ _ Hlmm Hsteps) as Hred'.
  so (confluence _#3 Hred Hred') as (q & Hredn & Hredv).
  so (bifurcation _ _ Hredn) as (q' & Hstepn & Hredn').
  exists q'.
  split; auto; [].
  eapply ireduces_to_val; eauto; [].
  eapply reduces_val; eauto; done.

  intro Hhalt.
  destruct Hhalt as (v & Hsteps & Hval).
  assert (star reduce m v) as Hred'.
    eapply star_trans; eauto; [].
    eapply steps_reduces; eauto; done.
  so (bifurcation _ _ Hred') as (q & Hstep' & Hred'').
  exists q.
  split; auto; [].
  eapply ireduces_to_val; eauto; done.
Qed.


Lemma adequacy_equiv :
  forall m n,
    equiv m n
    -> (halt m <-> halt n).
Proof.
intros m n Hequiv.
so (church_rosser _ _ Hequiv) as (p & Hredm & Hredn).
split.
  intros Hhalt.
  eapply adequacy_reduce; eauto; [].
  apply (adequacy_reduce m); auto; done.

  intros Hhalt.
  eapply adequacy_reduce; eauto; [].
  apply (adequacy_reduce n); auto; done.
Qed.
