
Require Import Tactics.
Require Import Arith.
Require Export Subst.
Require Import Sequence.
Require Import Static.



(* Generic development: parameterized over Syntax and Static:STATIC where: *)

Module Type STATIC.

Parameter isclx : term -> Prop.

Axiom subst_cl :
  forall s c, isclx c -> isclx (subst s c).

Notation "G , t" := (cons t G) (at level 65, left associativity).

End STATIC.


(*
To check parametrically, remove Static import and:

Declare Module Static : STATIC.
Import Static.
Definition iscl := isclx.
*)


(* Verify that Static satisfies STATIC. *)
Module StaticCheck <: STATIC := Static.

Goal Static.isclx = iscl.
Proof. auto. Qed.



Definition behaved (R : context -> term -> term -> Prop) : Prop :=
  forall G i c,
    iscl c
    -> index i G c
    -> R G (var i) (subst (sh (S i)) c).


Definition behavedrel (R : context -> term -> term -> term -> Prop) : Prop :=
  forall G i C,
    iscl C
    -> index i G C
    -> R G (var i) (var i) (subst (sh (S i)) C).


Inductive subof (R : context -> term -> term -> Prop) : context -> sub -> context -> Prop :=
| subof_dot {G t s c G'}
    : subof R G s G'
      -> R G t (subst s c)
      -> subof R G (dot t s) (G'; c)

| subof_sh {G n G'}
    : truncate n G G'
      -> subof R G (sh n) G'.


Inductive subrel (R : context -> term -> term -> term -> Prop) : context -> sub -> sub -> context -> Prop :=
| subrel_dot {G t1 t2 s1 s2 C G'}
    : subrel R G s1 s2 G'
      -> R G t1 t2 (subst s1 C)
      -> subrel R G (dot t1 s1) (dot t2 s2) (G'; C)

| subrel_sh {G n G'}
    : truncate n G G'
      -> subrel R G (sh n) (sh n) G'
.


Hint Constructors subof : static.


Lemma subof_mono :
  forall (R : _ -> _ -> _ -> Prop) (R' : _ -> _ -> _ -> Prop),
    (forall G1 t c, R G1 t c -> R' G1 t c)
    -> forall G1 G2 s, subof R G1 s G2 -> subof R' G1 s G2.
Proof.
intros R R' HR G1 G2 s Hs.
induction Hs.
(* dot *)
apply subof_dot.
apply IHHs.
apply HR, H.

(* sh *)
apply subof_sh.
auto.
Qed.


Lemma subrel_mono :
  forall (R : _ -> _ -> _ -> _ -> Prop) (R' : _ -> _ -> _ -> _ -> Prop),
    (forall G1 t t' c, R G1 t t' c -> R' G1 t t' c)
    -> forall G1 G2 s s', subrel R G1 s s' G2 -> subrel R' G1 s s' G2.
Proof.
intros R R' HR G1 G2 s s' Hs.
induct Hs.

(* dot *)
intros.
apply subrel_dot; auto; done.

(* sh *)
intros.
apply subrel_sh; auto; done.
Qed.


Lemma subrel_subof_mono :
  forall (R : _ -> _ -> _ -> _ -> Prop) (R' : _ -> _ -> _ -> Prop),
    (forall G1 t t' c, R G1 t t' c -> R' G1 t c)
    -> forall G1 G2 s s', subrel R G1 s s' G2 -> subof R' G1 s G2.
Proof.
intros R R' HR G1 G2 s s' Hs.
induct Hs.

(* dot *)
intros.
apply subof_dot; eauto; done.

(* sh *)
intros.
apply subof_sh; auto; done.
Qed.


(* Dominates subrel_mono; we keep the other around for backwards compatibility. *)
Lemma subrel_mono' :
  forall (R : _ -> _ -> _ -> _ -> Prop) (R' : _ -> _ -> _ -> _ -> Prop) G1,
    (forall t t' c, R G1 t t' c -> R' G1 t t' c)
    -> forall G2 s s', subrel R G1 s s' G2 -> subrel R' G1 s s' G2.
Proof.
intros R R' G1 HR G2 s s' Hs.
revert HR.
induct Hs.

(* dot *)
intros.
apply subrel_dot; auto; done.

(* sh *)
intros.
apply subrel_sh; auto; done.
Qed.


Hint Resolve subof_mono : substlem.


(* weakening closed *)
Definition wclo (R : context -> term -> term -> Prop) :=
  forall G t C C',
    R G t C
    -> R (G; C') (shift1 t) (shift1 C).

Definition wclorel (R : context -> term -> term -> term -> Prop) :=
  forall G t t' C C',
    R G t t' C
    -> R (G; C') (shift1 t) (shift1 t') (shift1 C).


(* weakening closure: need to strengthen the induction hypothesis for substitution *)
Definition wc (R : context -> term -> term -> Prop) (G : context) (t : term) (c : term) : Prop :=
  forall G' n, truncate n G' G -> R G' (subst (sh n) t) (subst (sh n) c).


Definition wcrel (R : context -> term -> term -> term -> Prop) G t m C :=
  forall G' n, truncate n G' G -> R G' (subst (sh n) t) (subst (sh n) m) (subst (sh n) C).


Lemma wclorel_wcrel :
  forall R,
    wclorel R
    -> forall G t t' c, R G t t' c -> wcrel R G t t' c.
Proof.
intros R Hwclo G t t' C HR.
intros G' n Htrunc.
revert HR.
induct Htrunc.

(* 0 *)
intros G.
simp_sub.
auto; done.

(* S *)
intros n D G G' _ IH HR.
replsub (sh (S n)) with (compose (sh n) sh1).
rewrite -> !subst_compose; [].
apply Hwclo; eauto; done.
Qed.
  

Lemma subof_wclo_weaken1 :
  forall R G1 G2 s C,
    wclo R
    -> subof R G1 s G2
    -> subof R (G1; C) (compose s sh1) G2.
Proof.
intros R G1 G2 s C Hclo Hs.
induct Hs.

(* dot *)
intros G1 t s c G2 _ IH HR.
simp_sub.
apply subof_dot; eauto; [].
rewrite -> subst_compose; [].
apply Hclo; assumption.

(* sh *)
intros G1 n G2 Htrunc.
simp_sub.
apply subof_sh; eauto with sequence; done.
Qed.  
  

Lemma subrel_wclo_weaken1 :
  forall R G1 G2 s s' C,
    wclorel R
    -> subrel R G1 s s' G2
    -> subrel R (G1; C) (compose s sh1) (compose s' sh1) G2.
Proof.
intros R G1 G2 s s' C Hclo Hs.
induct Hs.

(* dot *)
intros G1 t1 t2 s1 s2 C' G2 H H0 H1.
simp_sub.
apply subrel_dot; eauto; [].
rewrite -> subst_compose; [].
apply Hclo; assumption.

(* sh *)
intros G1 n G2 Htrunc.
simp_sub.
apply subrel_sh; eauto with sequence; done.
Qed.


Lemma subof_wclo_weaken :
  forall R G1 G2 G s,
    wclo R
    -> subof R G1 s G
    -> subof R (G1;; G2) (compose s (sh (length G2))) G.
Proof.
intros R G1 G2 G s Hclo Hsubof.
induct G2.

(* nil *)
simpl.
simp_sub.
assumption.

(* cons *)
intros c G2 IH.
simpl length.
rewrite <- app_comm_cons.
so (subof_wclo_weaken1 _#4 c Hclo IH) as Hsubof'.
simp_sub in Hsubof'.
assumption.
Qed.
  

Lemma subof_wclo_bind :
  forall R,
    behaved R
    -> wclo R
    -> forall G G' s c,
         iscl c
         -> subof R G s G'
         -> subof R (G; subst s c) (dot (var 0) (compose s sh1)) (G'; c).
Proof.
intros R HR Hclo G G' s c Hc Hs.
#2 apply subof_dot.
  apply subof_wclo_weaken1; auto; done.

  replace (subst (compose s sh1) c) with (subst (sh 1) (subst s c)) by (simp_sub; auto).
  apply HR; eauto using subst_cl with sequence; done.
Qed.


Lemma subof_wclo_bind2 :
  forall R,
    behaved R
    -> wclo R
    -> forall G G' s c1 c2,
         iscl c1
         -> iscl c2
         -> subof R G s G'
         -> subof R (G; subst s c1; subst (dot (var 0) (compose s sh1)) c2) (dot (var 0) (dot (var 1) (compose s (sh 2)))) (G'; c1; c2).
Proof.
intros R HR Hclo G G' s c1 c2 Hc1 Hc2 Hs.
#2 apply subof_dot.
  repl (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1) by (simp_sub; reflexivity).
  apply subof_wclo_weaken1; auto; [].
  apply subof_wclo_bind; auto; done.

  repl (subst (dot (var 1) (compose s (sh 2))) c2) with (shift1 (subst (dot (var 0) (compose s sh1)) c2)) by (simp_sub; reflexivity).
  apply HR; eauto using subst_cl with sequence; done.
Qed.


Lemma subrel_wclo_bind :
  forall R,
    behavedrel R
    -> wclorel R
    -> forall G G' s s' c,
         iscl c
         -> subrel R G s s' G'
         -> subrel R (G; subst s c) (dot (var 0) (compose s sh1)) (dot (var 0) (compose s' sh1)) (G'; c).
Proof.
intros R HR Hclo G G' s s' c Hc Hs.
#2 apply subrel_dot.
  apply subrel_wclo_weaken1; auto; done.

  replace (subst (compose s sh1) c) with (subst (sh 1) (subst s c)) by (simp_sub; auto).
  apply HR; eauto using subst_cl with sequence; done.
Qed.


Lemma subrel_wclo_bind2 :
  forall R,
    behavedrel R
    -> wclorel R
    -> forall G G' s s' c1 c2,
         iscl c1
         -> iscl c2
         -> subrel R G s s' G'
         -> subrel R (G; subst s c1; subst (dot (var 0) (compose s sh1)) c2) (dot (var 0) (dot (var 1) (compose s (sh 2)))) (dot (var 0) (dot (var 1) (compose s' (sh 2)))) (G'; c1; c2).
Proof.
intros R HR Hclo G G' s s' c1 c2 Hc1 Hc2 Hs.
#2 apply subrel_dot.
  repl (dot (var 1) (compose s (sh 2))) with (compose (dot (var 0) (compose s sh1)) sh1) by (simp_sub; reflexivity).
  repl (dot (var 1) (compose s' (sh 2))) with (compose (dot (var 0) (compose s' sh1)) sh1) by (simp_sub; reflexivity).
  apply subrel_wclo_weaken1; auto; [].
  apply subrel_wclo_bind; auto; done.

  repl (subst (dot (var 1) (compose s (sh 2))) c2) with (shift1 (subst (dot (var 0) (compose s sh1)) c2)) by (simp_sub; reflexivity).
  apply HR; eauto using subst_cl with sequence; done.
Qed.


Lemma wc_mono :
  forall (R1 : _ -> _ -> _ -> Prop) (R2 : _ -> _ -> _ -> Prop),
    (forall G t c, R1 G t c -> R2 G t c)
    -> forall G t c,
         wc R1 G t c -> wc R2 G t c.
Proof.
intros R1 R2 HR.
intros G t c H.
unfold wc.
auto.
Qed.


Hint Resolve wc_mono : substlem.


Lemma wc_elim :
  forall R G t c,
    wc R G t c -> R G t c.
Proof.
intros R G t c H.
unfold wc in H.
assert (truncate 0 G G) as Htrunc by auto with sequence.
pose proof (H _ _ Htrunc) as H'.
simp_sub in H'.
assumption.
Qed.


Lemma wcrel_elim :
  forall R G t1 t2 C,
    wcrel R G t1 t2 C -> R G t1 t2 C.
Proof.
intros R G t1 t2 C H.
#2 sor (refine (H G 0 _)) as H'.
  apply truncate_0; done.
simp_sub in H'.
assumption.
Qed.


Lemma wc_weaken :
  forall R G1 G2 n t c,
    truncate n G1 G2
    -> wc R G2 t c
    -> wc R G1 (subst (sh n) t) (subst (sh n) c).
Proof.
intros R G1 G2 n t c Htr Ht.
have (truncate n G1 G2) as Htr.
have (wc R G2 t c) as Ht.
toshow (wc R G1 (subst (sh n) t) (subst (sh n) c)).
intros G3 m Htr'.
have (truncate m G3 G1) as Htr'.
toshow (R G3 (subst (sh m) (subst (sh n) t)) (subst (sh m) (subst (sh n) c))).
simp_sub.
apply Ht.
replace (n+m) with (m+n) by omega.
eapply truncate_sum; eauto.
Qed.


Lemma wcrel_weaken :
  forall R G1 G2 n t t' c,
    truncate n G1 G2
    -> wcrel R G2 t t' c
    -> wcrel R G1 (subst (sh n) t) (subst (sh n) t') (subst (sh n) c).
Proof.
intros R G1 G2 n t t' c Htr Ht.
intros G3 m Htr'.
simp_sub.
apply Ht; [].
replace (n+m) with (m+n) by omega.
eapply truncate_sum; eauto.
Qed.


Lemma wc_behaved :
  forall R,
    behaved R -> behaved (wc R).
Proof.
intros R HR.
unfold behaved.
intros G i c Hc Hindex.
toshow (wc R G (var i) (subst (sh (S i)) c)).
intros G' n Htrunc.
toshow (R G' (subst (sh n) (var i)) (subst (sh n) (subst (sh (S i)) c))).
repeat simp_sub.
replace (1 + (i + n)) with (S (n+i)) by omega.
apply HR.
apply Hc.
eapply truncate_index_sum; eauto.
Qed.


Lemma wcrel_behaved :
  forall R, behavedrel R -> behavedrel (wcrel R).
Proof.
intros R HR G i C HC Hindex.
intros G' n Htrunc.
simp_sub.
repl (i + n) with (n + i) by omega.
apply HR; auto; [].
eapply truncate_index_sum; eauto; done.
Qed.


Lemma wc_wclo :
  forall R, wclo (wc R).
Proof.
intros R G t C C' HR.
eapply wc_weaken; eauto with sequence.
Qed.


Lemma wcrel_wclorel :
  forall R, wclorel (wcrel R).
Proof.
intros R G t t' C C' HR.
eapply wcrel_weaken; eauto with sequence.
Qed.


Lemma subof_wc_bind :
  forall R,
    behaved R
    -> forall G G' s c,
         iscl c
         -> subof (wc R) G s G'
         -> subof (wc R) (G; subst s c) (dot (var 0) (compose s sh1)) (G'; c).
Proof.
intros R HR G G' s c Hc Hs.
apply subof_wclo_bind; auto using wc_behaved, wc_wclo.
Qed.


Lemma subrel_wc_bind :
  forall R,
    behavedrel R
    -> forall G G' s1 s2 C,
         iscl C
         -> subrel (wcrel R) G s1 s2 G'
         -> subrel (wcrel R) (G; subst s1 C) (dot (var 0) (compose s1 sh1)) (dot (var 0) (compose s2 sh1)) (G'; C).
Proof.
intros R HR G G' s1 s2 C HC Hs.
apply subrel_wclo_bind; auto using wcrel_behaved, wcrel_wclorel.
Qed.


Lemma subof_index :
  forall R,
    behaved R
    -> forall G1 G2 s i c,
         iscl c
         -> subof R G1 s G2
         -> index i G2 c
         -> R G1 (project s i) (subst (compose (sh (S i)) s) c).
Proof.
intros R HR G1 G2 s i c Hc Hs Hindex.
generalize dependent i.
induction Hs; intros i Hindex.
(* dot *)
destruct i.
simp_sub.
inversion Hindex as [x l |]; subst x l c0.
apply H.

rewrite -> project_dot.
rewrite -> compose_sh_succ_dot.
apply IHHs.
inversion Hindex; subst.
assumption.

(* sh *)
simp_sub.
replace (1+(i+n)) with (S (n+i)) by omega.
apply HR.
apply Hc.
eapply truncate_index_sum; eauto.
Qed.


Lemma subrel_index :
  forall R,
    behavedrel R
    -> forall G1 G2 s s' i C,
         iscl C
         -> subrel R G1 s s' G2
         -> index i G2 C
         -> R G1 (project s i) (project s' i) (subst (compose (sh (S i)) s) C).
Proof.
intros R HR G1 G2 s s' i C HC Hs Hindex.
revert i C HC Hindex.
induct Hs.

(* dot *)
intros G1 t1 t2 s1 s2 D G2 _ IH H i C HC Hindex.
simp_sub.
#2 invert Hindex.
  intros <- ->.
  simp_sub.
  assumption.

  intros i' Hindex' <-.
  simp_sub.
  apply IH; auto; done.

(* sh *)
intros G1 n G2 Htrunc i C HC Hindex.
simp_sub.
repl (i + n) with (n + i) by omega.
apply HR; auto; [].
eapply truncate_index_sum; eauto; done.
Qed.


Lemma wc_intro :
  forall R,
    (forall G1 G2 s t c,
       subof (wc R) G1 s G2
       -> R G2 t c
       -> R G1 (subst s t) (subst s c))
    -> forall G t c,
         R G t c -> wc R G t c.
Proof.
intros R HR G t c H.
have (R G t c) as H.
toshow (wc R G t c).
intros G' n Htrunc.
have (truncate n G' G) as Htrunc.
toshow (R G' (subst (sh n) t) (subst (sh n) c)).
thus (subof (wc R) G' (sh n) G) as Hn by subof_sh.
apply (HR _#5 Hn).
assumption.
Qed.


Lemma wcrel_intro :
  forall R,
    (forall G1 G2 s s' t t' C,
       subrel (wcrel R) G1 s s' G2
       -> R G2 t t' C
       -> R G1 (subst s t) (subst s' t') (subst s C))
    -> forall G t t' C,
         R G t t' C -> wcrel R G t t' C.
Proof.
intros R HR G t t' C H.
intros G' n Htrunc.
eapply HR; eauto; [].
apply subrel_sh; auto; done.
Qed.


Lemma subof_id :
  forall R G, subof R G id G.
Proof.
intros R G.
apply subof_sh.
apply truncate_0.
Qed.


Lemma subrel_id :
  forall R G, subrel R G id id G.
Proof.
intros R G.
apply subrel_sh.
apply truncate_0.
Qed.


Hint Resolve subof_id : static.


Lemma subof_sh_all :
  forall R G1 G2 s,
    subof R G1 s G2
    -> compose (sh (length G2)) s = sh (length G1).
Proof.
intros R G1 G2 s Hs.
induction Hs.
(* dot *)
simpl length.
simp_sub.
apply IHHs.

(* sh *)
thus (length G = n + length G') as H' by truncate_length.
simp_sub.
f_equal.
omega.
Qed.


Lemma subof_close :
  forall R G,
    subof R G (sh (length G)) nil.
Proof.
intros R G.
apply subof_sh.
apply truncate_all.
Qed.



Fixpoint subst_ctx (s : sub) (G : context) {struct G} : context :=
  (match G with
   | nil => nil
   | G'; c => subst_ctx s G'; subst (under (length G') s) c
   end).
    

Lemma subst_ctx_cons :
  forall s G c,
    subst_ctx s (G; c) = subst_ctx s G; subst (under (length G) s) c.
Proof.
intros.
reflexivity.
Qed.


Lemma subst_ctx_length :
  forall s G,
    length (subst_ctx s G) = length G.
Proof.
intros s G.
induction G.
(* 0 *)
reflexivity.

(* S *)
simpl.
f_equal; apply IHG.
Qed.


Lemma index_subst_ctx :
  forall G i c s,
    index i G c
    -> index i (subst_ctx s G) (subst (under (length G - i - 1) s) c).
Proof.
intros G i c s Hindex.
induct Hindex.
(* 0 *)
intros c G.
simpl.
replace (length G - 0) with (length G) by omega.
apply index_0.

(* S *)
intros i c G c' _ IH.
thus (i < length (subst_ctx s G)) as Hlt by index_length.
rewrite -> subst_ctx_length in Hlt.
replace (length (G; c) - S i) with (length G - i) by (simpl length; omega).
apply index_S.
exact IH.
Qed.


Lemma subst_ctx_compose :
  forall G s1 s2,
    subst_ctx (compose s2 s1) G = subst_ctx s1 (subst_ctx s2 G).
Proof.
intros G.
induct G.

(* nil *)
intros s1 s2.
reflexivity.

(* cons *)
intros C G IH s1 s2.
simpl.
f_equal; auto; [].
simp_sub.
rewrite -> subst_ctx_length; [].
rewrite -> dist_compose_under; [].
reflexivity.
Qed.


Lemma subst_ctx_id :
  forall G, subst_ctx id G = G.
Proof.
intros G.
induct G; auto; [].
(* cons *)
intros C G IH.
simpl.
f_equal; auto; [].
rewrite -> subst_under_id; auto; done.
Qed.


Lemma subof_under :
  forall R,
    behaved R
    -> (forall G t c, R G t c -> wc R G t c)
    -> forall G1 G1' G2 s,
         Forall iscl G2
         -> subof R G1 s G1'
         -> subof R (G1;; subst_ctx s G2) (under (length G2) s) (G1';; G2).
Proof.
intros R Hbehaved Hwc G1 G1' G2 s Hallcl Hofs.
induction G2 as [| c G2].
(* nil *)
simpl.
assumption.

(* cons *)
simpl subst_ctx.
simpl length.
simpl app.
simp_sub.
apply (subof_mono (wc R)); [apply wc_elim |].
invert Hallcl; intros Hiscl Hallcl'.
apply subof_wc_bind; auto.
apply (subof_mono R); [apply Hwc |].
apply IHG2; auto.
Qed.


Lemma subrel_under :
  forall R,
    behavedrel R
    -> (forall G t t' c, R G t t' c -> wcrel R G t t' c)
    -> forall G1 G1' G2 s s',
         Forall iscl G2
         -> subrel R G1 s s' G1'
         -> subrel R (G1;; subst_ctx s G2) (under (length G2) s) (under (length G2) s') (G1';; G2).
Proof.
intros R Hbehaved Hwc G1 G1' G2 s s' Hallcl Hofs.
induction G2 as [| c G2].
(* nil *)
simpl.
assumption.

(* cons *)
simpl subst_ctx.
simpl length.
simpl app.
simp_sub.
apply (subrel_mono (wcrel R)); [apply wcrel_elim |].
invert Hallcl; intros Hiscl Hallcl'.
apply subrel_wc_bind; auto.
apply (subrel_mono R); [apply Hwc |].
apply IHG2; auto.
Qed.


Lemma weaken1 :
  forall R G A,
    subof R (G; A) sh1 G.
Proof.
intros R G A.
apply subof_sh; auto with sequence.
Qed.


Lemma weaken2 :
  forall R G A B,
    subof R (G; A; B) (sh 2) G.
Proof.
intros R G A B.
apply subof_sh; auto with sequence.
Qed.


Lemma subof_invert_as_dot :
  forall R G1 G2 A s,
    behaved R
    -> iscl A
    -> subof R G1 s (G2; A)
    -> exists m s',
         subof R G1 s' G2
         /\ R G1 m (subst s' A)
         /\ eqsub s (dot m s').
Proof.
intros R G1 G2 A s Hbehaved Hiscl Hs.
#2 invert Hs.
  intros m s' Hs' HR <-.
  exists m, s'.
  auto using eqsub_refl; done.

  intros i Htrunc <-.
  exists (var i), (sh (S i)).
  do2 2 split.
    apply subof_sh; [].
    eapply truncate_succ; eauto; done.
    
    apply Hbehaved; auto; [].
    eapply truncate_index; eauto; done.

    intros n.
    apply eqsub_expand_sh; done.
Qed.
